{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name cdenoise}
{@description Multi channel denoising by low-pass filtering the color information (=channel ratios)}

{$I directives.pas}
uses simage,getparam,sysutils;

var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,iwfn,imfn,ofn   : TAAStr;
    cw,cm,cq,rw,aw      : double;
    pm                  : longint;

    i,j                 : longint;
    fwhm                : ^double;
    fi,wi               : array of TSImg_FPImage;
    mi                  : array of TSImg_RDImage;
    mii                 : array of CSImg_MetaInfo;
    wi2                 : TSImg_FPImage;
    mi2                 : TSImg_RDImage;
    s                   : shortstring;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='';
        ifn.size:=0;    defparam('-i','Input file, can be specified multiple times',ifn);       deffmt('-i','<infile>');
        iwfn.size:=0;   defparam('-w','Input weight, can be specified multiple times',iwfn);    deffmt('-w','<infile>');
        imfn.size:=0;   defparam('-m','Input interpolation mask. Can be specified multiple times',imfn);    deffmt('-m','<infile>');
        ofn.size:=0;    defparam('-o','Output file, can be specified multiple times',ofn);      deffmt('-o','<outfile>');
        cw:=2.0;        defparam('-c','Color (channel ratio) FWHM relative to maximum FWHM (default: 2)',cw);
        cm:=-1;         defparam('-cm','Color (channel ratio) FWHM for dark regions relative to maximum FWHM (default: Same value as -c)',cm);
        cq:=0.8;        defparam('-cq','Reference quantile if cm is different from c. In that case color FWHM for values larger than the cq-quantile \\'+
                                       'will be smaller then -c and values larger than (cq/4)-quantile will be equal to cm. Default: 80%)',cq);
        rw:=1.25;       defparam('-r','Maximum channel FWHM relative to input FWHM. Default 1.25',rw);
        aw:=1.5;        defparam('-a','Maximum absolute channel FWHM. The maximum of the values given by -a and -r is used. Default 1.5',aw);
        pm:=1;          defparam('-p','Pattern for interpolation mask. Interpolation if mask and pattern is nonzero. Default: 1',pm);
                        defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': ifn.size='+int2str(ifn.size)+' iwfn.size='+int2str(iwfn.size)+' ofn.size='+int2str(ifn.size)+' owfn.size='+int2str(ifn.size)+' cw='+real2str(cw)+' cm='+real2str(cm)+' cq='+real2str(cq)+' rw='+real2str(rw)+' aw='+real2str(aw)+' pm='+int2str(pm));
    if ofn.size=0 then
    begin
        cc.warning(cc.name+': No output defined -- nothing to do');
        SImg_Halt;
    end;
    if ifn.size=0 then
    begin
        cc.warning(cc.name+': No input defined -- nothing to do');
        SImg_Halt;
    end;
    if (ifn.size<>iwfn.size) then cc.error(cc.name+': Number of input images must be equal to number of input weights');
    if (ofn.size>ifn.size) then cc.warning(cc.name+': More outputs than inputs defined, ignoring supernumerary outputs');
    if cm<1e-3 then cm:=cw;

    { input }
    getmem(fwhm,ifn.size*sizeof(double));
    setlength(fi,ifn.size);
    setlength(wi,ifn.size);
    setlength(mi,ifn.size);
    setlength(mii,ifn.size);
    for i:=0 to ifn.size-1 do
    begin
        fi[i].pt:=nil;
        wi[i].pt:=nil;
        mi[i].pt:=nil;
        mii[i]:=CSImg_MetaInfo.create();
        if (i<imfn.size) and (pm and 255<>0) then s:=', interpolation mask '+imfn.aa[i]
        else s:='';
        cc.Info(cc.name+': Reading input image ' + ifn.aa[i] + s + ' and weight ' + iwfn.aa[i]);

        SImg_ReadImageFile(cc, ifn.aa[i], fi[i], mii[i]);
        if (fi[i].nx<>fi[0].nx) or (fi[i].ny<>fi[0].ny) or (fi[i].nc<>1)  then cc.error(cc.name+': Invalid dimensions of input image'+ int2str(i+1) +' (' + ifn.aa[i] + ')');
        if not mii[i].get('FWHMPIX',fwhm[i]) then fwhm[i]:=-1;

        SImg_ReadImageFile(cc, iwfn.aa[i], wi[i]);
        if (wi[i].nx<>fi[0].nx) or (wi[i].ny<>fi[0].ny) or (fi[i].nc<>1) then cc.error(cc.name+': Invalid dimensions of weight image ' + int2str(i+1) +' (' + iwfn.aa[i] + ')');

        if s<>'' then
        begin
            SImg_ReadImageFile(cc, imfn.aa[i], mi[i]);
            if mi[i].fmt<>SImg_RawByte then cc.error(cc.name+': Invalid dimensions of interpolation mask ' + int2str(i+1) +' (' + imfn.aa[i] + ')');
            if (mi[i].nx<>fi[0].nx) or (mi[i].ny<>fi[0].ny) or (mi[i].nc<>1) then cc.error(cc.name+': Invalid dimensions of interpolation mask ' + int2str(i+1) +' (' + imfn.aa[i] + ')');
            wi2:=wi[i];
            mi2:=mi[i];
            for j:=0 to mi2.size-1 do
                if mi2.ptb[j] and pm<>0 then wi2.pt[j]:=0;
//          SImg_WriteImageFile(cc, 'cdnoise-w-'+int2str(i)+'.fits',wi2);
        end;
    end;

    { denoising }
    cc.Info('cdenoise: Denoising using ' + int2str(cc.threads) + ' threads');
    SImg_CDenoise(cc, fi,wi, fwhm, ofn.size, cq,cw,cm, rw,aw);

    for i:=0 to iwfn.size-1 do
        SImg_Free(cc, wi[i]);

    { output }
    for i:=0 to ifn.size-1 do
    begin
        if i<ofn.size then
        begin
            if mi[i].pt=nil then cc.Info('cdenoise: Writing output image '+ofn.aa[i])
            else begin
                cc.Info('cdenoise: Interpolation and writing output image '+ofn.aa[i]);
                SImg_InterpolateMasked(cc, fi[i], mi[i], pm);
            end;
            mii[i].add('FWHMPIX',real2str(fwhm[i],-4), 'Star FWHM in pixels');
            SImg_WriteImageFile(cc, ofn.aa[i], fi[i], mii[i]);
        end;
        SImg_Free(cc, mi[i]);
        SImg_Free(cc, fi[i]);
        mii[i].destroy;
    end;
end.
