{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name rawcorrect}
{@description Raw data correction, i.e. image calibrations done before dark current subtraction}
{@longdescription Supported calibration methods are: bias subtraction, power supply noise reduction, photon counting and background noise calculation.
All processing methods (except of bias subtraction) require exact bias correction, i.e. a pixel value of 0 corresponds to zero electrons (no dark current,
no photon current). Power supply noise reduction (-s) may cause artifacts in the order of 1/100 electrons. That becomes relevant if huge amounts of images
are added and some care should be taken if reduced data is archived.}

{$I directives.pas}
uses simage,getparam,;
var par             : tparams;
    cc              : CSImg_Config;
    fi,bi,zi,oi     : TSImg_FPImage;
    ri              : TSImg_RDImage;
    mi              : CSImg_MetaInfo;
    mask            : ^byte;  { 1:overflow, 2:underflow }
    fni,fno,fnb,fnz : ansistring;
    sa,key,comment  : ansistring;
    fn1             : ansistring;
    pc,q,nq         : boolean;
    ma              : TAAStr;
    i,j,ja,jz       : longint;
    r,s,rn,gain     : double;
    dmin,dmax,fmin  : double;

begin
    par.init;
    par.paramusage:='<infile1> [<infile2> [<outfile>]]';
    fni:='';      par.defparam('<infile>','Input image (default: stdin)',fni);
    fno:='';      par.defparam('<outfile>','Output image (default: stdout)',fno);
    fnb:='';      par.defparam('-b','Bias image',fnb);                                          par.deffmt('-b','<infile>');

    fmin:=-1;     par.defparam('-s','Power supply noise reduction (parameter is the frequency limit, recommended: 25..100).',fmin);

    pc:=false;    par.defparam('-c','Photon counting. Output will be integer image. Requires exact zero current calibration (true if -s is not specified).',pc);
    nq:=false;    par.defparam('-cn','Enable noise quantization for photon counting (increases compression at costs of 0.5% SNR loss).',nq);
    fnz:='';      par.defparam('-d','Dark current image (only used to increase the accuracy of power supply noise reduction)' ,fnz);    par.deffmt('-d','<infile>');
    fn1:='';      par.defparam('-1','Flatfield image (only used to increase the accuracy of power supply noise reduction)' ,fn1);       par.deffmt('-1','<infile>');

    ma.size:=0;   par.defparam('-m','Add meta info (can be defined multiple times)',ma);        par.deffmt('-m','<key>=<value>[/<comment>]');

                  par.defparam('-h','Help',par.helpf);
                  par.defparam('-o','Output image',fno);        par.deffmt('-o','<outfile>');
    par.run;
    par.done;

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('rawcorrect: in='+fni+' out='+fno+' bias='+fnb+' dark='+fnz+' flat='+fno+' fmin='+real2str(fmin)+' c='+boolean2str(pc)+' nq='+boolean2str(nq));

    { input image }
    fi.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info('rawcorrect: Reading input image '+fni);
    SImg_ReadImageFile(cc,fni,fi,mi);

    { process meta infos }
    for i:=0 to ma.size-1 do
    begin
        sa:=trim(ma.aa[i]);
        j:=pos('=',sa);
        if (j<=1) or (j>=length(sa)) then cc.warning('rawcorrect: Invalid format of -m '+sa+', expected -mr <key>=<value>[/<comment>]')
        else
        begin
            key:=trim(copy(sa,1,j-1));
            sa:=trim(copy(sa,j+1,length(sa)-j));
            j:=1;
            q:=false;
            while (j<=length(sa)) and (q or (sa[j]<>'/')) do
            begin
                if sa[j]='''' then q:=not q;
                inc(j);
            end;
            if j=1 then cc.warning('rawcorrect: Invalid format of -m '+sa+', <value> must not be empty, syntax: -m <key>=<value>[/<comment>] ')
            else
            begin
                comment:='';
                if j<length(sa) then comment:=trim(copy(sa,j+1,length(sa)-j));
                if j=0 then j:=length(sa)+1;
                mi.add(key, trim(copy(sa,1,j-1)), comment);
            end;
        end;
    end;

    if not mi.get('DATAMIN',dmin) then dmin:=-1e301;
    if not mi.get('DATAMAX',dmax) then dmax:=1e301;
    if not mi.get('RDNOISE',rn) then rn:=-1;
    if not mi.get('EGAIN',gain) then cc.error('rawcorrect: ADC gain (key "EGAIN") not defined');
    if pc and (gain>=0.50001) then cc.warning('rawcorrect: Minimum feasible gain for photon counting is 0.5 electron/ADU');
    if pc and (rn>0) and (rn*gain>=1.0001) then cc.warning('rawcorrect: Read out noise larger than 1 electron, photon counting does reduce read out noise and causes information loss');
    if (rn<0) then cc.warning('rawcorrect: Read out noise not defined. Unable to estimate background noise');

    { calculate overflow / underflow mask }
    mask:=nil;
    if (dmin>-1e300) or (dmax<1e300) then
    begin
        getmem2(mask,fi.size);
        for i:=0 to fi.size-1 do
        begin
            r:=fi.pt[i];
            if r<dmin then j:=2
            else if r>dmax then j:=1
            else j:=0;
            mask[i]:=j;
        end;
    end;

    { bias }
    bi.pt:=nil;
    if fnb<>'' then
    begin
        cc.Info('rawcorrect: Reading and subtracting bias image '+fni);
        SImg_ReadImageFile(cc,fnb,bi,nil);
        if bi.size<>fi.size then cc.error('Invalid size of bias image');
        s:=0;
        for i:=0 to fi.size-1 do
        begin
            r:=bi.pt[i];
            s+=r;
            fi.pt[i]-=r;
        end;
        s/=fi.size;
        dmin-=s;
        dmax-=s;
        SImg_Free(cc,bi);
    end;

    { reduce power supply noise }
    if fmin>0 then
    begin
        { dark current image }
        zi.pt:=nil;
        if fnz<>'' then
        begin
            cc.Info('rawcorrect: Reading dark current image '+fnz);
            SImg_ReadImageFile(cc,fnz,zi,nil);
        end;
        { flatfield image }
        oi.pt:=nil;
        if fn1<>'' then
        begin
            cc.Info('rawcorrect: Reading flat field '+fn1);
            SImg_ReadImageFile(cc,fn1,oi,nil);
        end;
        cc.Info('rawcorrect: Reducing power supply noise');
        SImg_PSDenoise(cc,fi,zi,oi,fi,fmin,false);
        if zi.pt<>nil then SImg_Free(cc,zi);
        if oi.pt<>nil then SImg_Free(cc,oi);
    end;

    { calculate background noise }
    if (rn>=0) then
    begin
        r:=SImg_BGNoise(cc,fi,rn,gain);
        cc.info('rawcorrect: Background noise: '+real2str(r,-1)+' ADU');
        mi.add('BGNOISE',real2str(r,15),'Background noise');
    end;

    { photon counting + output conversion }
    ri.pt:=nil;
    if pc then
    begin
        cc.Info('rawcorrect: Apply photon counting');
        ri.pt:=nil;
        SImg_PhotonCount(cc,fi,ri,gain,nq);
        ja:=65535;
        jz:=0;
        for i:=0 to ri.size-1 do
        begin
            j:=ri.ptw[i];
            if j<ja then ja:=j;
            if j>jz then jz:=j;
        end;
        if ja<1 then ja:=1;
        if jz>65534 then jz:=65534;
        dmin:=(ja-0.5)*ri.cf+ri.co;
        dmax:=(jz+0.5)*ri.cf+ri.co;
        mi.add('DATAMIN',real2str((ja-0.5)*ri.cf+ri.co,15),'Minimum valid value');
        mi.add('DATAMAX',real2str((jz+0.5)*ri.cf+ri.co,15),'Maximum valid value');
        if mask<>nil then
        begin
            ja-=2;
            jz+=2;
            for i:=0 to fi.size-1 do
            begin
                j:=mask[i];
                if j=1 then ri.ptw[i]:=jz
                else if j=2 then ri.ptw[i]:=ja;
            end;
        end;
        cc.Info('rawcorrect: Writing output image '+fno);
        SImg_WriteImageFile(cc,fno,ri,mi);
        SImg_Free(cc,ri);
    end else
    begin
        if mask<>nil then
        begin
            r:=dmin - (abs(dmin)+1e-6)*1e-6;
            s:=dmax + (abs(dmax)+1e-6)*1e-6;
            for i:=0 to fi.size-1 do
            begin
                j:=mask[i];
                if j=1 then fi.pt[i]:=s
                else if j=2 then fi.pt[i]:=r;
            end;
        end;
        if dmin>-1e300 then mi.add('DATAMIN',real2str(dmin,15),'Minimum valid value');
        if dmax< 1e300 then mi.add('DATAMAX',real2str(dmax,15),'Maximum valid value');
        cc.Info('rawcorrect: Writing output image '+fno);
        SImg_WriteImageFile(cc,fno,fi,mi);
    end;
    if mask<>nil then freemem2(mask);
    SImg_Free(cc,fi);
    cc.destroy;
end.
