{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name sdbadd}
{@description Adds one or multiple images to stack database}

{$I directives.pas}
uses simage,getparam,sysutils;

const unmask_bncorr = 1;        { correction of minimum square background noise }
      unmask_lamb   = 2;        { smoothing factor relative to background noise in order to prevent overweighting small values }
      mosaic_approx = 2;        { boundary overweight in mosaic mode }
//      mosaic_approx = 1;      { boundary overweight in mosaic mode }

var par                         : tparams;
    cc                          : CSImg_Config;
    sfn,ofn,flfn,rfn,cs,cso     : ansistring;
    wb,dt,dr,we,bw              : double;
    fmfn,afn,wfn,ifn            : TAAStr;
    bg,sa,c                     : longint;
    cr                          : boolean;

    fs,ffl,fr,fi,fw,ffm,fj,fk,fa: TSImg_FPImage;
    sl                          : TSImg_RDImage;
    res                         : TSImg_AlignmentParams;
    ms,ma,mi,mw,mf,msl          : CSImg_MetaInfo;
    si                          : TSImg_SDBInfo;
    i,j,x,y                     : longint;
    s,t,adr,bn,rbn,sbn          : double;
    dtd,dtd2                    : boolean;
    bi                          : TSImg_BayerInfo;
    ps,ca                       : ansistring;
    wcs                         : TSImg_WCS;
    bbw,bbw2                    : ^double;      { buffer for reference source list output }

begin
{define sdbadd_debug}
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='<stack database> [<FWHM limits>]';
        sfn:='';        defparam('<stack database>','Stack database',@sfn,tpastr,rapdreq);
        flfn:='';       defparam('<FWHM limits>','FWHM limit image',flfn);
        afn.size:=0;    defparam('-a','Alignment data, can be specified multiple times',afn);   deffmt('-a','<image>');
        ifn.size:=0;    defparam('-i','Input image, can be specified multiple times',ifn);      deffmt('-i','<image>');
        wfn.size:=0;    defparam('-w','Pixel weight image, optional, can be specified multiple times',wfn);                             deffmt('-w','<image>');
        fmfn.size:=0;   defparam('-f','FWHM map, can be specified multiple times',fmfn);        deffmt('-f','<image>');
        rfn:='';        defparam('-r','Reference image for 2nd and later pass.',rfn);           deffmt('-r','<image>');
        bg:=1;          defparam('-b','Background subtraction (0:none, 1:constant, 2:gradient, 3:curved gradient, 4+:1+ radial polynomials; default:1)',bg);
        wb:=-1;         defparam('-wb','Underweight boundaries. Parameter is relative to half of the smaller image image dimensions, i.e. range is 0..1.',wb);
        cr:=false;      defparam('-cr','Factor calibration using reference image. Only recommended if brightness of reference image is reliable\'+
                                       'which is usually not true if pixel weight with photon noise estimation or unmask functionality was used.',cr);
        cs:='';         defparam('-cs','Factor calibration using reference source list which is only recommended if pixel weight is used\'+
                                       'If both, -cr and -cs are given tho optimal method is used',cs); deffmt('-cs','<source list>');
        cso:='';        defparam('-cso','Output corrected reference source list for next pass.',cso);           deffmt('-cso','<source list>');
        dt:=-1;         defparam('-u','Mask out unlikely pixels in 2nd and later passes. Parameter is the threshold as noise factor. Feasible > 1',dt);
        dr:=1.5;        defparam('-ur','Absolute value is the unmask threshold relative to reference. If <0 lower threshold is sqrt(1+sqr(ur)). Default: 1.5',dr);
        sa:=2;          defparam('-s','Area brightness compensation. 0:none, 1:WCS, 2:determinant of alignment matrix. Default 2.\\' +
                                      'If brightness is calibrated using stars, the scaling factor includes everything but\\' +
                                      '(angular) pixel area. Due to flatfield corrections the correct correction factor is\\' +
                                      '<pixel area of reference image in center> / <pixel area of destination image in center>\\' +
                                      'Numerator is normalized to 1"^2 for method 1. Method 2 is more or less inaccurate due to\\'+
                                      'distortions. If WCS data is available, method 1 should be used.',sa);
        we:=0;          defparam('-we','Brightness weight exponent, feasible: 0..1. Should be 0 for final pass at max. SNR. Default: 0',we);
        c:=1;           defparam('-c','Use this color if Bayer patter image is recognized. Default: 1 which is the 2nd color',c);
        bw:=0.1;        defparam('-bw','Weight factor for undefined pixels of Bayer pattern images. Default: 0.1', bw);

        ofn:='';        defparam('-o','Output stack database (default: same as input)',ofn);            deffmt('-o','<outfile>');
        run;
        done;
    end;
    cc.debug(cc.name+': sfn='+sfn+' flfn='+flfn+' rfn='+rfn+' ofn='+ofn+' afn.size='+int2str(afn.size)+' ifn.size='+int2str(ifn.size)+' wfn.size='+int2str(wfn.size)+' fmfn.size='+int2str(fmfn.size)+
        ' bg='+int2str(bg)+' wbr='+real2str(wb)+' dt='+real2str(dt)+' cr='+boolean2str(cr)+' cs='+cs+' cso='+cso+' sa='+int2str(sa)+' we='+real2str(we)+' c='+int2str(c)+' bw='+real2str(bw));
    if wb>1 then wb:=1;

    if afn.size<>ifn.size then cc.error(cc.name+': Number of alignment data sets must be equal to number if input images');
    if ifn.size<1 then
    begin
        cc.warning('No input defined -- nothing to do');
        SImg_Halt;
    end;
    if (wfn.size<>0) and (wfn.size<>ifn.size) then cc.error(cc.name+': Number of weight images must be equal to number if input images');
    if (flfn='') and (fmfn.size<>0) then
    begin
        fmfn.size:=0;
    end;
    if (flfn<>'') and (fmfn.size=0) then
    begin
        cc.error(cc.name+': No FWHM maps given, ignoring FWHM limit');
        flfn:='';
    end;
    if (fmfn.size<>0) and (fmfn.size<>ifn.size) then cc.error(cc.name+': Number FWHM maps must be equal to number if input images');
    dtd:=dt>0.1;

    { factor calibration initialization }
    sl.pt:=nil;  bbw:=nil;  bbw2:=nil;
    msl:=CSImg_MetaInfo.create();
    if (cso<>'') and (cs='') then cc.error(cc.name+': Output reference source list (parameter -cso) requires input reference source list (parameter -cs)');
    if (cs<>'') and cr then
    begin
        if wfn.size<>0 then
        begin
            cc.info(cc.name+': Choosing factor calibration using reference source list');
            cr:=false
        end
        else begin
            cc.info(cc.name+': Choosing factor calibration using reference image');
            if cso='' then cs:='';      { (input) reference source list is required for output source list }
        end;
    end;
    if cr and (wfn.size<>0) then cc.warning(cc.name+': Factor calibration using reference image (option -cr) is only recommended if brightness of reference image is reliable which is usually not true if pixel weight with photon noise estimation or unmask functionality was used');
    if cs<>'' then
    begin
        if (not cr) and (wfn.size=0) then cc.warning(cc.name+': Factor calibration using reference source list may not work properly without pixel weight');
        cc.Info(cc.name+': Reading reference source list '+cs);
        SImg_ReadImageFile(cc,cs,sl,msl);
        if not (SImg_Recover(cc,sl) and (sl.fmt=SImg_RawDouble) and (sl.nx>=5) and (sl.nc=1)) then cc.error(cc.name+': Invalid reference source list');
    end;
    if cso<>'' then
    begin
        getmem2(bbw,sl.ny*3*sizeof(double));
        getmem2(bbw2,sl.ny*3*sizeof(double));
        for i:=0 to sl.ny*3-1 do
            bbw[i]:=0;
    end;

    { stack database }
    ms:=CSImg_MetaInfo.create();
    fs.pt:=nil;
    cc.Info(cc.name+': Reading stack database '+sfn);
    SImg_ReadImageFile(cc,sfn,fs,ms);
    if (fs.nc and 1)<>0 then cc.error(cc.name+': Invalid stack database: '+sfn);
    SImg_StackDB_GetInfo(cc, si, ms, fs.nx,fs.ny);
    if fs.nc<4 then
    begin
        if (fmfn.size>0) or (flfn<>'') then cc.warning(cc.name+': Stack database contains less than 2 FWHM ranges, ignoring FWHM parameters');
        fmfn.size:=0;
        flfn:='';
    end;

    { reference image }
    fr.pt:=nil;
    if rfn<>'' then
    begin
        cc.Info(cc.name+': Reading reference image '+rfn);
        SImg_ReadImageFile(cc,rfn,fr);
        if (fr.nc<>1) or (fr.nx<>fs.nx) or (fr.ny<>fs.ny) then cc.error(cc.name+': Invalid dimensions of reference image');
    end;

    { fwhm limits }
    ffl.pt:=nil;
    if flfn<>'' then
    begin
        cc.Info(cc.name+': Reading FWHM limits '+flfn);
        SImg_ReadImageFile(cc,flfn,ffl);
    end;

    { input images }
    ma:=CSImg_MetaInfo.create();
    mi:=CSImg_MetaInfo.create();
    mw:=CSImg_MetaInfo.create();
    mf:=CSImg_MetaInfo.create();
    fi.pt:=nil;  ffm.pt:=nil;
    fw.pt:=nil; fw.nx:=0;
    for i:=0 to 3 do
        bi.cf[i]:=1;
//    for i:=0 to 10 do
    for i:=0 to ifn.size-1 do
    begin
        { alignment data }
        cc.Info(cc.name+': Reading alignment data '+afn.aa[i]);
        ma.reset;
        SImg_RHeader(cc,afn.aa[i],ma);
        if not SImg_ReadAlignmentParams(cc,res,ma) then cc.error(cc.name+': Align parameters not set');

        { input image }
        cc.Info(cc.name+': Reading input image '+ifn.aa[i]);
        mi.reset;
        bi.pattern[0,0]:=-1;    // means no Bayer pattern
        SImg_ReadImageFile(cc,ifn.aa[i],fi,mi);
        if fi.nc<>1 then cc.Info(cc.name+': input image contains more than one color');
        if mi.get('BAYERPAT',ps) then
        begin
            cc.Info(cc.name+': Interpolate color ' + int2str(c));
            bi.pattern:=SImg_BayerParsePattern(cc,trimquotes(ps));
            if ((sl.pt<>nil) or cr) and (bw<0.699) then j:=0                    { in calibration mode non-linear interpolation is only exact if bw=1 }
            else if (fr.pt<>nil) and (dr<=-simg_mina) and (bw<0.101) then j:=-1 { q25 interpolation in unsymmetrical mode }
            else j:=1;                                                          { otherwise median }
            SImg_BayerInterpolation(cc,bi, fi,fi, j, c);
{$ifdef sdbadd_debug}
            if i=0 then SImg_WriteImageFile(cc,'sdbadd-bayer-i.fits',fi);
{$endif}
        end;

        { fwhm map }
        mf.reset;
        if fmfn.size>0 then
        begin
            cc.Info(cc.name+': Reading FWHM map '+fmfn.aa[i]);
            SImg_ReadImageFile(cc,fmfn.aa[i],ffm,mf);
        end;

        { pixel weight }
        if wfn.size>0 then
        begin
            cc.Info(cc.name+': Reading pixel weight image '+wfn.aa[i]);
            mw.reset;
            SImg_ReadImageFile(cc,wfn.aa[i],fw,mw);

            if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1) then cc.error(cc.name+': Invalid dimensions of weight image '+wfn.aa[i]);
            if bi.pattern[0,0]>=0 then
            begin
                SImg_BayerInterpolation(cc,bi, fw,fw, -2, c);
{$ifdef sdbadd_debug}
                if i=0 then SImg_WriteImageFile(cc,'sdbadd-bayer-w.fits',fw);
{$endif}
            end;
        end
        else begin
            if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) then SImg_New(cc, fw, fi.nx,fi.ny,1, false);
            for j:=0 to fw.size-1 do
                fw.pt[j]:=1;
        end;

        { calibration using reference source list }
        if sl.pt<>nil then
        begin
            if mi.get('FWHMPIX',s) or mf.get('FWHMPIX',s) then cc.Info(cc.name+': Read PSF FWHM width from meta data: ' +real2str(s,-1))
            else begin
                s:=SImg_GetSharpnessFWHM(cc,fi);
                cc.Info(cc.name+': Auto-detected FWHM: ' +real2str(s,-1));
            end;
            s:=SImg_StackDB_calib(cc, fi,fw, sl, res, s, bbw2);
            if not cr then
            begin
                if s<simg_mina then cc.warning(cc.name+': Could not determine image factor')
                else begin
                    cc.Info(cc.name+': Image factor adjusted to '+real2str(s,5)+' (was '+real2str(res.fact,5)+select(res.fact>simg_mina,'  quotient: '+real2str(s/res.fact,5),'')+')');
                    res.fact:=s;
                end;
            end;
            { update data for output reference list }
            if (bbw<>nil) and (res.fact>simg_mina) then
            begin
                s:=1/res.fact;
                for j:=0 to sl.ny-1 do
                begin
                    x:=j*3;
                    t:=bbw2[x+2]*s;         { weight }
                    bbw[x  ]+=bbw2[x  ]*t;  { brightness }
                    bbw[x+1]+=bbw2[x+1]*t;  { background }
                    bbw[x+2]+=s*t;
                end;
            end;
        end;

        { weight exponent }
        if res.fact>=simg_mina then res.weight*=exp(ln(res.fact)*we);

        { area correction }
        if sa=1 then
        begin
            if not SImg_GetWCSData(cc, ma, wcs) then cc.error(cc.name+': WCS data required for area brightness compensation method 1');
            SImg_GetPixelScale(wcs, s,t);
            res.fact/=abs(s*t);
        end
        else if sa=2 then res.fact*=res.b11*res.b22-res.b12*res.b21;

        { shift reference image }
        fa.pt:=nil;
        if (fr.pt<>nil) and ((bg>0) or dtd) then
        begin
            cc.Info(cc.name+': Shift reference image using '+int2str(cc.threads) + ' threads');
            SImg_New(cc, fa, fi.nx,fi.ny,1);    // undefined pixels will be 0
            if not SImg_StackDB_shift(cc, si, fr,fa,fw, res) then
            begin
                SImg_Free(cc, fa);
                continue;
            end;
{$ifdef sdbadd_debug}
            if i=0 then SImg_WriteImageFile(cc,'sdbadd-fa.fits',fa);
{$endif}

            { brightness calibration }
            if cr then
            begin
                s:=SImg_ImgFactor(cc, fi,fa,fw, bg);
                if s<1e-2 then cc.warning(cc.name+': Unlikely image calibration factor: 1/'+real2str(s,6)+'. Ignoring it.')
                else begin
                    for j:=0 to fa.size-1 do
                        fa.pt[j]*=s;
                    s:=1/s;
                    cc.Info(cc.name+': Image factor adjusted to '+real2str(s,5)+' (was '+real2str(res.fact,5)+'  quotient: '+real2str(s,5)+')');
                    res.fact*=s;
                    res.weight*=exp(ln(s)*we);
                end;
            end;
        end;

        { gradient and background removal }
        if (bg>0) or dtd then
        begin
            if fa.pt<>nil then
            begin
                { data for noise estimation using reference image }
                bn:=-1; adr:=0; dtd2:=false; sbn:=0;
                if dtd then
                begin
                    if not mi.get('EGAIN',adr) then cc.warning('No gain (key "EGAIN") defined') { adr is gain }
                    else if adr<1e-50 then cc.warning('Gain to small: '+real2str(adr))
                    else begin
                        if not mi.get('BGNOISE',bn) then bn:=-1;
                        if mi.get('RDNOISE',t) then
                        begin
                            if (mi.get('BAYERPAT',ps) or (bn<simg_mina)) and mi.get('DCNOISE',s) then
                            begin
                                { calculate BG nouse using dark current noise, adr is gain }
                                bn:=sqrt(sqr(s)+sqr(SImg_BGNoise(cc,fi,t,adr)));
                                cc.info(cc.name+': estimated background noise: '+real2str(bn,5)+' image units = '+real2str(bn*adr,-1)+' e-');
//                              cc.info('bn='+real2str(bn)+' dn='+real2str(s)+' rn='+real2str(r)+'  gain='+real2str(adr)+'e/ADU  q30='+real2str(SImg_GetQuantile(cc,fi,0.3))+'ADU' );
                            end
                            else if bn<simg_mina then
                            begin
                                cc.warning('Neither background noise (key "BGNOISE") nor dark current noise (key "DCNOISE") defined, using read out noise*4');
                                bn:=t*4;
                            end;
                        end;
                        if bn>=0 then
                        begin
                            adr:=1/adr;
                            dtd2:=true;
                            sbn:=sqr(bn)*(1+unmask_bncorr);     { square background noise limit }
                        end
                        else cc.warning('Neither read out noise (key "RDNOISE") nor background noise (key "BGNOISE") defined');
                    end;
                end;

                { brightness adjustment and background removal }
                fk.pt:=nil;
                s:=SImg_ImgApprox(cc, fi,fa,fw, fk,fj, bg, 0, false);  { fk: residual, fj: corrected image }
{               if abs(s-1)>1e-3 then
                begin
                    cc.Info(cc.name+': Image factor adjusted to: '+real2str(res.fact*s,5)+' (was '+real2str(res.fact,5)+')');
                    res.fact*=s;
                    if s>=simg_mina then res.weight*=exp(ln(s)*we);
                    s:=1/s;
                    for j:=0 to fa.size-1 do
                        fa.pt[j]*=s;
                end; }
{$ifdef sdbadd_debug}
                if i=0 then SImg_WriteImageFile(cc,'sdbadd-fres.fits',fk);
{$endif}

                { mask out unlikely values. fi: estimated square noise}
                ca:='';  { makes the compiler happy }
                if dtd2 and (wfn.size>0) then
                begin
                    { This is the recommended algorithm. Estimates variance of pixel from from weight image and subtracting reg(fk)*adr,
                      where reg is a smooth regularization function which avoids artifacts caused by preferring smaller values (which result in greater variance estimation) }
                    ca:='Mask out unlikely values. Noise noise estimation using reference image, weight, gain and readout noise';
                    bn*=unmask_lamb;
                    if dr<=-simg_mina then bn*=0.5;
                    rbn:=1/bn;
                    for j:=0 to fi.size-1 do
                    begin
                        t:=fw.pt[j];
                        if t>simg_mina then
                        begin
                            s:=fk.pt[j];
                            t:=2/t;                             { maximum underestimation factor of square noise estimator is 2 }
                            if s>bn then t-=(2*s-bn)*adr        { Smooth function in order to reduce effect of dark values being overweighted }
                            else if s>0 then t-=sqr(s)*rbn*adr;
                            if t<sbn then t:=sbn;
                            fi.pt[j]:=t;
                        end
                        else fi.pt[j]:=simg_large;
                    end;
                end else if dtd2 then
                begin
                    { It is assumed that the background noise contains photon noise of background, dark current noise and read out noise, i.e. variance can be estimated from background noise and fr2.
                      This algorithm does not take into account noise caused by pixel defects, amplifier glow, scaling (vignetting), ... and it therefore not the preferred one. }
                    ca:='Mask out unlikely values. Noise estimation using reference image, background, gain and readout noise';
                    for j:=0 to fi.size-1 do
                    begin
                        t:=fi.pt[j]-fk.pt[j];
                        if t>0 then fi.pt[j]:=sbn+t*adr                 { square noise }
                        else fi.pt[j]:=sbn;
                    end;
                end else if dtd and (wfn.size>0) then
                begin
                    { Estimates variance of pixel from from weight image. This prefers higher values (which result on overestimation of noise)
                      and is therefore only used if no better algorithm is available.}
                    ca:='Mask out unlikely values. Noise estimation using weight';
                    for j:=0 to fi.size-1 do
                    begin
                        t:=fw.pt[j];
                        if t>simg_mina then fi.pt[j]:=2/t
                        else fi.pt[j]:=simg_large;
                    end;
                end
                else begin
                    SImg_Free(cc,fi);
                    fi.pt:=nil;
                end;

                if fi.pt=nil then SImg_Move(cc,fj,fi)
                else begin
                    s:=SImg_StackDB_unmask(cc, fj,fk,fi,fw, dt,dr,adr);
                    cc.info(cc.name+': '+ca+': '+real2str(s/fw.size*100,-2)+'%');
                    SImg_Move(cc,fj,fi);
                end;
                SImg_Free(cc,fk);

                { final background approximation with boundary overweight }
                for j:=0 to fi.size-1 do
                    fa.pt[j]:=fi.pt[j]-fa.pt[j];
                SImg_ApproxMosaic(cc, fa,fw, bg,mosaic_approx,false,true);
                for j:=0 to fi.size-1 do
                    fi.pt[j]-=fa.pt[j];
                SImg_Free(cc,fa);
            end
            else begin
                cc.Info(cc.name+': Gradient and/or background removal '+int2str(cc.threads) + ' threads');
                SImg_BGApprox3(cc, fi,fw, bg);
            end;
        end;
        if fj.pt<>nil then SImg_free(cc,fj);

        { underweight boundaries }
        if wb>1e-3 then SImg_UnderweightBoundaries(cc,fw,wb);

        { underweight undefined pixels }
        if (bi.pattern[0,0]>=0) and (bw<=0.999) then
        begin
            j:=0;
            for y:=0 to fi.ny-1 do
                for x:=0 to fi.nx-1 do
                begin
                    if bi.pattern[y and 1, x and 1]<>c then fw.pt[j]*=bw;
                    j+=1;
                end;
        end;

{$ifdef sdbadd_debug}
        if i=0 then SImg_WriteImageFile(cc,'sdbadd-fw.fits',fw);
{$endif}

        { add }
        cc.Info(cc.name+': Add image to stack database using '+int2str(cc.threads) + ' threads');
        SImg_StackDB_add(cc, si, fs,fi,fw, ffl,ffm, res);

//      if i=0 then cc.error(cc.name+': No error');
    end;
    SImg_Free(cc,fi);
    SImg_Free(cc,fw);
    SImg_Free(cc,ffl);
    SImg_Free(cc,ffm);

    { adjust factor }
    if fr.pt<>nil then
    begin
        SImg_StackDB_ConvertValWeight(cc, fs, fi,fw);
        s:=SImg_ImgApprox2(cc, fr, fi,fw);
        if (s<0.95) or (s>2) then cc.warning('Invalid correction factor: '+real2str(s,-3)+': Omit correction')
        else begin
            cc.info(cc.name+': Correction factor: '+real2str(s,-3));
            for i:=0 to fs.size div 2-1 do
                fs.pt[i*2]*=s;
        end;
    end;
    SImg_Free(cc,fr);
    SImg_Free(cc,fi);
    SImg_Free(cc,fw);

    { output }
    if ofn='' then ofn:=sfn;
    cc.Info(cc.name+': Write stack database '+ofn);
    SImg_WriteImageFile(cc,ofn,fs,ms);
    SImg_Free(cc,fs);

    { corrected reference source list }
    if cso<>'' then
    begin
        cc.Info(cc.name+': Write corrected reference source list '+cso);
        for i:=0 to sl.ny-1 do
        begin
            x:=i*3;
            s:=bbw[x+2];
            if s>simg_mina then
            begin
                s:=1/s;
                y:=i*sl.sl;
                sl.ptd[y+2]:=bbw[x  ]*s;        { brightness }
                sl.ptd[y+3]:=bbw[x+1]*s;        { background }
                sl.ptd[y+4]:=1;
            end;
        end;
        freemem2(bbw,sl.ny*3*sizeof(double));
        freemem2(bbw2,sl.ny*3*sizeof(double));
        SImg_WriteImageFile(cc,cso,sl,msl);
    end;
    SImg_Free(cc, sl);

    ms.destroy;
    ma.destroy;
    mi.destroy;
    mw.destroy;
    mf.destroy;
    msl.destroy;
    cc.destroy;
end.
