{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SImg_Denoise ****************************************************}
{*************************************************************************}
{ Denoising
    fi     : input and output image
    fwhm   : FWHM of input image
    rad    : radius parameter, relative to fwhm, >=0, recommended: 0..2
    sp     : structure preservation. 0..1
    thr    : denoising threshold as quantile
}
procedure SImg_Denoise(cc:CSImg_Config; fi:TSImg_FPImage; fwhm,rad,sp,thr:double);
var fl,fh,fj    : TSImg_FPImage;
    i,j         : longint;
    r,s,rthr    : double;
    mf          : boolean;
begin
    SImg_IncDebugLevel;
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_Denoise: fi='+SImg_InfoStr2(fi)+' fwhm='+real2str(fwhm)+' rad='+real2str(rad)+' sp='+real2str(sp)+' thr='+real2str(thr));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Denoise: Invalid image: fi');
    if fwhm<1 then fwhm:=1;
    if rad<0 then rad:=0;
    rad:=fwhm*(1.5+rad);        { rad >= 1.5 }
    if sp<0 then sp:=0;
    if sp>1 then sp:=1;
    sp:=sp*2*fwhm/(pi*rad) + 0.5*(1-sp);
    mf:=(0.5-sp)*pi*rad<0.5;    { median filter if (0.5-sp)*pi*sqr(rad) < 0.5*rad }
    if mf then sp:=0.5;

    { filters }
    fl.pt:=nil; fh.pt:=nil;
    cc.info(cc.name+': '+int2str(round(sp*100))+'%-quantile filter with radius '+real2str(rad,-1));
    SImg_Filter(cc, fi, fl, rad, -4, 1, sp);
    if mf then fh:=fl
    else begin
        cc.info(cc.name+': '+int2str(round((1-sp)*100))+'%-quantile filter with radius '+real2str(rad,-1));
        SImg_Filter(cc, fi, fh, rad, -4, 1, 1-sp);
    end;

    { threshold }
    fj.pt:=nil;
    SImg_New(cc, fj, fi.nx,fi.ny,fi.nc, false);
    j:=0;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i];
        s:=fh.pt[i];
        if r>s then fj.pt[i]:=r-s
        else begin
            s:=fl.pt[i];
            if r<s then fj.pt[i]:=s-r
            else begin
                fj.pt[i]:=0;
                j+=1;
            end;
        end;
    end;
    thr:=SImg_GetQuantile(cc, fj, (j+(fi.size-j)*thr)/fi.size );
    SImg_Free(cc,fj);
    if thr<simg_mina then cc.error(cc.name+': constant image');

    { noise filter }
    rthr:=1/thr;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i];
        s:=fh.pt[i];
        if r>s then
        begin
            r:=(r-s)*rthr;
            if r>=2 then continue;
            if r<1 then r*=0.5*sqr(r)
            else r*=1-0.5*sqr(2-r);
            fi.pt[i]:=s+r*thr;
            continue;
        end;
        s:=fl.pt[i];
        if r<s then
        begin
            r:=(s-r)*rthr;
            if r>=2 then continue;
            if r<1 then r*=0.5*sqr(r)
            else r*=1-0.5*sqr(2-r);
            fi.pt[i]:=s-r*thr;
        end;
    end;

    SImg_Free(cc,fl);
    SImg_Free(cc,fh);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_CDenoise ***************************************************}
{*************************************************************************}
{ Multi channel denoising by low-pass filtering the color information (=channel ratios).
  fi        : input/output channel images
  wi        : input/output pixel weight images
  ifwhm     : input FWHM, <=0 means undefined (will be calculated automatically if undefined)
  no        : maximum number of output channels to calculate
  qr        : reference quantile
  rcfwhm1   : output color FWHM relative to maximum input FWHM at qr-quantile
  rcfwhm0   : output color FWHM relative to maximum input FWHM below (qr/4)-quantile
  rfwhm     : maximum relative (to the input FWHM) FWHM which is taken into account during channel denoising. Values <=0 are interpreted as infinity.
  afwhm     : maximum absolute FWHM which is taken into account during channel denoising. The maximum max(afwhm,ifwhm[i]*rfwhm) will be used.
}
procedure SImg_CDenoise(cc:CSImg_Config; var afi,awi:array of TSImg_FPImage; ifwhm:pdouble; no:longint; qr,rcfwhm1,rcfwhm0,rfwhm,afwhm:double);
const reg_nc = 1e-2; { overweight for same channel Must be>0 }
var afwi2,awi2,aoi      : array of TSImg_FPImage;
    fi,wi,fwi2,wi2      : TSImg_FPImage;
    fwi2i,wi2i,fwi3,wi3 : TSImg_FPImage;
    rr                  : TSImg_RDImage;
    i,j,k,size          : longint;
    r,s,t,mfwhm         : double;
    ps,qs               : array[0..1] of double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_CDenoise: called with: fi.size='+int2str(high(afi)+1)+' wi.size='+int2str(high(awi)+1)+' no='+int2str(no)+' qr='+real2str(qr)+' rcfwhm1='+real2str(rcfwhm1)+' rcfwhm0='+real2str(rcfwhm0)+' rfwhm='+real2str(rfwhm)+' afwhm='+real2str(afwhm));
    if no<1 then
    begin
        cc.warning(cc.name+': No outputs defined. Nothing to do');
        SImg_DecDebugLevel;
        exit;
    end;
    if high(afi)<>high(awi) then cc.error('SImg_CDenoise: Number of images must be equal to number of weights');
    size:=high(afi)+1;
    for i:=0 to size-1 do
    begin
        if not SImg_Recover(cc,afi[i]) then cc.error('SImg_CDenoise: Invalid image: afi['+int2str(i)+']');
        if not SImg_Recover(cc,awi[i]) then cc.error('SImg_CDenoise: Invalid image: awi['+int2str(i)+']');
        if (afi[i].nx<>afi[0].nx) or (afi[i].ny<>afi[0].ny) or (afi[i].nc<>1) then cc.error('SImg_CDenoise: Invalid dimensions: afi['+int2str(i)+']');
        if (awi[i].nx<>afi[0].nx) or (awi[i].ny<>afi[0].ny) or (awi[i].nc<>1) then cc.error('SImg_CDenoise: Invalid dimensions: awi['+int2str(i)+']');
    end;
    if no>size then no:=size;
    if rcfwhm0<1.001 then rcfwhm0:=1.001;
    if rcfwhm1<1.001 then rcfwhm1:=1.001;
    if rcfwhm0<rcfwhm1 then
    begin
        cc.warning('Warning: SImg_CDenoise: FWHM factor for lower values smaller than FWHM factor to higher values. Swapping values');
        r:=rcfwhm0;
        rcfwhm0:=rcfwhm1;
        rcfwhm1:=r;
    end;

    { obtain FWHM }
    mfwhm:=0;
    r:=simg_large;
    for i:=0 to size-1 do
    begin
        if ifwhm[i]<0.01 then
        begin
            ifwhm[i]:=SImg_GetSharpnessFWHM(cc,afi[i]);
            cc.info(cc.name+': calculated FWHM for image '+int2str(i+1)+': '+real2str(ifwhm[i],-1));
        end;
        if ifwhm[i]>mfwhm then mfwhm:=ifwhm[i];
        if ifwhm[i]<r then r:=ifwhm[i];
    end;
    if (rfwhm<1) and (afwhm<r) then cc.error(cc.name+': Channel FWHM limits to small. Either relative limit rfwhm must be larger than 1 or absolute limit must be larger than '+real2str(r+0.01,-2));

    { calculate color fwhm }
    wi3.pt:=nil;
    if rcfwhm0-rcfwhm1<1e-2 then cc.info(cc.name+': Absolute color FWHM: '+real2str(rcfwhm1*mfwhm,-2))
    else begin
        cc.info(cc.name+': Calculating absolute color FWHM within range ' + real2str(mfwhm,-2) + '..' + real2str(rcfwhm0*mfwhm,-2));
        { weighted sum of all colors }
        SImg_New(cc, wi3,  afi[0].nx,afi[0].ny,1, true);
        for k:=0 to wi3.size-1 do
        begin
            s:=0;
            t:=simg_mina;
            for i:=0 to size-1 do
            begin
                r:=awi[i].pt[k];
                if r>0 then
                begin
                    s+=afi[i].pt[k]*r;
                    t+=r;
                end;
            end;
            wi3.pt[k]:=s/t;
        end;

        { LP filtering }
        SImg_Filter(cc, wi3,wi3, mfwhm*FWHM_TO_RAD);

        { color FWHM and weight }
        ps[0]:=0.25*qr;
        ps[1]:=qr;
        SImg_GetQuantile(cc, wi3, cc.threads, -1,ps,qs);
        ps[1]:=(qs[1]-qs[0])*(rcfwhm1-1)/(rcfwhm0-rcfwhm1);
        ps[0]:=rcfwhm0*ps[1];
        s:=1/rcfwhm0;
        for k:=0 to wi3.size-1 do
        begin
            r:=wi3.pt[k]-qs[0];
            if r<0 then r:=rcfwhm0
            else r:=(r+ps[0])/(r+ps[1]);
//define[cdenoise_sqrweight]
//ifdef[cdenoise_sqrweight]
            wi3.pt[k]:=sqr(mfwhm*r);
//else
            wi3.pt[k]:=mfwhm*r;
//endif
        end;
//        SImg_WriteImageFile(cc,'cdenoise-cfwhm.fits',wi3);
    end;

    { LP filtering }
    cc.info(cc.name+': Color (channel ratio) low-pass filtering');
    setlength(afwi2,size);
    setlength(awi2 ,size);
    rr.pt:=nil;
    for i:=0 to size-1 do
    begin
        fi:=afi[i];  wi:=awi[i];
        fwi2.pt:=nil; SImg_New(cc, fwi2, fi.nx,fi.ny,1, false);
        wi2.pt:=nil;  SImg_New(cc, wi2,  fi.nx,fi.ny,1, false);
        if wi3.pt=nil then
        begin
            for k:=0 to fi.size-1 do
            begin
                r:=wi.pt[k];
                if r<0 then r:=0;
                s:=fi.pt[k];
                if s<0 then s:=0;
                wi2.pt[k]:=r;
                fwi2.pt[k]:=s*r;
            end;
            r:=sqr(rcfwhm1*mfwhm)-ifwhm[i];
            if r>0.05 then
            begin
                r:=sqrt(r)*FWHM_TO_RAD;
                SImg_Filter(cc, fwi2,fwi2, r);
                SImg_Filter(cc, wi2 , wi2, r);
            end;
        end
        else begin
            SImg_New(cc, rr, SImg_RawByte, fi.nx,fi.ny,fi.nc, false);
            ps[0]:=ifwhm[i];
            ps[1]:=255/sqrt(sqr(rcfwhm0*mfwhm)-ps[0]);
            for k:=0 to fi.size-1 do
            begin
                r:=wi.pt[k];
                if r<0 then r:=0;
                s:=fi.pt[k];
                if s<0 then s:=0;
                t:=wi3.pt[k];
                r*=t;
                wi2.pt[k]:=r;
                fwi2.pt[k]:=s*r;
//ifdef[cdenoise_sqrweight]
                rr.ptb[k]:=255-round(sqrt(t-ps[0])*ps[1]);
//else
                rr.ptb[k]:=255-round(sqrt(sqr(t)-ps[0])*ps[1]);
//endif
            end;
            r:=sqr(rcfwhm0*mfwhm)-ifwhm[i];
            if r>0.05 then
            begin
                r:=sqrt(r)*FWHM_TO_RAD;
                SImg_Filter(cc, fwi2,fwi2, r, rr);
                SImg_Filter(cc, wi2 , wi2, r, rr);
            end;
//          SImg_WriteImageFile(cc,'cdenoise-rr-'+int2str(i)+'.fits',rr);
        end;
        afwi2[i]:=fwi2;
        awi2[i]:=wi2;
    end;
    SImg_Free(cc, wi3);
    SImg_Free(cc, rr);

    { combine input images }
    setlength(aoi, size);
    wi3.pt:=nil;
    SImg_New(cc, wi3,  afi[0].nx,afi[0].ny,1, false);
    for i:=0 to no-1 do
    begin
        fwi2i:=afwi2[i];  wi2i:=awi2[i];
        SImg_Clear(cc, wi3);
        fwi3.pt:=nil;
        Simg_New(cc, fwi3, wi3.nx,wi3.ny,1, true);
        mfwhm:=ifwhm[i]*rfwhm;
        if afwhm>mfwhm then mfwhm:=afwhm;
        cc.info(cc.name+': Combining color '+int2str(i+1)+'. Maximum FWHM: '+real2str(ifwhm[i],-2));
        for j:=0 to size-1 do
        begin
            if (i=j) or (ifwhm[j]>mfwhm) then continue;
            fi  :=afi[j];    wi :=awi[j];
            fwi2:=afwi2[j];  wi2:=awi2[j];
            for k:=0 to fi.size-1 do
            begin
                s:=fwi2i.pt[k]*wi2.pt[k];
                if s<simg_mina then continue;
                s:=fwi2.pt[k]*wi2i.pt[k]/s;     { s=1 for i=j, that's why it is handled separately }
                r:=wi.pt[k];
                fwi3.pt[k]+=fi.pt[k]*r*s;
                wi3.pt[k]+=r*sqr(s);
            end;
        end;
        t:=reg_nc;
        if ifwhm[i]<=mfwhm then t+=1;
        fi:=afi[i];    wi:=awi[i];
        for k:=0 to fi.size-1 do
        begin
            r:=wi3.pt[k];
            if r>simg_mina then
            begin
                s:=wi.pt[k]*t;
                fwi3.pt[k]:=(fi.pt[k]*s+fwi3.pt[k])/(s+r);
            end
            else fwi3.pt[k]:=fi.pt[k]
        end;
        aoi[i]:=fwi3;
    end;
    SImg_Free(cc, wi3);

    { output, release memory }
    for i:=0 to size-1 do
    begin
        if i<no then SImg_Move(cc, aoi[i], afi[i]);
        SImg_Free(cc, afwi2[i]);
        SImg_Free(cc, awi2[i]);
    end;
    setlength(afwi2,0);
    setlength(awi2 ,0);
    setlength(aoi  ,0);

    SImg_DecDebugLevel;
end;



{*************************************************************************}
{******* SImg_BGDenoise **************************************************}
{*************************************************************************}
{ Denoisung of dark regions
    fi     : input and output image
    fwhm   : input FWHM
    rad    : maximum low pass filter radius
    qh     : no denoising above that quantile
    ql     : maximum denoising below that quantile
}
procedure SImg_BGDenoise(cc:CSImg_Config; fi:TSImg_FPImage; fwhm,rad,qh,ql:double);
var fw,fiw      : TSImg_FPImage;
    rr          : TSImg_RDImage;
    ps,qs       : array[0..1] of double;
    i           : longint;
    r           : double;
begin
    SImg_IncDebugLevel;
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_BGDenoise: fi='+SImg_InfoStr2(fi)+' fwhm='+real2str(fwhm)+' rad='+real2str(rad)+' sp='+real2str(qh)+' thr='+real2str(ql));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BGDenoise: Invalid image: fi');
    if qh-ql<1e-3 then cc.error(cc.name+': Lower quantile larger than upper quantile');
    if fwhm<1.5 then fwhm:=1.5;

    { quantiles }
    fiw.pt:=nil;
    SImg_Filter(cc, fi,fiw, fwhm*FWHM_TO_SIG);
    ps[0]:=ql;
    ps[1]:=qh;
    SImg_GetQuantile(cc, fiw, cc.threads, -1,ps,qs);
    qs[1]-=qs[0];
    if qs[1]<simg_mina then cc.error(cc.name+': Quantiles to close or constant image');
    qs[1]:=1/qs[1];

    { weight and filter radius }
    fw.pt:=nil;
    rr.pt:=nil;
    SImg_New(cc, fw, fi.nx,fi.ny,fi.nc, false);
    SImg_New(cc, rr, SImg_RawByte, fi.nx,fi.ny,fi.nc, false);
    for i:=0 to fi.size-1 do
    begin
        r:=1-(fiw.pt[i]-qs[0])*qs[1];
        if r<0 then r:=0;
        if r>1 then r:=1;
        r:=sqr(r);
        rr.ptb[i]:=round((1-r)*255);
        fw.pt[i]:=r;
        fiw.pt[i]:=fi.pt[i]*r;
    end;
//    SImg_WriteImageFile(cc,'bgdenoise-rr.fits',rr);

    { filtering }
    rad/=SIG_TO_RAD;
    SImg_Filter(cc, fiw,fiw, rad,rr);
    SImg_Filter(cc, fw,fw, rad,rr);

    { output }
    rad/=SIG_TO_RAD;
    for i:=0 to fi.size-1 do
        if fw.pt[i]>simg_mina then fi.pt[i]:=fiw.pt[i]/fw.pt[i];
//      if rr.ptb[i]<255 then fi.pt[i]:=fiw.pt[i]/fw.pt[i];

    SImg_Free(cc, fw);
    SImg_Free(cc, fiw);
    SImg_Free(cc, rr);
    SImg_DecDebugLevel;
end;
