{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[debug_hdr]

{*************************************************************************}
{******* SImg_DRCompress *************************************************}
{*************************************************************************}
{ luminance conversion, black level adjustment.
  lq   : quantile for black level adjustment to <0 to skip that
  crop : crop negative values of one color images
  result: black level }
function SImg_DRCompress_convert(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; lq:double; crop:boolean):double;
var fj  : TSImg_FPImage;
    r   : double;
    i   : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    { one color conversion }
    fj.pt:=nil;
    if fi.nc=1 then fj:=fi
    else SImg_AddColors(cc, fi,fj);
    { black level adjustment }
    result:=0;
    if (fi.nc<>1) or crop then
    begin
        if lq<=-simg_mina then
        begin
            for i:=0 to fj.size-1 do
                if fj.pt[i]<0 then fj.pt[i]:=0;
        end else
        begin
            lq:=SImg_GetQuantile(cc,fj,lq);
            for i:=0 to fj.size-1 do
            begin
                r:=fj.pt[i]-lq;
                if r<0 then r:=0;
                fj.pt[i]:=r;
            end;
            result:=lq;
        end;
    end
    else if lq>-simg_mina then
    begin
        lq:=SImg_GetQuantile(cc,fj,lq);
        for i:=0 to fj.size-1 do
            fj.pt[i]-=lq;
        result:=lq;
    end;
    fo:=fj;
end;


{ convert 1-color image to output
  fj: input (1-color image)
  fi: output
  lq: black level returned by SImg_DRCompress_convert
  sp: saturation preservation
}
procedure SImg_DRCompress_output(fj,fi:TSImg_FPImage; lq,sp:double);
var i,c         : longint;
    r,s,t,u,v   : double;
    d           : ^simgfp;
    b           : boolean;
begin
    if sp<0 then sp:=0;
    if sp>1 then sp:=1;

    if fj.size=fi.size then { output is one color image }
    begin
        if fi.pt<>fj.pt then move(fj.pt^, fi.pt^, fj.size*sizeof(simgfp));
        exit;
    end;

    lq/=fi.nc;
    d:=fi.pt-fi.nc;
    v:=1/fi.nc;
    sp:=1-sp;
    b:=sp>1e-3;
    for i:=0 to fj.size-1 do
    begin
        d+=fi.nc;
        r:=0; t:=1e100;
        for c:=0 to fi.nc-1 do
        begin
            s:=d[c]-lq;
            if s<0 then s:=0;
            d[c]:=s;
            r+=s;
            if s<t then t:=s;
        end;
        if r<simg_mina then continue;
        { subtraction }
        s:=fj.pt[i];
        if s<0 then s:=0;
        if b and (t>simg_mina) then
        begin
            u:=(r-s)*v;
            if t<u then u:=t;
            u*=sp;
            for c:=0 to fi.nc-1 do
                d[c]-=u;
            r-=u*fi.nc;
            if r<1e-100 then continue;
        end;
        { multiplication }
        r:=s/r;
        for c:=0 to fi.nc-1 do
            d[c]*=r;
    end;
end;


{*************************************************************************}
{******* SImg_DRCompress1 ************************************************}
{*************************************************************************}
{ Dynamic range compression using non-linear high-pass filter and regularization method 1
    fi    : input and output image
    mask  : mask, optional, 1-color image
    lq    : black level quantile or <0 for no black level adjustment
    mq    : 25% reduction quantile
    ss    : sizes of details to emphasize
    sp    : saturation preservation (0..1)
}
procedure SImg_DRCompress1(cc:CSImg_Config; fi:TSImg_FPImage; mask:TSImg_RDImage; lq,mq:double; ss:array of double; sp:double);
var fj,fk       : TSImg_FPImage;
    i,k         : longint;
    r,s,t       : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_DRCompress1: called with: fi='+SImg_InfoStr2(fi)+' mask='+SImg_InfoStr2(mask)+' lq='+real2str(lq)+' mq='+real2str(mq)+' ss.size='+int2str(high(ss)+1)+' sp='+real2str(sp));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_DRCompress2: Invalid image: fi');
    if mask.pt<>nil then
    begin
        if not SImg_Recover(cc,mask) then cc.error('SImg_DRCompress1: Invalid image: mask');
        if (mask.nx<>fi.nx) or (mask.ny<>fi.ny) or (mask.nc<>1) then cc.error('SImg_DRCompress1: Incompatible dimensions: mask');
    end;

    { convert to 1 color image }
    lq:=SImg_DRCompress_convert(cc,fi,fj,lq,true);

    { bg-filter }
    fk.pt:=nil;
    for k:=0 to high(ss) do
    begin
        SImg_FBGFilter(cc, fj,fk, mask, ss[k],ss[k],0.5, true, false);

        r:=SImg_GetQuantile(cc,fk,mq);
        if r<1e-100 then r:=1e-100;
        s:=0.25/r;
        for i:=0 to fk.size-1 do
        begin
            t:=fk.pt[i];
            if t>=r then fj.pt[i]-=0.5*t-0.25*r
            else fj.pt[i]-=sqr(t)*s;
        end;

//ifdef[debug_hdr]
        SImg_WriteImageFile(cc,'drcompress1-fk-'+int2str(round(ss[k]))+'.fits',fk);
         SImg_WriteImageFile(cc,'drcompress1-fj-'+int2str(round(ss[k]))+'.fits',fj);
//endif
    end;
    SImg_Free(cc,fk);

    { output to 1 color image }
    SImg_DRCompress_output(fj,fi,lq,sp);
    if fi.pt<>fj.pt then SImg_Free(cc,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_DRCompress2 ************************************************}
{*************************************************************************}
{ ti : thread index
  tn : thread number
  r  : reduction factor, 1/exponent
  s  : maximum of fk
}
threadprocedure SImg_DRCompress2_thread(ti:longint; tn:longint; fj:TSImg_FPImage; fk:TSImg_FPImage; r:double; s:double);
var t,rr,rs     : double;
    x,y,ya,yz   : longint;
    d,e         : ^simgfp;
begin
    ya:=(fj.ny*ti) div tn;
    yz:=(fj.ny*(ti+1)) div tn-1;

    d:=fj.pt + ya*fj.sl;
    e:=fk.pt + ya*fk.sl;
    rr:=1/r;
    rs:=1/s;
    for y:=ya to yz do
        for x:=0 to fj.sl-1 do
        begin
            t:=e^*rs; e+=1;
            if t>0.99999999 then t-=r
            else t-=r*(1-exp(ln(1-t)*rr));
            d^-=s*t; d+=1;
        end;
end;


{ Dynamic range compression using non-linear high-pass filter and regularization method 2
    fi    : input and output image
    mask  : mask, optional, 1-color image
    lq    : black level quantile or <0 for no black level adjustment
    hq    : high quantile at which background will be reduced by 50%
    rmin  : Reduction factor limit. 0..1
    ss    : smallest details size to emphasize
    sp    : saturation preservation (0..1)
}
procedure SImg_DRCompress2(cc:CSImg_Config; fi:TSImg_FPImage; mask:TSImg_RDImage; lq,hq, rmin, ss, sp:double);
var fj,fk       : TSImg_FPImage;
    i,tn        : longint;
    r,s,t       : double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_DRCompress2: called with: fi='+SImg_InfoStr2(fi)+' mask='+SImg_InfoStr2(mask)+' lq='+real2str(lq)+' hq='+real2str(hq)+' rmin='+real2str(rmin)+' ss='+real2str(ss,-1)+' sp='+real2str(sp));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_DRCompress2: Invalid image: fi');
    if mask.pt<>nil then
    begin
        if not SImg_Recover(cc,mask) then cc.error('SImg_DRCompress2: Invalid image: mask');
        if (mask.nx<>fi.nx) or (mask.ny<>fi.ny) or (mask.nc<>1) then cc.error('SImg_DRCompress2: Incompatible dimensions: mask');
    end;
    if rmin>0.9999 then
    begin
        cc.warning('drcompress2: background reduction >=1, nothing to do');
        SImg_DecDebugLevel;
        exit;
    end;

    { convert to 1 color image }
    lq:=SImg_DRCompress_convert(cc,fi,fj,lq,true);
    hq:=SImg_GetQuantile(cc,fj,hq);

    { bg-filter }
    tn:=cc.threads;
    fk.pt:=nil;
    repeat
        SImg_FBGFilter(cc, fj,fk, mask, ss,ss,0.5, true, false);

        { calculate reduction factor (=1/exponent) }
        s:=0;
        for i:=0 to fk.size-1 do
            if fk.pt[i]>s then s:=fk.pt[i];
        if s<1e-100 then break;
        t:=hq/s;
        r:=0.5*t;
        if t<0.99999 then
        begin
            { solve using FP iteration, only very few iterations required }
            for i:=0 to 5 do
                r:=0.5*t/(1-exp(ln(1-t)/r));
//ifdef[debug_hdr]
            cc.info('drcompress2: '+ real2str(t,6) + ' --> ' + real2str(r*(1-exp(ln(1-t)/r)),6) );
//endif
        end;
        if r>1 then break;
        if r<rmin then r:=rmin;
        cc.info('drcompress2: maximum reduction at size=' + real2str(ss,-1) + ': ' + real2str(r,6));

        { subtraction }
        for i:=0 to tn-1 do
            dothread SImg_DRCompress2_thread(i,tn, fj,fk, r,s);
        threadwait;

//ifdef[debug_hdr]
        SImg_WriteImageFile(cc,'drcompress2-fk-'+int2str(round(ss))+'.fits',fk);
        SImg_WriteImageFile(cc,'drcompress2-fj-'+int2str(round(ss))+'.fits',fj);
//endif
        ss*=2;
    until (ss>=fj.nx) or (ss>=fj.ny);
    SImg_Free(cc,fk);

    { output to 1 color image }
    SImg_DRCompress_output(fj,fi,lq,sp);
    if fi.pt<>fj.pt then SImg_Free(cc,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_DRCompress3 ************************************************}
{*************************************************************************}
{ ti  : thread index
  tn  : thread number
  reg : regularization
}
threadprocedure SImg_DRCompress3_thread(ti:longint; tn:longint; fc:TSImg_FPImage; fj:TSImg_FPImage; fk:TSImg_FPImage; r:double; reg:double; a:psimgfp; b:psimgfp);
var s,t         : double;
    i,ia,iz     : longint;
begin
    ia:=((fj.ny*ti) div tn)*fj.sl;
    iz:=((fj.ny*(ti+1)) div tn)*fj.sl-1;
    r:=r-1;
    for i:=ia to iz do
    begin
        if fk.pt<>nil then
        begin
            t:=fk.pt[i];
            if t>a^ then a^:=t  { not 100% thread save but consequences are acceptable }
            else if t<b^ then b^:=t;
            t:=fj.pt[i]-t;
        end
        else t:=fj.pt[i];
        s:=fc.pt[i];
        if s<0 then
        begin
            t+=s;
            s:=0;
        end;
        if t<=0 then fc.pt[i]:=s+t
        else fc.pt[i]:=s+t+r*sqr(t)/(reg+t);
    end;
end;

{ Dynamic range compression using non-linear high-pass filter and regularization method 3
    fi    : input and output image
    mask  : mask, optional, 1-color image
    lq    : black level quantile or <0 for no black level adjustment. Recommended: 0..0.25
    smin  : smallest detail filter (as radius). Should be larger than superresolution factor. At least 1.
    smed  : used median filter below that size
    smax  : largest detail filter (as radius). At least smin*2. If less, it is set to 0.5*min(fi.nx,fi.ny).
    ravg  : average dynamic range compression
    reg   : regularization relative to noise
    sig   : noise floor. Auto-detect if input is <=0
    sp    : saturation preservation (0..1)
}
procedure SImg_DRCompress3(cc:CSImg_Config; var fi:TSImg_FPImage; mask:TSImg_RDImage; lq, smin,smed,smax,ravg, reg:double; var sig:double; sp:double);
const stop_limit = 0.5;  { abort when remaining component is less than this limit relative to noise floor }
      reg_fact_fin = 1;  { regularization factor for final iteration }
var fj,fk,fc            : TSImg_FPImage;
    i,j,k               : longint;
    r,s,s_prev,rr,sr,t  : double;
    a,b,c               : simgfp;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_DRCompress3: called with: fi='+SImg_InfoStr2(fi)+' mask='+SImg_InfoStr2(mask)+' lq='+real2str(lq)+' smin='+real2str(smin)+' smed='+real2str(smed)+' smax='+real2str(smax)+' ravg='+real2str(ravg,-1)+' reg='+real2str(reg,-2)+' sig='+real2str(sig)+' sp='+real2str(sp));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_DRCompress3: Invalid image: fi');
    if mask.pt<>nil then
    begin
        if not SImg_Recover(cc,mask) then cc.error('SImg_DRCompress3: Invalid image: mask');
        if (mask.nx<>fi.nx) or (mask.ny<>+fi.ny) or (mask.nc<>1) then cc.error('SImg_DRCompress3: Incompatible dimensions: mask');
    end;

    if ravg<1e-3 then ravg:=1e-3;
    if smin<1 then smin:=1;
    if fi.ny<fi.nx then a:=fi.ny
    else a:=fi.nx;
    if smax<2*smin then smax:=0.5*a;
    if reg<0 then reg:=0;
    k:=round(ln(smax/smin)/(0.5*ln(5)))+1;      { number of filters }
    if k<2 then k:=2;
    sr:=exp(ln(smax/smin)/(k-1));       { size step sr }
    { factor step rr: ravg*(k+1) = (1-rr^(k+1))/(1-rr) --> fixpoint iteration: rr_(l+1) = 1-(1-rr_l^(k+1))/(ravg*(k+1)) }
    rr:=0.5;
    for i:=0 to 10 do
        rr:=1 - (1-exp(ln(rr)*(k+1)))/(ravg*(k+1));
//  cc.info('k='+int2str(k)+' sr='+real2str(sr)+' rr='+real2str(rr)+'  smin='+real2str(smin)+' smax='+real2str(smax));

    if (sig<simg_mina) and (lq<=-simg_mina) or (lq>=0.2501) then cc.warning(cc.name+': Warning: Noise floor calculation may be incorrect if too much values are cropped');

    {  convert to 1 color image }
    fj.pt:=nil;
    lq:=SImg_DRCompress_convert(cc,fi,fj,lq,true);

    if sig<simg_mina then
    begin
        sig:=SImg_NoiseFloor(cc, fj, min(smin*3,smed))/fi.nc;
        cc.info(cc.name+': Estimated noise floor: '+real2str(sig,4));
    end;
    reg*=sig*fi.nc;

    { min-filter }
    fk.pt:=nil;
    s:=smin; s_prev:=0;
    r:=1;
    j:=0;
    fc.pt:=nil; SImg_New(cc, fc, fj.nx,fj.ny,1, false);
    repeat
        { fj: detail component s_prev..s }
        a:=sqrt(sqr(s)-sqr(s_prev));
        if a<=smed then SImg_Filter(cc, fj,fk, a, -4, 1,0.5)
        else SImg_FBGFilter(cc, fj,fk, mask, a,a,0.5, true, j=0);

        { add component }
        a:=fk.pt[0]; b:=a;
        cc.info(cc.name+': Max. reduction for component '+real2str(s_prev,-1)+'...'+real2str(s,-1)+': '+real2str(1-r,4));
        if j=0 then
            for i:=0 to fj.size-1 do
            begin
                t:=fk.pt[i];
                if t>a then a:=t
                else if t<b then b:=t;
                fc.pt[i]:=fj.pt[i]-t;   { r=1 }
            end
        else if reg<simg_mina then
            for i:=0 to fj.size-1 do
            begin
                t:=fk.pt[i];
                if t>a then a:=t
                else if t<b then b:=t;
                t:=fj.pt[i]-t;
                c:=fc.pt[i];
                if c<0 then
                begin
                    t+=c;
                    c:=0;
                end;
                if t<0 then fc.pt[i]:=c+t
                else fc.pt[i]:=c+t*r;
            end
        else begin
            for i:=0 to cc.threads-1 do
                dothread SImg_DRCompress3_thread(i,cc.threads, fc,fj,fk, r,reg, @a,@b);
            threadwait;
        end;
        SImg_Move(cc, fk,fj);

        { next step }
        r*=rr;
        s_prev:=s;
        s*=sr;
        j+=1;
    until (j>=k) or ((a-b)<reg*stop_limit);
    SImg_Free(cc,fk);

    { final step: residual }
    cc.info(cc.name+': Max. reduction for component >'+real2str(s_prev,-1)+': '+real2str(1-r,4));
    if reg<simg_mina then
        for i:=0 to fj.size-1 do
        begin
            t:=fj.pt[i];
            c:=fc.pt[i];
            if c<0 then
            begin
                t+=c;
                c:=0;
            end;
            if t<0 then fc.pt[i]:=0
            else fc.pt[i]:=c+t*r;
        end
    else  begin
        for i:=0 to cc.threads-1 do
            dothread SImg_DRCompress3_thread(i,cc.threads, fc,fj,fk, r,reg*reg_fact_fin, @a,@b);
        threadwait;
        for i:=0 to fc.size-1 do
            if fc.pt[i]<0 then fc.pt[i]:=0;
    end;
    SImg_Free(cc, fj);

    { output to 1 color image }
    if fi.nc=1 then SImg_Move(cc, fc,fi)
    else begin
        SImg_DRCompress_output(fc,fi,lq,sp);
        SImg_Free(cc,fc);
    end;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_DRCompose **************************************************}
{*************************************************************************}
threadprocedure SImg_DRCompose_thread(ti:longint; tn:longint; fc:TSImg_FPImage; fj:TSImg_FPImage; fk:TSImg_FPImage; j:longint; ra2:pdouble; reg:double);
var i,ia,iz,k,c : longint;
    a,r         : double;
begin
    ia:=((fj.ny*ti) div tn)*fj.sl;
    iz:=((fj.ny*(ti+1)) div tn)*fj.sl-1;
    if reg>simg_mina then
        for i:=ia to iz do
        begin
            if fk.pt<>nil then a:=fj.pt[i]-fk.pt[i]
            else a:=fj.pt[i];
            if a<0 then a:=0;  { may happen with mask }
            k:=i*fc.nc;
            r:=sqr(a)/(a+reg);
            for c:=0 to fc.nc-j do
                fc.pt[k+c]+=a+(ra2[-c]-1)*r;
            for c:=fc.nc-j+1 to fc.nc-1 do
                fc.pt[k+c]+=a-r;
        end
    else
        for i:=0 to fj.size-1 do
        begin
            if fk.pt<>nil then a:=fj.pt[i]-fk.pt[i]
            else a:=fj.pt[i];
            if a<0 then a:=0;  { may happen with mask }
            k:=i*fc.nc;
            for c:=0 to fc.nc-j do
                fc.pt[k+c]+=a*ra2[-c];
        end;
end;


{ Colorize image using different dynamic range compression for different colors
    fi    : input image
    fo    : output image
    mask  : mask, optional, 1-color image
    lq    : black level quantile (ignored if <0)
    reg   : reduction factor normalization relative to noise
    sa    : Detail filters (as radius). One per color. Last color last colors comes first. Details smaller than first element are added to all colors with factor 1. First element should be at least two times the superresolution factor.
    ra    : Compression factor per color. Last color colors comes first.
    result: noise floor
}
function SImg_DRCompose(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; mask:TSImg_RDImage; lq,reg:double; sa,ra:array of double):double;
var fj,fk,fc    : TSImg_FPImage;
    i,j,k,c     : longint;
    r           : double;
    a           : simgfp;
    ra2         : ^double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_DRCompose: called with: fi='+SImg_InfoStr2(fi)+' fo='+SImg_InfoStr2(fo)+' mask='+SImg_InfoStr2(mask)+' lq='+real2str(lq)+' reg='+real2str(reg)+' lq='+real2str(lq)+' sa.length='+int2str(1+high(sa))+' ra.length='+int2str(1+high(ra)));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_DRCompose: Invalid image: fi');
    if fi.nc<>1 then cc.Error('SImg_DRCompose: 1-color image expected: fi');
    if mask.pt<>nil then
    begin
        if not SImg_Recover(cc,mask) then cc.error('SImg_DRCompose: Invalid image: mask');
        if (mask.nx<>fi.nx) or (mask.ny<>fi.ny) or (mask.nc<>1) then cc.error('SImg_DRCompose: Incompatible dimensions: mask');
    end;
    if high(sa)<>high(ra) then cc.error('SImg_DRCompose: Arrays ''ra'' and ''sa'' have different lengths');
    if not SImg_OverwriteAlert(fi,fo) then SImg_Free(cc,fo);

    fc.nc:=high(ra)+1;
    ra2:=@ra[high(ra)];         { points to red factor }

    { black level }
    if lq>0 then r:=SImg_GetQuantile(cc,fi,lq)
    else r:=0;
    fj.pt:=nil;
    SImg_New(cc, fj, fi.nx,fi.ny,1, false);
    for i:=0 to fi.size-1 do
    begin
        a:=fi.pt[i]-r;
        if a<0 then a:=0;
        fj.pt[i]:=a;
    end;
    if SImg_OverwriteAlert(fi,fo) then SImg_Free(cc,fi);

    { bg-filter }
    fk.pt:=nil; fc.pt:=nil;
    SImg_New(cc, fc, fj.nx,fj.ny,fc.nc, false);
    result:=0;  { makes the compiler happy }
    for j:=0 to fc.nc-1 do
    begin
        r:=sa[j];
        cc.info(cc.name+': Filter size '+real2str(r,-1));
        SImg_FBGFilter(cc, fj,fk, mask, r,r,0.5, true, j=0);
        if j=0 then
        begin
            for i:=0 to fj.size-1 do { white: added to all colors with factor 1 }
            begin
                a:=fj.pt[i]-fk.pt[i];
                if a<0 then a:=0;  { may happen with mask }
                fj.pt[i]:=a;
                k:=i*fc.nc;
                for c:=0 to fc.nc-1 do
                    fc.pt[k+c]:=a;
            end;
            result:=SImg_NoiseFloor_int(cc,fj);
            if reg>=1e-6 then reg*=result
            else reg:=0;
        end
        else begin
            for i:=0 to cc.threads-1 do
                dothread SImg_DRCompose_thread(i,cc.threads, fc,fj,fk, j,ra2,reg);
            threadwait;
        end;
        SImg_Move(cc, fk,fj);
    end;
    r:=ra2[0];
    for i:=0 to cc.threads-1 do
        dothread SImg_DRCompose_thread(i,cc.threads, fc,fj,fk, fc.nc,ra2,reg);
    threadwait;
    SImg_Free(cc,fj);
    SImg_Move(cc, fc,fo);

    SImg_DecDebugLevel;
end;
