{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

//define[debug_histogr]
{*************************************************************************}
{*************************************************************************}
{******* histogram transformations ***************************************}
{*************************************************************************}
{*************************************************************************}

{*************************************************************************}
{******* SImg_GetMinMaxN *************************************************}
{*************************************************************************}
{ get the n-th brightest/darkest value
    n         : n
    c         : color, <0 for all
    minv,maxv : results
}
procedure SImg_GetMinMaxN(cc:CSImg_Config; fi:TSImg_FPImage; n,c:longint; var minv,maxv:double);
var i,j,iz,cd   : longint;
    r           : simgfp;
    d,bd,bb     : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetMinMaxN: Invalid image');
    if n<1 then cc.error('SImg_MinMaxN: n must be greater than 0');

    getmem2(bd,sizeof(simgfp)*n*2);
    bb:=bd+n;

    for i:=0 to n-1 do
    begin
        bd[i]:=simg_large;
        bb[i]:=-simg_large;
    end;

    cd:=1;
    if (c<0) or (c>=fi.nc) then c:=0
    else cd:=fi.nc;
    iz:=(fi.size div cd)-1;

    d:=fi.pt+c;
    for i:=0 to iz do
    begin
        r:=d^; d+=cd;
        { bb }
        if r>bb[0] then
        begin
            j:=1;
            while (j<n) and (r>bb[j]) do
            begin
                bb[j-1]:=bb[j];
                j+=1;
            end;
            bb[j-1]:=r;
        end;
        { bd }
        if r<bd[0] then
        begin
            j:=1;
            while (j<n) and (r<bd[j]) do
            begin
                bd[j-1]:=bd[j];
                j+=1;
            end;
            bd[j-1]:=r;
        end;
    end;
    minv:=bd[0];
    maxv:=bb[0];
    freemem2(bd,sizeof(simgfp)*n*2);
end;


{*************************************************************************}
{******* SImg_GetHistogramm **********************************************}
{*************************************************************************}
{ get the n-th brightest/darkest value
    tn         : number of threads
    c          : color, <0 for all
    minv,maxv  : minimum maximum value to consider
    hist       : histogram buffer
    histsize   : length of histogram buffer
}
threadprocedure SImg_GetHistogram_thread(fi:TSImg_FPImage; ya:longint; yz:longint; c:longint; minv:double; maxv:double; hist:plongint; histsize:longint);
var i,iz,cd,h0 : longint;
    s,hs2      : double;
    d          : ^simgfp;
    r          : double;
begin
    cd:=1;
    if (c<0) or (c>=fi.nc)
        then c:=0
        else cd:=fi.nc;
    iz:=fi.nx*(((yz-ya+1)*fi.nc) div cd)-1;

    fillchar(hist^,histsize*sizeof(longint),0);
    if abs(maxv-minv)<simg_mina then exit;

    hs2:=histsize-0.01;
    s:=hs2/(maxv-minv);
    d:=fi.pt + ya*fi.sl + c;
    h0:=0;
    for i:=0 to iz do
    begin
        r:=d^; d+=cd;
        if r<=minv then h0+=1
        else if r<=maxv then hist[trunc((r-minv)*s)]+=1;
    end;
    hist[0]+=h0;
end;

procedure SImg_GetHistogram(cc:CSImg_Config; fi:TSImg_FPImage; tn,c:longint; minv,maxv:double; hist:plongint; histsize:longint);
var i,t,ya,yz  : longint;
    histbuf,d  : ^longint;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetHistogram: illegal image');
    if tn<2 then
    begin
        SImg_GetHistogram_thread(fi, 0,fi.ny-1, c, minv,maxv, hist, histsize);
        exit;
    end;

    getmem2(histbuf,sizeof(longint)*histsize*(tn-1));
    yz:=fi.ny div tn;
    dothread Simg_GetHistogram_thread(fi, 0,yz-1, c, minv,maxv, hist, histsize);
    for t:=0 to tn-2 do
    begin
        ya:=yz;
        yz:=(fi.ny*(t+2)) div tn;
        dothread Simg_GetHistogram_thread(fi, ya,yz-1, c, minv,maxv, histbuf+t*histsize, histsize);
    end;
    threadwait;
    for t:=0 to tn-2 do
    begin
        d:=histbuf+t*histsize;
        for i:=0 to histsize-1 do
            hist[i]+=d[i];
    end;

    freemem2(histbuf,sizeof(longint)*histsize*(tn-1));
end;


{*************************************************************************}
{******* SImg_GetQuantile ************************************************}
{*************************************************************************}
{ get quantiles
    tn       : number of threads
    c        : color, <0 for all
    ps       : percentages
    qs       : quantiles (result, must be allocated and can be the same array as ps)
}
procedure SImg_GetQuantile(cc:CSImg_Config; fi:TSImg_FPImage; tn,c:longint; ps:array of double; var qs:array of double);
const histsize=16*1024;
var hist1,hist2 : ^longint;
    minv,maxv   : double;
    i,j,jz      : longint;
    a,b         : int64;
    ar,r,s,t    : double;
    nohist      : boolean;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetQuantile: illegal image');

    SImg_GetMinMaxN(cc, fi, 1+((fi.nx*fi.ny) shr 22), c, minv,maxv);

    jz:=high(ps);
    if high(qs)<jz then jz:=high(qs);
    nohist:=true;
    for j:=0 to jz do
        if abs(ps[j]-0.5) < 0.5-1e-8 then nohist:=false;
    if nohist then
    begin
        for j:=0 to jz do
            qs[j]:=select(ps[j]<0.5, minv, maxv);
        exit;
    end;

    if maxv-minv<1e-20 then
    begin
        cc.warning('SImg_GetQuantile: Constant image');
        for j:=0 to jz do
            qs[j]:=0.5*(minv+maxv);
        exit;
    end;

    getmem2(hist1,histsize*sizeof(longint));
    getmem2(hist2,histsize*sizeof(longint));
    SImg_GetHistogram(cc, fi, tn,c, minv,maxv, hist1,histsize);

    for j:=0 to jz do
        if ps[j]<=1e-8 then qs[j]:=minv
        else if ps[j]>=1-1e-8 then qs[j]:=maxv
        else begin
            a:=fi.nx*fi.ny;
            if (c<0) or (c>=fi.nc) then a*=fi.nc;
            ar:=ps[j]*a; a:=trunc(ar);

            b:=0; i:=-1;
            repeat
                i+=1;
                b+=hist1[i];
            until (b>a) or (i=histsize-1);

            r:=(i-0.05)/histsize*(maxv-minv)+minv;
            s:=(i+1.05)/histsize*(maxv-minv)+minv;
            SImg_GetHistogram(cc, fi, tn,c, r,s, hist2,histsize);

            b:=0; i:=-1;
            repeat
                i+=1;
                b+=hist2[i];
            until (b>a) or (i=histsize-1);
            if hist2[i]=0 then t:=i
            else t:=i + (ar-(b-hist2[i]) )/hist2[i];
            qs[j]:=r + t*(s-r)/histsize;
        end;

    freemem2(hist1,histsize*sizeof(longint));
    freemem2(hist2,histsize*sizeof(longint));
end;


function SImg_GetQuantile(cc:CSImg_Config; fi:TSImg_FPImage; q:double):double;
var ps,qs  : array[0..0] of double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    ps[0]:=q;
    SImg_GetQuantile(cc, fi, select(gettid=SImg_MainTid,1,cc.threads), -1,ps,qs);
    result:=qs[0];
end;


function SImg_GetQuantile(cc:CSImg_Config; fi:TSImg_FPImage; c:longint; q:double):double;
var ps,qs  : array[0..0] of double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if (c<0) or (c>fi.nc-1) then cc.error('SImg_GetQuantile: Invalid color');
    ps[0]:=q;
    SImg_GetQuantile(cc, fi, select(gettid=SImg_MainTid,1,cc.threads), c,ps,qs);
    result:=qs[0];
end;


{*************************************************************************}
{******* SImg_HistCut ****************************************************}
{*************************************************************************}
{ crop and scale brightness with linear tonal curve
  lq,uq    : lower/upper quantile
  cut      : cut values
  st       : stretch to [0,1]
  la,ua    : interpret lq/uq as absolute value (not as quantile) }
procedure SImg_HistCut(cc:CSImg_Config; fi:TSImg_FPImage; lq,uq:double; cut,st:boolean; la,ua:boolean);
var i          : longint;
    rl,ru      : double;
    r,t        : double;
    d          : ^simgfp;
    ps,qs      : array[0..1] of double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_HistCut: called with: img='+SImg_InfoStr2(fi)+'lq='+real2str(lq)+' uq='+real2str(uq)+
        ' cut='+boolean2str(cut)+' st='+boolean2str(st)+' la='+boolean2str(la)+' ua='+boolean2str(ua));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_HistCut: Invalid image: fi');
    if (not cut) and (not st) then cc.Warning('SImg_HistCut: Nothing to do');

    if lq>uq then
    begin
        cc.warning(cc.name+': Upper quantile smaller than lower quantile, swapping them. Use `convert -sc -1'' to invert image.');
        r:=lq; lq:=uq; uq:=r;
    end;

    ps[0]:=select(la,-1,lq);
    ps[1]:=select(ua,-1,uq);
    SImg_GetQuantile(cc,fi,cc.threads, -1,ps,qs);
    rl:=select(la,lq,qs[0]);
    ru:=select(ua,uq,qs[1]);

    if ru-rl<1e-20 then
    begin
        cc.warning(cc.name+': Quantiles are to close or constant image');
        t:=0;
    end else t:=1/(ru-rl);

    d:=fi.pt;
    for i:=0 to fi.size-1 do
    begin
        r:=d^;
        if cut then
        begin
            if r<rl then r:=rl
            else if r>ru then r:=ru;
        end;

        if st then d^:=(r-rl)*t
        else d^:=r;
        d+=1;
    end;

    SImg_DecDebugLevel;
end;

{*************************************************************************}
{******* SImg_HistCut ****************************************************}
{*************************************************************************}
threadprocedure SImg_HistCut_thread(ya:longint; yz:longint; fi:TSImg_FPImage; rl:double; ru:double; alp1:double; alp2:double; sp:double; st:boolean);
var i,iz,c      : longint;
    r,s,m,t,u,v : double;
    rl2,ru2     : double;
    d           : ^simgfp;
    rs          : boolean;
begin
    rs:=sp>1e-4;
    if not rs then sp:=0;
    t:=1/(ru-rl);
    u:=1/fi.nc;
    rl2:=rl+(ru-rl)*1e-10;
    ru2:=ru-(ru-rl)*1e-10;
    iz:=(yz-ya+1)*fi.nx-1;
    d:=fi.pt + ya*fi.sl;
    for i:=0 to iz do
    begin
        s:=0; m:=0;
        for c:=0 to fi.nc-1 do
            begin
            r:=d[c];
            if r<rl2 then r:=0
            else if r>ru2 then r:=1
            else r:=(r-rl)*t;
            d[c]:=r;
            s+=r;
            if r>m then m:=r;
        end;

        s*=u;
        v:=1-sp;
        if rs and (s>1e-10) and (1-s>1e-10) then
        begin
            s:=exp(ln( 1-exp(ln(1-s)*alp1) )*alp2)/s;
            if m*s>1 then { limit saturation preservation such that results are <= 1 }
            begin
                if 1-m<1e-6 then v:=1
                else if m>1e-6 then
                begin
                    m:=(m*s-1)/(m*s-exp(ln( 1-exp(ln(1-m)*alp1) )*alp2));
                    if m>v then v:=m;
                end;
            end;
        end;
        s*=1-v;

        for c:=0 to fi.nc-1 do
        begin
            r:=d^;
            if (r>1e-10) and (1-r>1e-10) then r:=exp(ln( 1-exp(ln(1-r)*alp1) )*alp2)*v + r*s;
            if st then d^:=r
            else d^:=r*(ru2-rl2)+rl2;
            d+=1;
        end;
    end;
end;

{ crop and scale brightness with non-linear tonal curve
  lq,mq,uq : lower/mid/upper quantile
  mv       : value for mid quantile
  sp       : color preservation (0..1)
  st       : stretch to [0,1]
  la,ma,ua : interpret lq/mq/uq as absolute value (not as quantile) }
procedure SImg_HistCut(cc:CSImg_Config; fi:TSImg_FPImage; lq,mq,uq,mv,sp:double; st:boolean; la,ma,ua:boolean);
var i,ya,yz    : longint;
    rl,ru,rh   : double;
    alp1,alp2  : double;
    r,t,u      : double;
    ps,qs      : array[0..2] of double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_HistCut: called with: source='+SImg_InfoStr2(fi)+
        ' lq='+real2str(lq)+' mq='+real2str(mq)+' uq='+real2str(uq)+' mv='+real2str(mv)+' sp='+real2str(sp)+
        ' st='+boolean2str(st)+' la='+boolean2str(la)+' ma='+boolean2str(ma)+' ua='+boolean2str(ua));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_HistCut: Invalid image: fi');

    if lq>uq then
    begin
        cc.warning(cc.name+': Upper quantile smaller than lower quantile, swapping them. Use `convert -sc -1'' to invert image.');
        r:=lq; lq:=uq; uq:=r;
    end;

    if (mq<=lq+1e-6) or (mq>=uq-1e-6) then cc.error(cc.name+': Middle quantile must be between upper and lower quantile, ignoring it.');
    if (mv<=1e-6) or (mv>=1-1e-6) then cc.error(cc.name+': Middle value out of range. Must be larger than 0 and less than 1.');

    ps[0]:=select(la,-1,lq);
    ps[1]:=select(ua,-1,uq);
    ps[2]:=select(ma,-1,mq);
    SImg_GetQuantile(cc,fi,cc.threads, -1,ps,qs);
    rl:=select(la,lq,qs[0]);
    ru:=select(ua,uq,qs[1]);
    rh:=select(ma,mq,qs[2]);
    if ru-rl<1e-20 then cc.error(cc.name+': Quantiles are to close or constant image');

    alp1:=1;
    if (mq>lq+1e-20) and (mq<uq-1e-20) then
    begin
        if (rh-rl>1e-18) and (ru-rh>1e-18) then
        begin
            t:=ln(mv); u:=ln((ru-rh)/(ru-rl));
            for i:=1 to 5 do
            begin
                r:=alp1;
                alp1-=(exp(u*alp1)+exp(t*alp1)-1)/(u*exp(u*alp1)+t*exp(t*alp1));
//              write(stderr,s:0:4);
                if alp1<0.5*r then alp1:=0.5*r;
                if alp1>2*r then alp1:=2*r;
//              writeln(stderr,'  ',s:0:4,'     ',exp(u*s)+exp(t*s)-1:0:6 );
            end;
        end;
    end;
    alp2:=1/alp1;
//ifdef[debug_histogr]
    cc.info(cc.name+': alp1='+real2str(alp1,-3)+' alp2='+real2str(alp2,-3));
//endif

    ya:=0;
    for i:=0 to cc.threads-1 do
    begin
        yz:=(fi.ny*(i+1)) div cc.threads;
        dothread Simg_HistCut_thread(ya,yz-1, fi, rl,ru, alp1,alp2, sp, st);
        ya:=yz;
    end;
    threadwait;

    SImg_DecDebugLevel;
end;

{ crop and scale brightness with non-linear tonal curve
  lq,mq,uq : lower/mid/upper quantile
  mv       : value for mid quantile
  sp       : saturation preservation (0..1)
  st       : stretch to [0,1]
  la,ma,ua : interpret lq/mq/uq as absolute value (not as quantile) }
procedure SImg_HistCut(cc:CSImg_Config; fi:TSImg_FPImage; lq,mq,uq,mv, gam, sp:double; st:boolean; la,ma,ua:boolean);
var i,ya,yz    : longint;
    rl,ru,r    : double;
    alp1,alp2  : double;
    ps,qs      : array[0..2] of double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_HistCut: called with: source='+SImg_InfoStr2(fi)+
        ' lq='+real2str(lq)+' mq='+real2str(mq)+' uq='+real2str(uq)+' mv='+real2str(mv)+' gam='+real2str(gam)+' sp='+real2str(sp)+
        ' st='+boolean2str(st)+' la='+boolean2str(la)+' ma='+boolean2str(ma)+' ua='+boolean2str(ua));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_HistCut: Invalid image: fi');
    if gam<-0.9999 then gam:=-0.9999;

    if lq>uq then
    begin
        cc.warning(cc.name+': Upper quantile smaller than lower quantile, swapping them. Use `convert -sc -1'' to invert image.');
        r:=lq; lq:=uq; uq:=r;
    end;

    if (mq<=lq+1e-6) or (mq>=uq-1e-6) then cc.error(cc.name+': Middle quantile must be between upper and lower quantile, ignoring it.');
    if (mv<=1e-6) or (mv>=1-1e-6) then cc.error(cc.name+': Middle value out of range. Must be larger than 0 and less than 1.');

    ps[0]:=select(la,-1,lq);
    ps[1]:=select(ua,-1,uq);
    ps[2]:=select(ma,-1,mq);
    SImg_GetQuantile(cc,fi,cc.threads, -1,ps,qs);
    rl:=select(la,lq,qs[0]);
    ru:=select(ua,uq,qs[1]);
    if ru-rl<1e-20 then cc.error(cc.name+': Quantiles are to close or constant image');

    alp2:=1+gam;
    alp1:=ln(1-exp(ln(mv)/alp2))/ln((ru-qs[2])/(ru-rl));
//ifdef[debug_histogr]
    cc.info(cc.name+': alp1='+real2str(alp1,-3)+' alp2='+real2str(alp2,-3));
//endif

    ya:=0;
    for i:=0 to cc.threads-1 do
    begin
        yz:=(fi.ny*(i+1)) div cc.threads;
        dothread Simg_HistCut_thread(ya,yz-1, fi, rl,ru, alp1,alp2, sp, st);
        ya:=yz;
    end;
    threadwait;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_HistCutGamma ***********************************************}
{*************************************************************************}
{ crop and scale brightness with non-linear tonal curve
  lq,uq    : lower/mid/upper quantile
  gam,bet  : gamma / beta
  mv       : value for mid quantile
  sp       : saturation preservation (0..1)
  st       : stretch to [0,1]
  sc       : process colors separately
  la,ua    : interpret lq/uq as absolute value (not as quantile) }
procedure SImg_HistCutGamma(cc:CSImg_Config; fi:TSImg_FPImage; lq,uq, gam,bet, sp:double; st:boolean; la,ua:boolean);
var i,ya,yz    : longint;
    r,rl,ru    : double;
    alp1,alp2  : double;
    ps,qs      : array[0..1] of double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_HistCutGamma: called with: img='+SImg_InfoStr2(fi)+
        ' lq='+real2str(lq)+' uq='+real2str(uq)+' gam='+real2str(gam)+' bet='+real2str(bet)+' sp='+real2str(sp)+
        ' st='+boolean2str(st)+' la='+boolean2str(la)+' ua='+boolean2str(ua));

    if not SImg_Recover(cc,fi) then cc.Error('SImg_HistCutGamma_int: Invalid image: fi');
    if lq>uq then
    begin
        cc.warning(cc.name+': Upper quantile smaller lower quantile, swap them. Use `convert -sc -1'' to invert image.');
        r:=lq; lq:=uq; uq:=r;
    end;

    ps[0]:=select(la,-1,lq);
    ps[1]:=select(ua,-1,uq);
    SImg_GetQuantile(cc,fi,cc.threads, -1,ps,qs);
    rl:=select(la,lq,qs[0]);
    ru:=select(ua,uq,qs[1]);
    if ru-rl<1e-20 then cc.error(cc.name+': Quantiles are to close or constant image');

    if abs(gam)+0.5*abs(bet)>0.995 then cc.error(cc.name+': Arguments out of range |gam|+|bet|/2 should be smaller than 1');

    { alp1 }
    r:=gam-bet*0.5;
    if r<-0.9999 then r:=-0.9999;
    if r> 0.9999 then r:= 0.9999;
    r:=pi*(1-r)/4; alp1:=sin(r)/cos(r);
    { alp2 }
    r:=gam+bet*0.5;
    if r<-0.9999 then r:=-0.9999;
    if r> 0.9999 then r:= 0.9999;
    r:=pi*(1-r)/4; alp2:=cos(r)/sin(r);
//ifdef[debug_histogr]
    cc.info(cc.name+': alp1='+real2str(alp1,-3)+' alp2='+real2str(alp2,-3));
//endif

    ya:=0;
    for i:=0 to cc.threads-1 do
    begin
        yz:=(fi.ny*(i+1)) div cc.threads;
        dothread Simg_HistCut_thread(ya,yz-1, fi, rl,ru, alp1,alp2, sp, st);
        ya:=yz;
    end;
    threadwait;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_HistCutLog *************************************************}
{*************************************************************************}
{ ti     : thread index
  tn     : thread number
  fi     : image input and output, input must be non-negative
  reg,ex : f(x) = (x+reg)^(ex-1)
}
threadprocedure SImg_HistCutLog_thread1(ti:longint; tn:longint; fi:TSImg_FPImage; reg:double; ex:double);
var i,ia,iz     : longint;
begin
    ia:=(fi.size*ti) div tn;
    iz:=(fi.size*(ti+1)) div tn-1;
    ex-=1;
    if reg<simg_mina then reg:=simg_mina;
    for i:=ia to iz do
        fi.pt[i]:=exp(ln(fi.pt[i]+reg)*ex);
end;

{ ti     : thread index
  tn     : thread number
  fi     : image (input and output)
  lq,uq  : upper/lower quantiles
  reg,ex : f(x) = (ln(x+reg^ex)-ex*ln(reg))*reg/ex
  mq,mv  : brightness adjustment of the form 1-(1-y)^b using mid quantile/value. Ignored if one the values is not within the range of 0 to 1.
}
threadprocedure SImg_HistCutLog_thread2(ti:longint; tn:longint; fi:TSImg_FPImage; lq:double; uq:double; reg:double; ex:double; mq:double; mv:double);
var a,b,q,r,s,rs        : double;
    i,ia,iz             : longint;
    c2                  : boolean;
begin
    ia:=(fi.size*ti) div tn;
    iz:=(fi.size*(ti+1)) div tn-1;
    uq-=lq;
    mq-=lq;
    q:=reg/ex;
    r:=ex*ln(reg);
    reg:=exp(r);
    s:=(ln(uq+reg)-r)*q;
    rs:=1/s;
    c2:=(mq>lq) and (mq<uq) and (mv>1e-6) and (mv<1-1e-6);
    if c2 then b:=ln(1-mv)/ln(1-(ln(mq+reg)-r)*q*rs )
    else b:=1;
//    if ti=0 then writeln(stderr,'b='+real2str(b));
    s/=b;
    for i:=ia to iz do
    begin
        a:=fi.pt[i]-lq;
        if a<simg_mina then a:=0
        else if a>=uq then a:=s
        else begin
            a:=(ln(a+reg)-r)*q;
            if c2 then a:=(1-exp(ln(1-a*rs)*b))*s;
        end;
        fi.pt[i]:=a;
    end;
end;

{ crop and scale brightness with regularized logarithmic tonal curve
  lq,uq : lower/upper quantile
  reg   : regularization
  ex    : saturation modification exponent
  mq,mv : brightness adjustment of the form 1-(1-y)^b using mid quantile/value. Ignored if one the values if not within the range of 0 to 1.
}
procedure SImg_HistCutLog(cc:CSImg_Config; fi:TSImg_FPImage; lq,uq, reg,ex, mq,mv:double);
var fj          : TSImg_FPImage;
    ps,qs       : array[0..2] of double;
    i,c         : longint;
    q,r,s       : double;
    d           : ^simgfp;
    c2          : boolean;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_HistCutLog: called with: img='+SImg_InfoStr2(fi)+' lq='+real2str(lq)+' uq='+real2str(uq)+' reg='+real2str(reg)+' ex='+real2str(ex)+' mq='+real2str(mq)+' mv='+real2str(mv));
    if not SImg_Recover(cc,fi) then cc.error('SImg_HistCutLog: Invalid image: fi');

    if lq<0 then lq:=0;
    if uq>1 then uq:=1;
    if uq-lq<1e-5 then cc.error(cc.name+': Lower quantile must be smaller than upper quantile');
    c2:=(mq-lq>1e-6) and (uq-mq>1e-6) and (mv>1e-6) and (mv<1-1e-6);
    if not c2 then mv:=-1;
    if reg<simg_mina then reg:=simg_mina;
    if ex<0.1 then ex:=0.1;
    if ex>4 then ex:=4;

    ps[0]:=lq;
    ps[1]:=uq;
    ps[2]:=select(c2,mq,-1);

    if (fi.nc<2) or (abs(ex-1)<1e-3) then ex:=1
    else begin
        if lq<=-simg_mina then lq:=0
        else lq:=SImg_GetQuantile(cc, fi, lq);
        { luminance image }
        fj.pt:=nil;
        SImg_New(cc, fj, fi.nx,fi.ny,1, false);
        d:=fi.pt;
        q:=1/fi.nc;
        for i:=0 to fj.size-1 do
        begin
            s:=0;
            for c:=0 to fi.nc-1 do
            begin
                r:=d[c]-lq;
                if r>0 then s+=r;
            end;
            fj.pt[i]:=s*q;
            d+=fi.nc;
        end;
        { f(x) = (x+reg)^(ex-1) }
        for i:=0 to cc.threads-1 do
            dothread SImg_HistCutLog_thread1(i,cc.threads, fj, reg,ex);
        threadwait;
        { apply transformation }
        q:=exp(ln(reg)*ex);
        d:=fi.pt;
        lq-=reg;
        for i:=0 to fj.size-1 do
        begin
            s:=fj.pt[i];
            for c:=0 to fi.nc-1 do
                d[c]:=(d[c]-lq)*s-q;
{           begin
                r:=d[c]-lq;
                if r<=0 then r:=0;
                d[c]:=(r+reg)*s-q;
            end;}
            d+=fi.nc;
        end;
        SImg_Free(cc,fj);
        ps[0]:=-1;
    end;

    if ps[0]<=-simg_mina then ps[0]:=-1;
    SImg_GetQuantile(cc, fi, cc.threads, -1,ps,qs);
    if ps[0]<=-simg_mina then qs[0]:=0;

//    cc.info('lq='+real2str(qs[0])+' uq='+real2str(qs[1])+' mq='+real2str(qs[2]));
    if qs[1]-qs[0]<simg_mina then cc.error(cc.name+': Difference between lower and upper quantile to small or constant image');
    if c2 then
    begin
        if qs[2]-qs[0]<simg_mina then cc.error(cc.name+': Difference between mid and lower quantile to small or constant image');
        if qs[1]-qs[2]<simg_mina then cc.error(cc.name+': Difference between mid and upper quantile to small or constant image');
    end;
    for i:=0 to cc.threads-1 do
        dothread Simg_HistCutLog_thread2(i,cc.threads, fi, qs[0],qs[1], reg,ex, qs[2],mv);
    threadwait;

    SImg_DecDebugLevel;
end;



{*************************************************************************}
{******* SImg_Log ********************************************************}
{*************************************************************************}
threadprocedure SImg_Log_thread(ti:longint; tn:longint; fi:TSImg_FPImage; lq:double; reg:double);
var a,r         : double;
    i,ia,iz     : longint;
begin
    ia:=(fi.size*ti) div tn;
    iz:=(fi.size*(ti+1)) div tn-1;

    if reg<simg_mina then reg:=simg_mina;
    r:=ln(reg);
    for i:=ia to iz do
    begin
        a:=fi.pt[i]-lq;
        if a<=0 then a:=0
        else a:=(ln(a+reg)-r)*reg;
        fi.pt[i]:=a;
    end;
end;

{ Logarithmically scale brightness of an image
  lq    : lower quantile
  reg   : regularization
}
procedure SImg_Log(cc:CSImg_Config; fi:TSImg_FPImage; lq,reg:double);
var i,tn     : longint;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Log: called with: img='+SImg_InfoStr2(fi)+' lq='+real2str(lq)+' reg='+real2str(reg));
    if not SImg_Recover(cc,fi) then cc.error('SImg_HistCutLog: Invalid image: fi');

    if lq<0 then lq:=0;
    lq:=SImg_GetQuantile(cc, fi,lq);

    tn:=cc.threads;
    for i:=0 to tn-1 do
        dothread Simg_Log_thread(i,tn, fi, lq,reg);
    threadwait;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_HistCutPow *************************************************}
{*************************************************************************}
{ ti     : thread index
  tn     : thread number
  fi     : image (input and output)
  lq,uq  : upper/lower limit
  reg,ex : f(x) = (x+reg^cex)^(ex/cex) - reg^ex
  cex    : color exponent
  mq,mv  : brightness adjustment of the form 1-(1-y)^b using mid quantile/value. Ignored if one the values is not within the range of 0 to 1.
}
threadprocedure SImg_HistCutPow_thread(ti:longint; tn:longint; fi:TSImg_FPImage; lq:double; uq:double; reg:double; ex:double; cex:double; mq:double; mv:double);
var a,b,r,s,rs  : double;
    i,ia,iz     : longint;
    c2          : boolean;
begin
    ia:=(fi.size*ti) div tn;
    iz:=(fi.size*(ti+1)) div tn-1;
    uq-=lq;
    mq-=lq;
    r:=0;
    if reg<simg_mina then
    begin
        r:=0;
        reg:=0;
    end
    else begin
        r:=exp(ln(reg)*ex);
        reg:=exp(ln(reg)*cex);
    end;
    ex/=cex;
    s:=exp(ln(uq+reg)*ex)-r;
    rs:=1/s;
    c2:=(mq>lq) and (mq<uq) and (mv>1e-6) and (mv<1-1e-6);
    if c2 then b:=ln(1-mv)/ln(1-(exp(ln(mq+reg)*ex)-r)*rs)
    else b:=1;
//    if ti=1 then writeln(stderr,real2str(b),'   ',real2str(r),'   ',real2str(ex),'       ',real2str(lq),'  ',real2str(mq),'  ',real2str(uq),'    ',real2str(mv),'    ',real2str((exp(ln(mq+reg)*ex)-r)*rs));
    for i:=ia to iz do
    begin
        a:=fi.pt[i]-lq;
        if a<simg_mina then a:=0
        else if a>uq then a:=s
        else begin
            a:=exp(ln(a+reg)*ex)-r;
            if c2 then a:=(1-exp(ln(1-a*rs)*b))*s;
        end;
        fi.pt[i]:=a;
    end;
end;

{ crop and scale brightness with regularized non-linear tonal curve
  lq,uq : lower/upper quantile
  reg   : regularization. Valid: >0
  ex    : exponent for the base tonal curve (x+reg)^ex
  cex   : saturation modification exponent. Valid: >0
  mq,mv : brightness adjustment of the form 1-(1-y)^b using mid quantile/value. Ignored if one the values if not within the range of 0 to 1.
}
procedure SImg_HistCutPow(cc:CSImg_Config; fi:TSImg_FPImage; lq,uq, reg,ex,cex, mq,mv:double);
var fj          : TSImg_FPImage;
    ps,qs       : array[0..2] of double;
    i,c         : longint;
    q,r,s       : double;
    d           : ^simgfp;
    c2          : boolean;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_HistCutPow: called with: img='+SImg_InfoStr2(fi)+' lq='+real2str(lq)+' uq='+real2str(uq)+' reg='+real2str(reg)+' ex='+real2str(ex)+' cex='+real2str(cex)+' mq='+real2str(mq)+' mv='+real2str(mv));
    if not SImg_Recover(cc,fi) then cc.error('SImg_HistCutPow: Invalid image: fi');

    if lq<0 then lq:=0;
    if uq>1 then uq:=1;
    if uq-lq<1e-5 then cc.error(cc.name+': Lower quantile must be smaller than upper quantile');
    c2:=(mq-lq>1e-6) and (uq-mq>1e-6) and (mv>1e-6) and (mv<1-1e-6);
    if not c2 then mv:=-1;
    if reg<0 then reg:=0;
    if ex<1e-6 then ex:=1e-6;
    if ex>10 then ex:=10;
    if cex<1e-6 then cex:=1e-6;
    if cex>10 then cex:=10;

    ps[0]:=lq;
    ps[1]:=uq;
    ps[2]:=select(c2,mq,-1);

    if (fi.nc<2) or (abs(cex-1)<1e-2) then cex:=1
    else begin
        if lq<=-simg_mina then lq:=0
        else lq:=SImg_GetQuantile(cc, fi, lq);
        { luminance image }
        fj.pt:=nil;
        SImg_New(cc, fj, fi.nx,fi.ny,1, false);
        d:=fi.pt;
        q:=1/fi.nc;
        for i:=0 to fj.size-1 do
        begin
            s:=0;
            for c:=0 to fi.nc-1 do
            begin
                r:=d[c]-lq;
                if r>0 then s+=r;
            end;
            fj.pt[i]:=s*q;
            d+=fi.nc;
        end;
        { f(x) = (x+reg)^(ex-1) }
        for i:=0 to cc.threads-1 do
            dothread SImg_HistCutLog_thread1(i,cc.threads, fj, reg,cex);
        threadwait;
        { apply transformation }
        d:=fi.pt;
        q:=exp(ln(reg)*cex);
        for i:=0 to fj.size-1 do
        begin
            s:=fj.pt[i];
            for c:=0 to fi.nc-1 do
            begin
                r:=d[c]-lq;
                if r<0 then r:=0;
                d[c]:=(r+reg)*s-q;
            end;
            d+=fi.nc;
        end;
        SImg_Free(cc,fj);
        ps[0]:=-1;
    end;

    if ps[0]<=-simg_mina then ps[0]:=-1;
    SImg_GetQuantile(cc, fi, cc.threads, -1,ps,qs);
    if ps[0]<=-simg_mina then qs[0]:=0;

    if qs[1]-qs[0]<simg_mina then cc.error(cc.name+': Difference between lower andupper quantile to small or constant image');
    if c2 then
    begin
        if qs[2]-qs[0]<simg_mina then cc.error(cc.name+': Difference between mid and lower quantile to small or constant image');
        if qs[1]-qs[2]<simg_mina then cc.error(cc.name+': Difference between mid and upper quantile to small or constant image');
    end;
    for i:=0 to cc.threads-1 do
        dothread Simg_HistCutPow_thread(i,cc.threads, fi, qs[0],qs[1], reg,ex,cex, qs[2],mv);
    threadwait;

    SImg_DecDebugLevel;
end;

