{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* system routines for FRDImages ***********************************}
{*************************************************************************}
{*************************************************************************}
//define[@rawptr ][;][
var #0   : ^byte;
    #0b  : ^byte absolute #0;
    #0w  : ^word absolute #0;
    #0s  : ^single absolute #0;
    #0d  : ^double absolute #0;
//ifdef[doubleprec]
    #0p  : ^double absolute #0;
//else
    #0p  : ^single absolute #0;
//endif ]


{*************************************************************************}
{******* fmt2str *********************************************************}
{*************************************************************************}
function fmt2str(fmt:TSImg_RawFormat):string16;
begin
    case fmt of
        SImg_RawByte   : result:='Byte';
        SImg_RawWord   : result:='Word';
        SImg_RawSingle : result:='Single';
        SImg_RawDouble : result:='Double';
        else result:='(unknown)';
    end;
end;


{*************************************************************************}
{******* SImg_GetImageTableEntry *****************************************}
{*************************************************************************}
function SImg_GetImageTableEntry(ri:TSImg_RDImage):longint;
begin
    result:=SImg_GetImageTableEntry(ri.pt, ri.dsize);
end;


{*************************************************************************}
{******* SImg_New ********************************************************}
{*************************************************************************}
procedure SImg_New(cc:CSImg_Config; var ri:TSImg_RDImage; clear:boolean);  { create new image; nx, ny, nc, fmt must be set; bole is set to SImg_SysBOLE  }
var i  : longint;
    ds : ^single;
    dd : ^double absolute ds;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_Free(cc,ri);
    SImg_IncDebugLevel;
    if (ri.nx<minxsize) or (ri.ny>maxxsize)
        or (ri.ny<minysize) or (ri.ny>maxysize)
        or (ri.nc<mincnsize) or (ri.nc>maxcnsize) then cc.Error('SImg_New (RD): illegal dimensions');
    ri.sl:=ri.nx*ri.nc;
    ri.size:=ri.sl*ri.ny;
    case ri.fmt of
        SImg_RawByte   : ri.dsize:=ri.size*sizeof(byte);
        SImg_RawWord   : ri.dsize:=ri.size*sizeof(word);
        SImg_RawSingle : ri.dsize:=ri.size*sizeof(single);
        SImg_RawDouble : ri.dsize:=ri.size*sizeof(double);
        else cc.Error('SImg_New: unknown data format');
    end;
    ri.co:=0; ri.cf:=1;
    ri.bole:=SImg_SysBOLE;

    i:=SImg_GetEmptyImageTableEntry;
    SImg_Sem_Lock(SImg_Sem_Img);
    try
        with SImg_ImageTable[i] do
        begin
            size:=ri.dsize;
            getmem2(pt,size);
            ri.pt:=pt;
        end;
    finally
        SImg_Sem_Unlock(SImg_Sem_Img);
    end;
    ri.iid:=i;
    cc.debug('SImg_New (RD): allocated '+base1024str(ri.dsize)+' bytes for '+SImg_InfoStr2(ri)+' at '+ptr2str(ri.pt));

    if not SImg_Integrity(ri) then cc.Error('SImg_New (RD): internal error');

    if clear then
    begin
        ds:=ri.pts;
        case ri.fmt of
            SImg_RawByte,
            SImg_RawWord   : fillchar(ri.pt^,ri.dsize,0);
            SImg_RawSingle : for i:=0 to ri.size-1 do
                                ds[i]:=0;
            SImg_RawDouble : for i:=0 to ri.size-1 do
                                dd[i]:=0;
        end;
    end;

    SImg_DecDebugLevel;
end;

procedure SImg_New(cc:CSImg_Config; var ri:TSImg_RDImage; fmt:TSImg_RawFormat; nx,ny,nc:longint; clear:boolean); { create new image }
begin
    ri.nx:=nx; ri.ny:=ny; ri.nc:=nc;
    ri.fmt:=fmt;
    SImg_New(cc,ri,clear);
end;

(*
procedure SImg_New(cc:CSImg_Config; var ri:TSImg_RDImage);  { create new image; nx, ny, nc, fmt, xc, yc must be set }
begin
    SImg_New(cc,ri,true);
end;

procedure SImg_New(cc:CSImg_Config; var ri:TSImg_RDImage; fmt:TSImg_RawFormat; nx,ny,nc:longint); { create new image }
begin
    ri.nx:=nx; ri.ny:=ny; ri.nc:=nc;
    ri.fmt:=fmt;
    SImg_New(cc,ri,true);
end; *)



{*************************************************************************}
{******* SImg_Free *******************************************************}
{*************************************************************************}
procedure SImg_Free(cc:CSImg_Config; var ri:TSImg_RDImage);
var i  : longint;
begin
    if cc=nil then cc:=SImg_Config_default;
    i:=SImg_GetImageTableEntry(ri);
    if (i>=0) then
    with SImg_ImageTable[i] do
    begin
        cc.debug('SImg_Free (RD): deallocating '+base1024str(size)+'bytes for '+SImg_InfoStr2(ri)+' at '+ptr2str(pt));
        SImg_Sem_Lock(SImg_Sem_Img);
        try
            freemem2(pt,size);
            ri.size:=0; ri.pt:=nil; ri.dsize:=0;
             pt:=nil; size:=0;
        finally
            SImg_Sem_Unlock(SImg_Sem_Img);
        end;
    end;
    ri.size:=0; ri.pt:=nil; ri.iid:=-1;
end;


{*************************************************************************}
{******* SImg_Move *******************************************************}
{*************************************************************************}
procedure SImg_Move(cc:CSImg_Config; var ri,rj:TSImg_RDImage);  { ri -> rj }
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Overwritealert(ri,rj) then
    begin
        SImg_Free(cc,rj);
        rj:=ri;
        ri.pt:=nil; ri.size:=0; ri.iid:=-1;
    end;
end;


{*************************************************************************}
{******* SImg_Integrity **************************************************}
{*************************************************************************}
function SImg_Integrity(ri:TSImg_RDImage):boolean;
var s : longint;
begin
    case ri.fmt of
        SImg_RawByte   : s:=sizeof(byte);
        SImg_RawWord   : s:=sizeof(word);
        SImg_RawSingle : s:=sizeof(single);
        SImg_RawDouble : s:=sizeof(double);
        else s:=0;
    end;
    SImg_Integrity:=(ri.iid>=0) and (ri.iid<SImg_ImageTableSize) and
        (SImg_GetImageTableEntry(ri)=ri.iid) and
        (ri.nx>=minxsize) and (ri.nx<=maxxsize) and
        (ri.ny>=minysize) and (ri.ny<=maxysize) and
        (ri.nc>=mincnsize) and (ri.nc<=maxcnsize) and
        (ri.sl>=ri.nx*ri.nc) and (ri.size>=ri.ny*ri.sl) and
        (ri.size*s<=ri.dsize) and (s<>0);
end;


{*************************************************************************}
{******* SImg_Recover ****************************************************}
{*************************************************************************}
{ check integrity, try to recover errors, true if ri valid }
function SImg_Recover(cc:CSImg_Config; var ri:TSImg_RDImage):boolean;
var s,i : longint;
begin
    if cc=nil then cc:=SImg_Config_default;
    case ri.fmt of
        SImg_RawByte   : s:=sizeof(byte);
        SImg_RawWord   : s:=sizeof(word);
        SImg_RawSingle : s:=sizeof(single);
        SImg_RawDouble : s:=sizeof(double);
        else s:=0;
     end;
    i:=SImg_GetImageTableEntry(ri);
    SImg_Recover:=(i>=0) and
        (ri.nx>=minxsize) and (ri.nx<=maxxsize) and
        (ri.ny>=minysize) and (ri.ny<=maxysize) and
        (ri.nc>=mincnsize) and (ri.nc<=maxcnsize);

    if SImg_Recover then
    begin
        if ri.iid<>i then
        begin
            cc.Warning('SImg_Recover: invalid iid: corrected');
            ri.iid:=i;
        end;
        if ri.sl<ri.nx*ri.nc then
        begin
            cc.warning('SImg_Recover: invalid sl: corrected');
            ri.sl:=ri.nc*ri.nc;
        end;
        if ri.size<ri.sl*ri.ny then
        begin
            cc.warning('SImg_Recover: invalid size: corrected');
            ri.size:=ri.sl*ri.ny;
        end;
        if ri.dsize<ri.size*s then
        begin
            cc.warning('SImg_Recover: invalid data size: corrected');
            ri.dsize:=ri.size*s;
        end;
    end;
end;

{ check integrity, try to recover errors, true if ri valid, set DATAMIN/DATAMAX for integer images }
function SImg_Recover(cc:CSImg_Config; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;
var i : longint;
    r : double;
begin
    result:=SImg_Recover(cc,ri);
    if (not result) or (mi=nil) then exit;
    if ri.fmt=SImg_RawByte then i:=255
    else if ri.fmt=SImg_RawWord then i:=65535
    else exit;
    if not mi.get('DATAMIN',r) then mi.add('DATAMIN',real2str(0.5*ri.cf+ri.co,10),'Minimum valid value');
    if not mi.get('DATAMAX',r) then mi.add('DATAMAX',real2str((i-0.5)*ri.cf+ri.co,10),'Maximum valid value');
end;


{*************************************************************************}
{******* SImg_OverwriteAlert *********************************************}
{*************************************************************************}
function SImg_OverwriteAlert(ri,rj:TSImg_RDImage):boolean;
var i,j:longint;
begin
    i:=SImg_GetImageTableEntry(ri);
    j:=SImg_GetImageTableEntry(rj);
    SImg_OverwriteAlert:=(i=j) and (i>=0);
end;

function SImg_OverwriteAlert(ri:TSImg_RDImage;fi:TSImg_FPImage):boolean;
var i,j:longint;
begin
    i:=SImg_GetImageTableEntry(ri);
    j:=SImg_GetImageTableEntry(fi);
    SImg_OverwriteAlert:=(i=j) and (i>=0);
end;


{*************************************************************************}
{******* SImg_InfoStr ****************************************************}
{*************************************************************************}
function SImg_InfoStr(ri:TSImg_RDImage):string63;
begin
    if SImg_GetImageTableEntry(ri)>=0
    then SImg_InfoStr:=int2str(ri.nx)+'x'+int2str(ri.ny)+'x'+int2str(ri.nc)+' '+fmt2str(ri.fmt)+'-'+condstring(ri.bole,'LE','BE')
    else SImg_InfoStr:='(unused)';
end;

function SImg_InfoStr2(ri:TSImg_RDImage):string63;
begin
    SImg_InfoStr2:='RDImg(id='+int2str(SImg_GetImageTableEntry(ri))+', fmt='+SImg_InfoStr(ri)+')';
end;


{*************************************************************************}
{******* SImg_SwapBytes **************************************************}
{*************************************************************************}
procedure SImg_SwapBytes(var ri:TSImg_RDImage);
var i            : longint;
    d            : pbyte;
    b0,b1,b2,b3  : byte;   // makes it easier to optimize
begin
    d:=ri.pt;
    case ri.fmt of
        SImg_RawDouble :
            for i:=0 to ri.size-1 do
            begin
                b0:=d[0]; d[0]:=d[7]; d[7]:=b0;
                b1:=d[1]; d[1]:=d[6]; d[6]:=b1;
                b2:=d[2]; d[2]:=d[5]; d[5]:=b2;
                b3:=d[3]; d[3]:=d[4]; d[4]:=b3;
                d+=8;
            end;
        SImg_RawSingle :
            for i:=0 to ri.size-1 do
            begin
                b0:=d[0]; d[0]:=d[3]; d[3]:=b0;
                b1:=d[1]; d[1]:=d[2]; d[2]:=b1;
                d+=4;
            end;
        SImg_RawWord :
            for i:=0 to ri.size-1 do
            begin
                b0:=d[0]; d[0]:=d[1]; d[1]:=b0;
                d+=2;
            end;
    end;
    ri.bole:=not ri.bole;
end;

{*************************************************************************}
{******* SImg_Convert ****************************************************}
{*************************************************************************}
{******* FP -> RD ********************************************************}
{ fi->ri; FP-Image -> RD-Image;
  mv=true: move instead copy, free fi
  ke=true: keep fi }
procedure SImg_Convert(cc:CSImg_Config; var fi:TSImg_FPImage; var ri:TSImg_RDImage; mv,ke:boolean);
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Convert: called with fi='+SImg_InfoStr2(fi)+' --> ri='+SImg_InfoStr2(ri)+' mv='+boolean2str(mv)+' ke='+boolean2str(ke));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Convert: Invalid image: fi');
    if not SImg_OverwriteAlert(ri,fi) then SImg_Free(cc,ri);
    if mv or (not ke) then
    begin
        ri.pt:=fi.pt;
        ri.fmt:=SImg_RawFP;
        ri.nx:=fi.nx; ri.ny:=fi.ny; ri.nc:=fi.nc;
        ri.sl:=fi.sl;
        ri.size:=fi.size;
        ri.dsize:=fi.size*sizeof(simgfp);
        ri.cf:=1; ri.co:=0;
        ri.iid:=fi.iid;
        ri.bole:=SImg_SysBOLE;
        if not ke then
        begin
            fi.iid:=-1;
            fi.pt:=nil;
            fi.size:=0;
        end;
    end
    else begin
        ri.pt:=nil;
        SImg_New(cc, ri, SImg_RawFP, fi.nx, fi.ny, fi.nc, false);
        move(fi.pt^,ri.pt^,ri.dsize);
    end;
    SImg_DecDebugLevel;
end;

{******* RD -> RD ********************************************************}
{ ri->rj; RD-Image -> RD-Image
  mi     : meta info for negative crop compensation
  mv=true: move instead of copy, free ri
  ke=true: keep ri }
procedure SImg_Convert(cc:CSImg_Config; var ri,rj:TSImg_RDImage; fmt:TSImg_RawFormat; mi:CSImg_MetaInfo; mv,ke:boolean);
var i            : longint;
    f,o, rn      : double;
    p,q,r,s,y    : double;
    p1,q1        : double;
    rk           : TSImg_RDImage;
    sa           : ansistring;
    cropcomp     : boolean;
@rawptr d;
@rawptr e;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Convert: called with ri='+SImg_InfoStr2(ri)+' --> rj='+SImg_InfoStr2(rj)+' fmt='+fmt2str(fmt)+' mv='+boolean2str(mv)+' ke='+boolean2str(ke));
    if not SImg_Recover(cc,ri) then cc.error('SImg_Convert: Invalid image: ri');
    if not SImg_OverwriteAlert(ri,rj) then SImg_Free(cc,rj);
    if not ri.bole then SImg_SwapBytes(ri);

    rn:=1e-100;
    cropcomp:=cc.cropcomp and (mi<>nil);
    if cropcomp then
    begin
        cropcomp:=((ri.fmt=SImg_RawWord) or (ri.fmt=SImg_RawByte)) and ((fmt=SImg_RawSingle) or (fmt=SImg_RawDouble)) and mi.get('RDNOISE',rn) and (-0.5*ri.cf+ri.co<rn*2) and mi.get('CROPCOMP',sa) and (sa='T');
        mi.delete('CROPCOMP');
    end;
    p:=ri.co;
    if cropcomp then
    begin
        f:=0.5*ri.cf;
        o:=ri.co;
        s:=1/(2*sqr(rn));
        p:=0; q:=0;
        p1:=0; q1:=0;
        y:=0;
        r:=-5*rn;
        repeat
            y+=exp(-sqr(r)*s);
            p+=r*y; q+=y;
            if r>o-f then
            begin
                p1+=r*y; q1+=y;
            end;
            r+=0.01*rn;
        until r>o+f;
        p:=p/q-p1/q1;
    end;

    if ri.fmt=fmt then
    begin
        if mv or (not ke) then
        begin
            rk:=ri;
            if not ke then
            begin
                ri.iid:=-1;
                ri.pt:=nil;
                ri.size:=0;
                ri.dsize:=0;
            end;
            if not SImg_Integrity(rk) then cc.error('SImg_Convert (RD->RD): Internal error');
        end
        else begin
            rk.pt:=nil;
            SImg_New(cc, rk, fmt, ri.nx, ri.ny, ri.nc, false);
            rk.cf:=ri.cf;
            rk.co:=ri.co;
            rk.bole:=ri.bole;
            move(ri.pt^, rk.pt^, ri.dsize);
        end;
    end
    else begin
        rk.pt:=nil;
        SImg_New(cc, rk, fmt, ri.nx, ri.ny, ri.nc, false);
        d:=ri.pt;
        e:=rk.pt;
        case ri.fmt of
            SImg_RawByte :
                case fmt of
                    SImg_RawWord:
                        begin
                            rk.cf:=ri.cf/256.0;
                            rk.co:=ri.co;
                            for i:=0 to ri.size-1 do
                                ew[i]:=db[i] shl 8;
                        end;
                    SImg_RawSingle :
                        begin
                            f:=ri.cf;
                            o:=ri.co;
                            if cropcomp then
                            begin
                                for i:=0 to ri.size-1 do
                                    if db[i]=0 then es[i]:=p
                                    else es[i]:=db[i]*f+o;
                            end
                            else for i:=0 to ri.size-1 do
                                es[i]:=db[i]*f+o;
                        end;
                    SImg_RawDouble :
                        begin
                            f:=ri.cf;
                            o:=ri.co;
                            if cropcomp then
                            begin
                                for i:=0 to ri.size-1 do
                                if db[i]=0 then ed[i]:=p
                                           else ed[i]:=db[i]*f+o;
                            end
                            else for i:=0 to ri.size-1 do
                                ed[i]:=db[i]*f+o;
                        end;
                    else cc.error('SImg_Convert: illegal output raw format');
                end;
            SImg_RawWord :
                case fmt of
                    SImg_RawByte :
                        begin
                            rk.cf:=ri.cf*256;
                            rk.co:=ri.co;
                            for i:=0 to ri.size-1 do
                                eb[i]:=dw[i] shr 8;
                        end;
                    SImg_RawSingle :
                        begin
                            f:=ri.cf;
                            o:=ri.co;
                            if cropcomp then
                            begin
                                for i:=0 to ri.size-1 do
                                    if dw[i]=0 then es[i]:=p
                                    else es[i]:=dw[i]*f+o;
                            end
                            else for i:=0 to ri.size-1 do
                                es[i]:=dw[i]*f+o;
                        end;
                    SImg_RawDouble :
                        begin
                            f:=ri.cf;
                            o:=ri.co;
                            if cropcomp then
                            begin
                                for i:=0 to ri.size-1 do
                                    if dw[i]=0 then ed[i]:=p
                                    else ed[i]:=dw[i]*f+o;
                            end
                            else for i:=0 to ri.size-1 do
                                ed[i]:=dw[i]*f+o;
                        end;
                    else cc.error('SImg_Convert: illegal output raw format');
                end;
            SImg_RawSingle :
                case fmt of
                    SImg_RawByte :
                        begin
                            SImg_GetMinMax(cc,ri,o,f);
                            f:=(f-o)/255.0;
                            rk.co:=o; rk.cf:=f;
                            f:=1/f;
                            for i:=0 to ri.size-1 do
                            begin
                                r:=(ds[i]-o)*f;
                                if r<0 then r:=0;
                                if r>255 then r:=255;
                                eb[i]:=round(r);
                            end;
                        end;
                    SImg_RawWord :
                        begin
                            SImg_GetMinMax(cc,ri,o,f);
                            f:=(f-o)/65535;
                            rk.co:=o; rk.cf:=f;
                            f:=1/f;
                            for i:=0 to ri.size-1 do
                            begin
                                r:=(ds[i]-o)*f;
                                if r<0 then r:=0;
                                if r>65535 then r:=65535;
                                ew[i]:=round(r);
                            end;
                        end;
                    SImg_RawDouble :
                        begin
                            f:=ri.cf;
                            o:=ri.co;
                            for i:=0 to ri.size-1 do
                                ed[i]:=ds[i]*f+o;
                        end
                    else cc.error('SImg_Convert: illegal output raw format');
                end;
            SImg_RawDouble :
                case fmt of
                    SImg_RawByte :
                        begin
                            SImg_GetMinMax(cc,ri,o,f);
                            f:=(f-o)/255.0;
                            rk.co:=o; rk.cf:=f;
                            f:=1/f;
                            for i:=0 to ri.size-1 do
                            begin
                                r:=(dd[i]-o)*f;
                                if r<0 then r:=0;
                                if r>255 then r:=255;
                                eb[i]:=round(r);
                            end;
                        end;
                    SImg_RawWord :
                        begin
                            SImg_GetMinMax(cc,ri,o,f);
                            f:=(f-o)/65535;
                            rk.co:=o; rk.cf:=f;
                            f:=1/f;
                            for i:=0 to ri.size-1 do
                            begin
                                r:=(dd[i]-o)*f;
                                if r<0 then r:=0;
                                if r>65535 then r:=65535;
                                ew[i]:=round(r);
                            end;
                        end;
                    SImg_RawSingle :
                        begin
                            f:=ri.cf;
                            o:=ri.co;
                            for i:=0 to ri.size-1 do
                                es[i]:=dd[i]*f+o;
                        end
                    else cc.error('SImg_Convert: illegal output raw format');
                end;
            else cc.error('SImg_Convert: illegal input raw format');
        end;
        if not ke then SImg_Free(cc,ri);
    end;
    SImg_Move(cc,rk,rj);
    SImg_DecDebugLevel;
end;

{******* RD -> FP ********************************************************}
{ ri->fi; RD-Image -> FP-Image
  mi     : meta info for negative crop compensation
  mv=true: move instead of copy, free ri
  ke=true: keep ri }
procedure SImg_Convert(cc:CSImg_Config; var ri:TSImg_RDImage; var fi:TSImg_FPImage; mi:CSImg_MetaInfo; mv,ke:boolean);
var rj  : TSImg_RDImage;
    d   : psimgfp;
    i   : longint;
    o,f : double;
begin
    if cc=nil then cc:=SImg_Config_default;

    SImg_IncDebugLevel;
    cc.debug('SImg_Convert: called with ri='+SImg_InfoStr2(ri)+' --> fi='+SImg_InfoStr2(fi)+' mv='+boolean2str(mv)+' ke='+boolean2str(ke));
    if not SImg_Recover(cc,ri) then cc.error('SImg_Convert: Invalid image: ri');
    if not SImg_OverwriteAlert(ri,fi) then SImg_Free(cc,fi);

    if ri.fmt<>SImg_RawFP then
    begin
        rj.pt:=nil;
        SImg_Convert(cc,ri,rj,SImg_RawFP, mi, mv, ke);
        mv:=true;
    end
    else rj:=ri;

    if rj.bole<>SImg_SysBOLE then SImg_SwapBytes(rj);

    if mv or (not ke) then
    begin
        fi.pt:=rj.pt;
        fi.nx:=rj.nx; fi.ny:=rj.ny; fi.nc:=rj.nc;
        fi.sl:=rj.sl;
        fi.size:=rj.size;
        fi.iid:=rj.iid;
        f:=rj.cf;
        o:=rj.co;
        if not ke then
        begin
            ri.iid:=-1;
            ri.pt:=nil;
            ri.size:=0;
            ri.dsize:=0;
        end;
        rj.iid:=-1;
        rj.pt:=nil;
        rj.size:=0;
        rj.dsize:=0;
        if not SImg_Recover(cc,fi) then cc.error('SImg_Convert (RD->FP): Internal error 1');
    end
    else begin
        fi.pt:=nil;
        SImg_New(cc, fi, ri.nx, ri.ny, ri.nc, false);
        move(rj.pt^,fi.pt^,ri.dsize);
        f:=rj.cf;
        o:=rj.co;
    end;

    if abs(o)+abs(f-1)>1e-20 then
    begin
        d:=fi.pt;
        for i:=0 to fi.size-1 do
            d[i]:=d[i]*f+o;
    end;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_RDCut ******************************************************}
{*************************************************************************}
{ simple cut, no expansions }
procedure SImg_RDCut(cc:CSImg_Config; ri:TSImg_RDImage; var rj:TSImg_RDImage; x1,y1,x2,y2:longint);
var rk    : TSImg_RDImage;
    d,e   : ^byte;
    i,j,y : longint;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    SImg_DebugMessage('SImg_RDCut: called with ri='+SImg_InfoStr2(ri)+' --> rj='+SImg_InfoStr2(rj)+' x1='+int2str(x1)+' y1='+int2str(y1)+' x2='+int2str(x2)+' y2='+int2str(y2));
    if not SImg_Recover(cc,ri) then cc.error('SImg_RDCut: Invalid image: ri');
    if SImg_OverwriteAlert(ri,rj) and (x1=0) and (y2=0) and (x2=ri.nx-1) and (y2=ri.ny-1) then
    begin
        SImg_DecDebugLevel;
        exit;
    end;
    if not SImg_OverwriteAlert(ri,rj) then SImg_Free(cc,rj);
    if (x1<0) or (x1>=x2) or (x2>=ri.nx) then cc.error('SImg_RDCut: Invalid X coordinates');
    if (y1<0) or (y1>=y2) or (y2>=ri.ny) then cc.error('SImg_RDCut: Invalid Y coordinates');

    rk.pt:=nil;
    SImg_New(cc, rk, ri.fmt, x2-x1+1, y2-y1+1, ri.nc, false);
    rk.cf:=ri.cf;
    rk.co:=ri.co;
    rk.bole:=ri.bole;

    i:=ri.dsize div ri.size;
    d:=ri.ptb + (y1*ri.nx+x1)*ri.nc*i;
    e:=rk.ptb;
    j:=rk.nx*rk.nc*i;
    i*=ri.nx*ri.nc;
    for y:=y1 to y2 do
    begin
        move(d^,e^,j);
        d+=i;
        e+=j;
    end;

    SImg_Move(cc,rk,rj);
    SImg_DecDebugLevel;
end;

