{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

//ignorecase

//define[{][}][//noexpand[!]{!
//nolf//noexpand[}]#0}
//nolf//noexpand[!]}!]

//define[']['][//noexpand[!]'!
//nolf//noexpand[']#0'
//nolf//noexpand[!]'!]

{$mode objfpc}

//include[configure.inc]

unit simage;
interface
uses classes,sysutils,urmatrix2,ctypes;

{$I-}
@DEFAULT_PACKRECORDS;

type psingle    = ^single;
     pdouble    = ^double;
     pextended  = ^extended;
     pbyte      = ^byte;
     pword      = ^word;
     pmatrix    = ^matrix;
//ifdef[doubleprec]
     simgfp     = double;
//else
     simgfp     = single;
//endif
     psimgfp    = ^simgfp;
     fftfp      = double;
     pfftfp     = ^fftfp;
     simgstring = ansistring;
     psimgstring = ^simgstring;
//     stackfp    = single;
//     pstackfp   = ^stackfp;
{$ifdef CPU64}
     aint       = int64;
     aword      = qword;
{$else}
     aint       = longint;
     aword      = longword;
{$endif}

    string71   = string[71];
    string63   = string[63];
    string37   = string[37];
    string31   = string[31];
    string19   = string[19];
    string16   = string[16];
    string15   = string[15];
    string8    = string[8];
    string7    = string[7];

//include[macros.inc]

//define[constants]
//include[configure.inc]

//ifdef[doubleprec]
const simg_large = 1e250;
      simg_mina  = 1e-250;
//else
const simg_large = 1e30;
      simg_mina  = 1e-30;
//endif

const rez65535 : double = 1/65535;
      rez255   : double = 1/255;

//include[config.inc]

{*************************************************************************}
{******* routines and types defined in utils.inc *************************}
{*************************************************************************}
{ dig: number of valid digits
    <=0   : -number of digits behind comma
    1..15 : absolute number of digits
    >15   : raw (freepascal) output }
function real2str(d:extended; dig:longint):string31;
function real2str(d:extended):string31;  { dig=4 }

function time2str(t:extended):string15;

function int2str(i:int64):string15;
function int2str(i:int64;j:longint):string15;

function base1024str(d:aword):string31;
function hex2str(i:dword):string15;
function ptr2str(pt:pointer):string15;
function boolean2str(b:boolean):string7;
function onoff2str(b:boolean):string7;
procedure upstr(var s:shortstring);

function condstring(b:boolean;s1,s2:shortstring):shortstring;

{ clear spaces
   l: leading
   e: ending }
procedure trimstr(var s:ansistring; l,e:boolean);
procedure trimstr(var s:shortstring; l,e:boolean);
procedure trimstr(var s:ansistring);
procedure trimstr(var s:shortstring);

{ trim single quotes }
function trimquotes(var s:ansistring):ansistring;

function min(const a,b:double):double; inline;
function max(const a,b:double):double; inline;

function str2pchar(var str:shortstring):pchar;
function str2pchar(var str:ansistring):pchar;

function gettid:longint;

{ return false on error }
function val2(s:shortstring; var d:double):boolean;

function select(const cond:boolean; const vtrue,vfalse:shortstring):shortstring; inline;
function select(const cond:boolean; const vtrue,vfalse:ansistring):ansistring; inline;
function select(const cond:boolean; const vtrue,vfalse:longint):longint; inline;
function select(const cond:boolean; const vtrue,vfalse:double):double; inline;
function select(const cond:boolean; const vtrue,vfalse:single):single; inline;
function select(const cond:boolean; const vtrue,vfalse:byte):byte; inline;

{*************************************************************************}
{******* routines defined in messages.inc ********************************}
{*************************************************************************}
{ deprecated, use cc.*}

{ messages }
procedure SImg_Info(const s:ansistring);            { emits a Info message }
procedure SImg_Hint(const s:ansistring);            { emits a Hint message }
procedure SImg_Warning(const s:ansistring);         { emits a Warning message }
procedure SImg_Error(const s:ansistring);           { emits a Error message }
procedure SImg_FatError(const s:ansistring);        { emits a Fatal message }
{ t : type of message:
       <-3 : debug
       -3  : Info
       -2  : Hint
       -1  : Warning
        0  : Error
       >0  : Fatal
  mes  : message String
  print: write to stdout }
procedure SImg_StdPrintMessage(t:longint; const mes:ansistring; print:boolean);

{ debugging }
const SImg_HomeDir  : string = '';

procedure SImg_IncDebugLevel;
procedure SImg_DecDebugLevel;


{*************************************************************************}
{******* routines defined in thread.inc **********************************}
{*************************************************************************}
type CSImg_Thread=class(TThread)
       procedure Wakeup;             { wake-up thread (delay interrupting) }
       end;

type CSImg_AutoThread = class(TThread)
         constructor create(ow:boolean);
         destructor destroy; override;
         procedure wait;
       protected
         isrunning,oneway,isrunning2 : boolean;
         end;

procedure SImg_Halt(ec:longint);
procedure SImg_Halt;
procedure SImg_WakeUp(tid:longint);   { for delay interrupting }


{*************************************************************************}
{******* routines defined in configclass.inc *****************************}
{*************************************************************************}
const SImg_ALign_top = 0;
      SImg_ALign_mitop = 0;

type TSImg_PrintMessageProc=procedure(t:longint; const mes:simgstring; print:boolean);

type CSImg_Config=class  { configuration class }
         enableDebug,enableInfos, enableHints,
         enableWarnings, enableErrors  : boolean; { enable/disable printing different kind of messages }
         { logging }
         logFileNum      : longint;               { number of log files, <0 to disable logging }
         logFileBase     : string[31];            { log file base name }
         { endianess }
         pnmLE           : boolean;               { default PNM endianess for reading }
         forceBE,forceLE : boolean;               { force bid/little endian for writing, if file format supports this }
         { other }
         cropComp        : boolean;               { compensate value cropping during int->fp conversion if RDNOISE is given }
         { threads }
         threads   : longint;                     { number of threads }
         name      : string[31];                  { program name }
         { draw settings }
         textVAlign   : single;                            { absolute value: vertical alignment relative to height: 0 means top. sign: <0 to enables position adjustment if border would be reached  }
         textHAlign   : single;                            { absolute value: horizontal alignment relative to height: 0 means top. sign: <0 to enables position adjustment if border would be reached }
         textBack     : boolean;                           { draw background for text }
         drawColor    : array[0..maxcnsize-1] of simgfp;   { drawing color }
         backColor    : array[0..maxcnsize-1] of simgfp;   { background color }
         constructor create(cc:CSImg_Config);     { create new configuration class with values from cc }
         { message types:
            <-3 : Debug
            -3  : Info
            -2  : Hint
            -1  : Warning
             0  : Error
            >0  : Fatal }
         procedure Debug(const s:ansistring);             { generates debug message }
         procedure Info(const s:ansistring);              { generates info message }
         procedure Hint(const s:ansistring);              { generates hint }
         procedure Warning(const s:ansistring);           { generates warning }
         procedure Error(const s:ansistring);             { generates error }
         procedure FatError(const s:ansistring);          { generates fatal error }
         procedure PrintMessage(t:longint; const mes:ansistring; print:boolean);    { print message }
         procedure SetMessageProc(mp:TSImg_PrintMessageProc);
    { ***** private ***** }
    private
         prev            : CSImg_Config;       { descendant }
         { messages internal }
         MessageProc     : TSImg_PrintMessageProc;
         MessageProcSet  : boolean;
     end;

const SImg_Config_default : CSImg_Config = nil;


{*************************************************************************}
{******* routines defined in system.inc **********************************}
{*************************************************************************}
{ Exceptions }
type CSImg_ErrorException=Class(Exception);
     CSImg_QuiteException=Class(Exception);

function SImg_ExceptionToMessage(obj:tobject; addr:pointer; framecount:longint; frames:ppointer):simgstring;
function SImg_ExceptionToMessage:simgstring;
procedure SImg_QuiteException;
procedure SImg_OutputCatchedException(cc:CSImg_Config);

{ memory management }
procedure getmem2(var ptr:pointer;size:aword);
procedure freemem2(var ptr:pointer;size:aword);
procedure freemem2(var ptr:pointer);


{*************************************************************************}
{******* routines defined in meta-info.inc *******************************}
{*************************************************************************}
type CSImg_MetaInfo=class  { class storing universal meta information }
       private
         cnt,reserved   : longint;
         keys,vals,coms : pansistring;
         cs             : TRTLCriticalSection;
       public
         constructor create;
         constructor create(mi:CSImg_MetaInfo);         { create as a clone }
         destructor destroy; override;
         procedure reset;
         procedure add(const theKey,theVal,theComment:ansistring);
         function add(sa:ansistring):ansistring;                { add encoded key of the form <key>=<value>[/<comment>] and return an error string if an error occurred or '' otherwise }
         procedure add(mi:CSImg_MetaInfo);
         procedure delete(const theKey:ansistring);
         procedure deleteRaw;   { delete standard raw attributes }
         procedure rename(const oldKey,newKey:ansistring);
         function get(const theKey:ansistring; var theVal:ansistring; theComment:pansistring):boolean;  // returns false if meta information does not exist
         function get(const theKey:ansistring; var theVal:ansistring):boolean;
         function get(const theKey:ansistring; var d:double; theComment:pansistring):boolean;
         function get(const theKey:ansistring; var d:double):boolean;
         function get(const theKey:ansistring; var s:single):boolean;
         { Get value in degree. Sexagesimal values (separators ':' or ' ') are multiplied with cf. }
         function getDeg(const theKey:ansistring; cf:double; var d:double; theComment:pansistring):boolean;
         { Get telescope position in degree. Return false if unknown }
         function getRaDec(var ra,dec:double):boolean;                  // approximate equatorial position in degree, false if unknown
         { Get start of observation date and time as unix epoch time. Return false if unknown }
         function getDateTime(var dt:int64):boolean;
         property size    : longint read cnt;
         property key     : pansistring read keys;
         property val     : pansistring read vals;
         property comment : pansistring read coms;
         end;

{ WCS data. SIP distortions up to order 2 are supported }
type TSImg_WCS = record
       ref_x,ref_y                    : double;  // reference pixels
       ref_dec,ref_ra                 : double;  // coordinates of reference pixel in degrees
       cd1_1,cd1_2,cd2_1,cd2_2        : double;  // transformation matrix in degree/pixel
       a0_0,a0_1,a1_0,a0_2,a1_1,a2_0  : double;  // order 0..2 SIP distortion factors for x axis
       b0_0,b0_1,b1_0,b0_2,b1_1,b2_0  : double;  // order 0..2 SIP distortion factors for y axis
       t                              : string[3];  // currently supported: 'tan', 'sin', 'arc', 'stg'
       end;

{ returns false if meta WCS data is incomplete }
function SImg_GetWCSData(cc:CSImg_Config; mi:CSImg_MetaInfo; var wcs:TSImg_WCS):boolean;

{ calculates equatorial position of a pixel using WCS data
  x,y :  WCS coordinates, bottom left pixel is (1,1) }
procedure SImg_GetRaDec(cc:CSImg_Config; const wcs:TSImg_WCS; x,y:double; var ra,dec:double);

{ calculates equatorial position of a pixel using WCS data
  x,y :  picture coordinate, to left pixel is (0,0) }
procedure SImg_GetRaDec(cc:CSImg_Config; const wcs:TSImg_WCS; x,y,height:double; var ra,dec:double);

{ Calculates approximate pixel position in fits coordinates using WCS data. (Distortions are ignored.)
  ra,dec  : equatorial position
  result  : false if WCS transformation matrix is singular
}
function SImg_GetPos(const wcs:TSImg_WCS; ra,dec:double; var xo,yo:double):boolean;

{ Calculates approximate pixel position in fits coordinates using WCS data. (Distortions are ignored.)
  ra,dec  : equatorial position
  height  : picture height
  result  : false if WCS transformation matrix is singular
}
function SImg_GetPos(const wcs:TSImg_WCS; ra,dec,height:double; var xo,yo:double):boolean;

{ calculates pixel scale at reference point in arcsecs. dx is positive, dy has the same sign as the determinant }
procedure SImg_GetPixelScale(const wcs:TSImg_WCS; var dx,dy:double);


{*************************************************************************}
{******* routines defined in fp-sys.inc **********************************}
{*************************************************************************}
type TSImg_FPImage=record    { floating point image }
       pt        : ^simgfp;  { pointer to data }
       nx,ny,nc  : longint;  { dimensions }
       sl        : longint;  { nx*nc }
       size      : int64;    { nx*ny*nc }
       iid       : longint;  { image id (must be equal to SImg_GetImageTableEntry(self) }
       end;
     PSImg_FPImage=^TSImg_FPImage;

procedure SImg_New(cc:CSImg_Config; var r:TSImg_FPImage; clear:boolean);                        { create new image; nx, ny and nc must be set }
procedure SImg_New(cc:CSImg_Config; var r:TSImg_FPImage;nx,ny,nc:longint; clear:boolean);       { mx=my=-1 }
procedure SImg_New(cc:CSImg_Config; var r:TSImg_FPImage);                                       { create new image; nx, ny and nc must be set }
procedure SImg_New(cc:CSImg_Config; var r:TSImg_FPImage;nx,ny,nc:longint);
procedure SImg_Free(cc:CSImg_Config; var fi:TSImg_FPImage);                                     { destroy image }
procedure SImg_Move(cc:CSImg_Config; var fi,fj:TSImg_FPImage);                                  { fi -> fj (fi will be destroyed) }
function SImg_CheckConstant(fi:TSImg_FPImage):boolean;                                          { checks whether image is constant }
function SImg_Integrity(fi:TSImg_FPImage):boolean;                                              { check integrity, true if valid }
function SImg_Recover(cc:CSImg_Config;var fi:TSImg_FPImage):boolean;                            { true, if valid }
function SImg_OverwriteAlert(fi,fj:TSImg_FPImage):boolean;                                      { true, if images are the same }
procedure SImg_Duplicate(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; dat:boolean); { clone an image; fi-->fj; dat=true: copy data }
function SImg_InfoStr(fi:TSImg_FPImage):string19;
function SImg_InfoStr2(fi:TSImg_FPImage):string37;

{*************************************************************************}
{******* routines defined in rd-sys.inc **********************************}
{*************************************************************************}
type TSImg_RawFormat=(SImg_RawByte:=0,SImg_RawWord,SImg_RawSingle,SImg_RawDouble);
const
//ifdef[doubleprec]
      SImg_RawFP   = SImg_RawDouble;
//else
      SImg_RawFP   = SImg_RawSingle;
//endif

{$IFDEF ENDIAN_BIG}
    SImg_SysBOLE   = false;     { natural byte order is big endian }
{$ELSE}
    SImg_SysBOLE   = true;      { natural byte order is little endian }
{$ENDIF}


type TSImg_RDImage=record       { raw date image }
       case smallint of
         0  : ( pt       : pointer;
                fmt      : TSImg_RawFormat; { format of the image data }
                nx,ny,nc : longint;         { dimensions }
                sl       : longint;         { size of a line in pixels = nx*nc }
                size     : int64;           { size in pixels: nx*ny*nc }
                dsize    : int64;           { size in bytes }
                cf,co    : double;          { conversion out=in*fact+offs }
                iid      : longint;         { image id (must be SImg_GetImageTableEntry(self) }
                bole     : boolean;         { true: byte order is little endian }
              );
         1  : ( ptb : ^byte; );
         2  : ( ptw : ^word; );
         3  : ( pts : ^single; );
         4  : ( ptd : ^double; );
         5  : ( ptp : ^simgfp; );
       end;

function fmt2str(fmt:TSImg_RawFormat):string16;                                 { format to string }

procedure SImg_New(cc:CSImg_Config; var ri:TSImg_RDImage; clear:boolean);       { create new image; nx, ny, nc, fmt must be set; bole is set to SImg_SysBOLE  }
procedure SImg_New(cc:CSImg_Config; var ri:TSImg_RDImage; fmt:TSImg_RawFormat; nx,ny,nc:longint; clear:boolean); { create new image }
procedure SImg_Free(cc:CSImg_Config; var ri:TSImg_RDImage);                     { destroy image }
procedure SImg_Move(cc:CSImg_Config; var ri,rj:TSImg_RDImage);                  { ri -> rj }
function SImg_Integrity(ri:TSImg_RDImage):boolean;                              { check integrity, true if ri valid }
function SImg_Recover(cc:CSImg_Config; var ri:TSImg_RDImage):boolean;           { check integrity, try to fix errors, true if ri valid }
function SImg_Recover(cc:CSImg_Config; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;        { check integrity, try to recover errors, true if ri valid, set DATAMIN/DATAMAX for integer images }
function SImg_OverwriteAlert(ri,rj:TSImg_RDImage):boolean;                      { true, if images are the same }
function SImg_OverwriteAlert(ri:TSImg_RDImage;fi:TSImg_FPImage):boolean;
function SImg_InfoStr(ri:TSImg_RDImage):string63;
function SImg_InfoStr2(ri:TSImg_RDImage):string63;
procedure SImg_SwapBytes(var ri:TSImg_RDImage);
procedure SImg_Convert(cc:CSImg_Config; var fi:TSImg_FPImage; var ri:TSImg_RDImage; mv,ke:boolean);             { fi->ri; FP-Image -> RD-Image; mv=true: move (instead copy), free fi; ke=true: keep fi }
procedure SImg_Convert(cc:CSImg_Config; var ri,rj:TSImg_RDImage; fmt:TSImg_RawFormat; mi:CSImg_MetaInfo; mv,ke:boolean); { ri->rj; RD-Image -> RD-Image; mi: meta info for negative crop compensation;  mv=true: move (instead copy), free ri; ke=true: keep ri }
procedure SImg_Convert(cc:CSImg_Config; var ri:TSImg_RDImage; var fi:TSImg_FPImage; mi:CSImg_MetaInfo; mv,ke:boolean);   { ri->fi; RD-Image -> FP-Image; mi: meta info for negative crop compensation;  mv=true: move (instead copy), free ri; ke=true: keep ri }
procedure SImg_RDCut(cc:CSImg_Config; ri:TSImg_RDImage; var rj:TSImg_RDImage; x1,y1,x2,y2:longint);             { simple cut, no expansions, no color cut }


{*************************************************************************}
{******* routines defined in fp-base0.inc ********************************}
{*************************************************************************}
{ todo: cc versions }
procedure SImg_Clear(cc:CSImg_Config; fi:TSImg_FPImage; k:longint);                     { fi(*,*,l):=0; }
procedure SImg_Clear(cc:CSImg_Config; fi:TSImg_FPImage);                                { fi(*,*,*):=0; }

procedure SImg_Copy(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint;xt,yt:longint);    { fi(*,*,l):=0; copy fh(y-yt,x-xt,k) --> fi(y,x,l), shift & cut }
procedure SImg_Copy(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint);                  { fi(*,*,l):=0; copy fh(y   ,x   ,k) --> fi(y,x,l), cut }

procedure SImg_Add(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint;xt,yt:longint);     { fi(y,x,l)+=fh(y-yt,x-xt,k), shift & cut }
procedure SImg_Add(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint);                   { fi(y,x,l)+=fh(y   ,x   ,k), cut }

procedure SImg_Mul(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint;xt,yt:longint);     { fi(y,x,l)*=fh(y-yt,x-xt,k), shift & cut, undefined pixels assumed 1 }
procedure SImg_Mul(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint);                   { fi(y,x,l)*=fh(y   ,x   ,k), cut, undefined pixels assumed 1 }

procedure SImg_Gauss(cc:CSImg_Config; fi:TSImg_FPImage; k,xc,yc,rx,ry:longint; sx,sy:double);  { k:color; xc,yc: center; 2*rx+1 x 2*ry+1: size; sx,sy: sigma }
function SImg_GetSum(cc:CSImg_Config; fi:TSImg_FPImage):double;
function SImg_GetSum(cc:CSImg_Config; fi:TSImg_FPImage; c:longint):double;
function SImg_GetMean(cc:CSImg_Config; fi:TSImg_FPImage):double;
function SImg_GetMean(cc:CSImg_Config; fi:TSImg_FPImage; c:longint):double;
procedure SImg_AddScalar(cc:CSImg_Config; fi:TSImg_FPImage; a:double);
procedure SImg_AddScalar(cc:CSImg_Config; fi:TSImg_FPImage; c:longint; a:double);
procedure SImg_MulScalar(cc:CSImg_Config; fi:TSImg_FPImage; a:double);
procedure SImg_MulScalar(cc:CSImg_Config; fi:TSImg_FPImage; c:longint; a:double);
procedure SImg_CropNegative(cc:CSImg_Config; fi:TSImg_FPImage);
procedure SImg_CropNegative(cc:CSImg_Config; fi:TSImg_FPImage; c:longint);


{*************************************************************************}
{******* routines defined in fp-base1.inc ********************************}
{*************************************************************************}
procedure SImg_MinMax(cc:CSImg_Config; mi:boolean; src1,src2:TSImg_FPImage; var dest:TSImg_FPImage; f1,f2:double);
procedure SImg_Mul(cc:CSImg_Config; src1,src2:TSImg_FPImage; var dest:TSImg_FPImage; e1,e2:double);  { dest=src1*|src1|^(e1-1) * src2*|src2|^(e2-1) }
procedure SImg_Power(cc:CSImg_Config; src:TSImg_FPImage; var dest:TSImg_FPImage; ex:double);   { dest=src*|src|^(ex-1) }
procedure SImg_FlipC(cc:CSImg_Config; fi:TSImg_FPImage); { fi(x,y,c) ---> fi(x,y,nc-1-c) }

{ fi    : input
  fo    : output
  rot   : CW rotation angle in degree. Rounded to 90° steps.
  flipx : flip in X direction (before rotation)
  flipy : flip in Y direction (before rotation)
}
procedure SImg_RotateFlip(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; rot:longint; flipx,flipy:boolean);

{ cut an image
   x1,x2,y1,y2,c1,c2  :
         * area to cut
         * if values are <0 or >=fi.n[xyc], a border with background value bg added
         * if x1>x2, y1>y2, c1>c2, image is flipped
   bg  : background value
}
procedure SImg_Cut(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; x1,y1,x2,y2,c1,c2,rot:longint; bg:double);

procedure SImg_FindCenter(cc:CSImg_Config; fi:TSImg_FPImage; var xc,yc:double);

procedure SImg_FindMaximum(cc:CSImg_Config; fi:TSImg_FPImage; var xc,yc:double);

procedure SImg_AddColors(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage);

{ fi     : input
  fj     : output
  xc,yc  : center
  dr,da  : radius/angle step
  a1,a2  : start/end angle (in degree) }
procedure SImg_CartesianToPolar(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; xc,yc, dr,da,a1,a2:double);

{ fi     : input
  fj     : output
  nx,ny  : center
  xc,yc  : center
  dr,da  : radius/angle step
  a1     : start angle (in degree) }
procedure SImg_PolarToCartesian(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; nx,ny:longint; xc,yc, dr,da,a1:double);

{ fi        : input
  fj1,fj2   : output
  threshold : interpolate, if deviation is larger then threshold (absolute value)
  swap      : swap half frames }
procedure SImg_DeInterlace(cc:CSImg_Config; fi:TSImg_FPImage; var fj1,fj2:TSImg_FPImage; threshold:double; swap:boolean);

procedure SImg_WeightToNoise(cc:CSImg_Config; src:TSImg_FPImage; var dest:TSImg_FPImage);
procedure SImg_WeightToSNoise(cc:CSImg_Config; src:TSImg_FPImage; var dest:TSImg_FPImage);


{*************************************************************************}
{******* rawcorrect.inc **************************************************}
{*************************************************************************}
{ Calculate (dead/bad) pixel mask using average and square noise image as calculated by SImg_AvgSet
    avg   : avg image
    sn    : square noise image
    dest  : mask image output
    rn    : read out noise in image units
    gain  : gain in electrons per image unit
}
procedure SImg_GetPixelMask(cc:CSImg_Config; avg,sn:TSImg_FPImage; var dest:TSImg_FPImage; rn,gain:double);

{ Calculates background brightness
  fi    : input
  gain  : gain in electrons per image unit
}
function SImg_BGBrightness(cc:CSImg_Config; fi:TSImg_FPImage; gain:double):double;

{ Calculates background noise
  fi    : input
  rn    : read out noise in image units
  gain  : gain in electrons per image unit
}
function SImg_BGNoise(cc:CSImg_Config; fi:TSImg_FPImage; rn,gain:double):double;

{ single image square noise estimator, see ../test/noise-estimator.pas, y5
    fi          : input image
    dest        : output image
    bn          : Background noise, can be calculated using SImg_BGNoise(cc,fi,rn,gain);
    gain        : gain in electrons per image unit
    add         : square noise estimation is added to dest
    result      : maximum underestimation error
}
function SImg_NoiseEstimator(cc:CSImg_Config; fi:TSImg_FPImage; var dest:TSImg_FPImage; bn,gain:double; add:boolean):double;

{ subtract variable dark current src2 from src1 (difference-fitting)
  w1,w2  : weights
  st     : step size (2 for Bayer images, 0 otherwise
  result : subtraction factor
}
function SImg_DarkCurrentSubtraction(cc:CSImg_Config; src1,src2,w1,w2:TSImg_FPImage; var dest:TSImg_FPImage; st:longint):double;

{ subtract variable amp glow src2 from src1 (normal-fitting)
  w1,w2  : weights
  result : subtraction factor
}
function SImg_AmpGlowSubtraction(cc:CSImg_Config; src1,src2,w1,w2:TSImg_FPImage; var dest:TSImg_FPImage):double;

{ Reduces line patterns caused by power supply noise
  fi    : input
  zi    : dark field, zi.pt=nil to ignore it
  oi    : flat field, oi.pt=nil to ignore it
  fj    : output
  fmin  : limiting frequency
  o1    : enable order 1 corrections
}
procedure SImg_PSDenoise(cc:CSImg_Config; fi,zi,oi:TSImg_FPImage; var fj:TSImg_FPImage; fmin:double; o1:boolean);

{ Count photons. Output image is an integer image (16 unsigned) with an offset of 16 (16: 0 electrons, 17: 1 electron, ...)
  Scaling parameters preserve original scaling.
  fi    : input
  ri    : output
  gain  : gain
  nq    : noise quantization for better compression, about 0.5% SNR loss.
}
procedure SImg_PhotonCount(cc:CSImg_Config; fi:TSImg_FPImage; var ri:TSImg_RDImage; gain:double; nq:boolean);


{*************************************************************************}
{******* routines defined in filter.inc **********************************}
{*************************************************************************}
const FWHM_TO_SIG = 1/sqrt(8*ln(2));
      SIG_TO_RAD  = sqrt(2);
      FWHM_TO_RAD = FWHM_TO_SIG * SIG_TO_RAD;

{ fi    : input
  fj    : output
  rad   : filter radius
  algo  : -5 : reciprocal average ( 1/(1/x1+1/x2+...)) ) of round region
          -4 : quantile of round region
          -3 : average of round region
          -2 : quantile of square region
          -1 : average of square region
           0 : Gaussian: exp(-sqr(r/rad))
           1 : drop min of square region
           2 : drop max of square region
           3 : drop min and max of square region
  ss    : step size
  q     : percentage for quantile filter
}
procedure SImg_Filter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; algo:longint; ss:longint; q:double);

{ q=0.5 }
procedure SImg_Filter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; algo:longint; ss:longint);

{ Gaussian filter: exp(-sqr(r/rad)) }
procedure SImg_Filter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double);

{ Gauss filter with variable radius : exp( -sqr(r) / (2*sqr(sig*frad)) )
  for internal use only.
  The radius is controlled with iterations. This is not verry exact but fast.
  fi    : input
  fj    : output
  sig   : sigma
  frad  : radius factor: 0 means 1.0, 255 means 0.0, frad.size=fi.size or frad.nx*frad.ny=fi.nx*fi.ny
}
procedure SImg_Filter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; sig:double; frad:TSImg_RDImage);

{ Foreground / background approximation filter
  fi    : input
  fj    : output
  mask  : mask, one byte per pixel, 0 means masked out, optional
  rx,ry : approximation radius
  cs    : corner approximation factor, feasible 0..1
  bg    : true for background
  rerun : input is already filtered
}
procedure SImg_FBGFilter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; mask:TSImg_RDImage; rx,ry,cs:double; bg,rerun:boolean);
procedure SImg_FBGFilter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; bg,rerun:boolean);
procedure SImg_FBGFilter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; bg:boolean);

{ Calculate the noise floor
  fi    : input image
  sr    : super resolution factor >=1
}
function SImg_NoiseFloor(cc:CSImg_Config; fi:TSImg_FPImage; sr:double):double;

{*************************************************************************}
{******* routines defined in denoise.inc *********************************}
{*************************************************************************}
{ Denoising
    fi     : input and output image
    fwhm   : fwhm of input image
    rad    : radius parameter, relative to fwhm, >=0, recommended: 0..2
    sp     : structure preservation. 0..1
    thr    : denoising threshold as quantile
}
procedure SImg_Denoise(cc:CSImg_Config; fi:TSImg_FPImage; fwhm,rad,sp,thr:double);

{ Multi channel denoising by low-pass filtering the color information (=channel ratios).
  fi        : input/output channel images
  wi        : input/output pixel weight images
  ifwhm     : input FWHM, <=0 means undefined (will be calculated automatically if undefined)
  no        : maximum number of output channels to calculate
  qr        : reference quantile
  rcfwhm1   : output color FWHM relative to maximum input FWHM at qr-quantile
  rcfwhm0   : output color FWHM relative to maximum input FWHM below (qr/4)-quantile
  rfwhm     : maximum relative (to the input FWHM) FWHM which is taken into account during channel denoising. Values <=0 are interpreted as infinity.
  afwhm     : maximum absolute FWHM which is taken into account during channel denoising. The maximum max(afwhm,ifwhm[i]*rfwhm) will be used.
}
procedure SImg_CDenoise(cc:CSImg_Config; var afi,awi:array of TSImg_FPImage; ifwhm:pdouble; no:longint; qr,rcfwhm1,rcfwhm0,rfwhm,afwhm:double);

{ Denoisung of dark regions
    fi     : input and output image
    fwhm   : input FWHM
    rad    : maximum low pass filter radius
    qh     : no denoising above that quantile
    ql     : maximum denoising below that quantile
}
procedure SImg_BGDenoise(cc:CSImg_Config; fi:TSImg_FPImage; fwhm,rad,qh,ql:double);

{*************************************************************************}
{******* routines defined in deconv.inc **********************************}
{*************************************************************************}
{ Deconvolution and denoising
    fi     : input image (1 color image)
    fo     : output image
    ifwhm  : input FWHM
    ofwhm  : output FWHM
    dn     : denoising strength (0..1)
    ddn    : Additional denoising of regions darker than that quantile.
    bdn    : Additional denoising at boundaries. Relative to image dimension. 0..0.5
    bg     : extra lower constraints. Recommended: 0..1. <0.1 means disable
    ns     : no star mode
    snr    : SNR parameter for wiener deconvolution
}
procedure SImg_DeconvDenoise(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; ifwhm,ofwhm, dn,ddn,bdn, bg:double; ns:boolean; snr:double);


{*************************************************************************}
{******* routines defined in interpolate.inc *****************************}
{*************************************************************************}
{ Interpolate masked pixels
  fi      : input and output
  mask    : interpolate if mask and pm<>0
  pm      : mask pattern
  fd      : filter divider, as larger as less low pass filtering. Disabled if >127.
}
procedure SImg_InterpolateMasked(cc:CSImg_Config; var fi:TSImg_FPImage; mask:TSImg_RDImage; pm,fd:byte);
procedure SImg_InterpolateMasked(cc:CSImg_Config; var fi:TSImg_FPImage; mask:TSImg_RDImage; pm:byte); { fd=3 }


{*************************************************************************}
{******* routines defined in starsubtract.inc ****************************}
{*************************************************************************}
type TSImg_StarSubtract_info = record
        sig,sig2   : simgfp; { noise floor, star detection limit }
        sat        : simgfp; { saturation level }
     end;

{ Calculate PSF
    fi     : input image (1 color image)
    fpsf   : output PSF
    tol    : FWHM tolerance as factor
    rad    : PSF radius
    rm     : star mask. Or with mask at position found by SImg_FindStars (optional)
    mask   : see rm
}
procedure SImg_GetPSF(cc:CSImg_Config; fi:TSImg_FPImage; var fpsf:TSImg_FPImage; tol:double; rad:longint; rm:TSImg_RDImage; mask:byte);

{ Subtract stars
    fi     : input (1 color image)
    fpsf   : input PSF
    rms    : mask input and output. Possible inputs:
        1. undefined
        2. mask
            bit 0: interpolated pixels (output only)
            bit 1: star (input and output)
            bit 2: save star (input and output )
        3. floating point (SImg_RawFP) containing deconvoluted stars that will be brightness filtered
    fob    : output image: background
    fos    : output image: deconvoluted stars
    fod    : output image: difference between input and fob where saturated regions are extrapolated
    ofob   : true: output fob; false: do not touch fob
    ofos   : true: output fos; false: do not touch fos
    ofod   : true: output fod; false: do not touch fod
    getmask: bit 0: find new stars
             bit 1: filter input candidates, otherwise they will be set to save stars.
    tol    : psf width tolerance as factor in center
    tolc   : psf width tolerance as factor in corners
    ff     : Source filter strength >=0, high values ==> less objects are considered as stars
    nf     : Noise correction factor. Usually 1.
    ithr   : Interpolation threshold relative to noise
    iwdt   : Interpolation width
    cwdt   : PSF correction width relative to PSF. Valid 0..1
    tn     : PSF correction tiles into direction of larger dimension; <0.5 to disable PSF correction
}
function SImg_StarSubtract(cc:CSImg_Config; fi,fpsf:TSImg_FPImage; var rms:TSImg_RDImage; var fob,fos,fod:TSImg_FPImage; ofob,ofos,ofod:boolean; getmask:longint; tol,tolc: double; ff,nf,ithr,iwdt,cwdt:double; tn:double):TSImg_StarSubtract_Info;

procedure SImg_StarSubtract_WidthFilter_tests(cc:CSImg_Config);


{*************************************************************************}
{******* routines defined in stars.inc ***********************************}
{*************************************************************************}
{ Calculate star FWHM (in pixels) map
  fi      : input
  fj      : output
  dmin    : minimum FWHM
  rad     : search radius. Recommended 1.5..20 times the PSF width
  res     : weight smoothing relative to image dimension. Recommended: 2..10
  result  : total FWHM in pixels
}
function SImg_FWHMMap(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; dmin:double; rad:longint; res:double):double;


{ Calculate star FWHM in pixels
  fi      : input
  dmin    : minimum FWHM
  rad     : search radius. Recommended 20*dmin
  fwhm    : array to store the FWHM per color
  result  : total FWHM in pixels
}
function SImg_GetSharpnessFWHM(cc:CSImg_Config; fi:TSImg_FPImage; dmin:double; rad:longint; var fwhm:array of double):double;
function SImg_GetSharpnessFWHM(cc:CSImg_Config; fi:TSImg_FPImage; dmin:double; rad:longint):double;
function SImg_GetSharpnessFWHM(cc:CSImg_Config; fi:TSImg_FPImage):double;


{*************************************************************************}
{******* routines defined in info.inc ************************************}
{*************************************************************************}
type TSImg_StatInfo = record  { some statistic information }
       minv,maxv : double;    { minimal, maximal value }
       av        : double;    { average value }
       md        : double;    { mean deviation }
       rms       : double;    { root mean square (deviation) }
       end;

procedure SImg_GetStatInfo(cc:CSImg_Config; fi:TSImg_FPImage; var si:TSImg_StatInfo);

{ Get minval and maxval.
  For 8 bit images these are 0 and 255.
  For 16 bit images these are 0 and 65536.
  cf anf co are ignored. }
procedure SImg_GetMinMax(cc:CSImg_Config; ri:TSImg_RDImage; var minv,maxv:double);
procedure SImg_GetMinMax(cc:CSImg_Config; fi:TSImg_FPImage; var min,max:double);

{ Egde detect algorithm, best choice if sensor resolution is higher than optical resolution
  thr: threshold; feasible: 0..1; 0.999..1 for stars
  returned value is the geometric mean of single color values }
function SImg_GetSharpness1(cc:CSImg_Config; fi:TSImg_FPImage; thr:double; var sharpness:array of double):double;
function SImg_GetSharpness1(cc:CSImg_Config; fi:TSImg_FPImage; thr:double):double;

{ Egde detect algorithm, optical resolution must be close or higher than sensor resolution
  thr: threshold; feasible: 0..1; 0.999..1 for stars
  returned value is the geometric mean of single color values }
function SImg_GetSharpness2(cc:CSImg_Config; fi:TSImg_FPImage; thr:double; var sharpness:array of double):double;
function SImg_GetSharpness2(cc:CSImg_Config; fi:TSImg_FPImage; thr:double):double;

{ Peak detect algorithm optimized for stars, optical resolution must be close to or higher than sensor resolution
  thr: threshold; valid: 0..1, feasible: 0.999..1
  returned value is the geometric mean of single color values }
function SImg_GetSharpness3(cc:CSImg_Config; fi:TSImg_FPImage; thr:double; var sharpness:array of double):double;
function SImg_GetSharpness3(cc:CSImg_Config; fi:TSImg_FPImage; thr:double):double;

type TSImg_PSFInfo = record
        xc,yc   : single;       { center positions }
        xm,ym   : word;         { positions of brightest pixel }
        sig     : single;       { RMS radius. Gaussian PSF: exp(-sqr(rad)/(2*sqr(sig)) }
        max     : simgfp;       { maximum }
        sum     : simgfp;       { sum. 1 for normalized PSF }
        zrad    : single;       { smallest radius with a zero pixel }
        end;

{ Get PSF info }
procedure Simg_GetPSFInfo(cc:CSimg_Config; fs:TSImg_FPImage; var inf:TSImg_PSFInfo);
function SImg_PSFInfoString(const inf:TSImg_PSFInfo):ansistring;
//function SImg_PSFInfo_test(cc:CSImg_Config;sig:double):ansistring;

{ Return PSF width at a given brightness level relative to max }
function Simg_GetPSFWidth(cc:CSimg_Config; fs:TSImg_FPImage; l:double):double;


{*************************************************************************}
{******* routines defined in bayer.inc ***********************************}
{*************************************************************************}
type TSImg_BayerPattern = array[0..1,0..1] of longint; { colors for Bayer pattern: 0:red/color0, 1:green/color2, 2:blue/color3, 4:color4; first index is y, second is x }

type TSImg_BayerInfo = record
       pattern  : TSImg_BayerPattern;    { Bayer pattern }
       colors   : longint;               { number of colors }
       cf       : array[0..3] of double; { color factors }
       end;
{ check Bayer pattern and set number of colors
  return true if Bayer pattern is valid }
function SImg_BayerCheckInfo(var bi:TSImg_BayerInfo):boolean;

{ generates a TSImg_BayerPattern from a string
  s : pattern, examples:
      * '1223', 'RGGB' (both are the same)
      * '1243', 'RGWB' (both are the same) }
function SImg_BayerParsePattern(cc:CSImg_Config; s:string7):TSImg_BayerPattern;
{ generates a TSImg_BayerPattern from a string  }
procedure SImg_BayerParsePattern(cc:CSImg_Config; s:string7; var bi: TSImg_BayerInfo);

function SImg_BayerFlipPattern(cc:CSImg_Config; flipx,flipy:boolean; s:string7):string7;

{ color interpolation
  bi     : Bayer pattern information
  fi     : input
  fj     : output
  algo
    < -1 : weight mask
    = -1 : q25
    =  0 : linear
    >  0 : median
  c
    <  0 : interpolate all colors
    >= 0 : interpolate color c only, where 0 denotes 1st color }
procedure SImg_BayerInterpolation(cc:CSImg_Config; bi:TSImg_BayerInfo; fi:TSImg_FPImage; var fj:TSImg_FPImage; algo,c:longint);

{ down scaling
  bi     : bayer pattern information
  fi     : input
  fj     : output }
procedure SImg_BayerDownScale(cc:CSImg_Config; bi:TSImg_BayerInfo; fi:TSImg_FPImage; var fj:TSImg_FPImage);

{ set color factors of a TSImg_BayerInfo such that average of each color will be 1
  bi     : Bayer pattern information
  fi     : input }
procedure SImg_BayerFlat(cc:CSImg_Config; fi:TSImg_FPImage; var bi:TSImg_BayerInfo);

{ apply color factors to Bayer pattern image
  bi     : Bayer pattern information
  fi     : input }
procedure SImg_BayerScale(cc:CSImg_Config; fi:TSImg_FPImage; var bi:TSImg_BayerInfo);

{ convert multi color image to Bayer pattern image }
procedure SImg_ToBayer(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; bi:TSImg_BayerInfo);


{*************************************************************************}
{******* routines defined in ff.inc **************************************}
{*************************************************************************}
{$PACKENUM 4}  { important, must be 4 }
@DEFAULT_PACKRECORDS;

type TSImg_FileBuf=record
       case smallint of
        0 : ( pt          : ^byte;
              size,reserv : qword;
            );
        1 : ( ptb         : ^byte );
        2 : ( ptw         : ^word );
        3 : ( ptc         : ^char );
       end;

type CSImg_Output=class
       info   : string;
       constructor init;
       procedure close(cc:CSImg_Config);
       procedure open(cc:CSImg_Config);
       procedure write(cc:CSImg_Config; pt:pointer; num:int64);
      private
       isopen : boolean;       { set to true by open }
       procedure open_int(cc:CSImg_Config); virtual; abstract;   { called from open }
       procedure close_int(cc:CSImg_Config); virtual; abstract;  { called from close if output is open (isopen=true) }
       procedure write_int(cc:CSImg_Config; pt:pointer; num:int64); virtual; abstract;  { called from write if output is open }
       end;

     CSImg_FileOutput=class(CSImg_Output)
       constructor init(const theFn:ansistring);
      private
       fn  : ansistring;
       fd  : longint;
       procedure open_int(cc:CSImg_Config); override;
       procedure close_int(cc:CSImg_Config); override;
       procedure write_int(cc:CSImg_Config; pt:pointer; num:int64); override;
       end;

{ read an image  (no conversions)
   fn: filename }
procedure SImg_ReadImageFile(cc:CSImg_Config; fn:ansistring; var ri:TSImg_RDImage; mi:CSImg_MetaInfo);
procedure SImg_ReadImageFile(cc:CSImg_Config; fn:ansistring; var ri:TSImg_RDImage);

{ read an image
   fn: filename }
procedure SImg_ReadImageFile(cc:CSImg_Config; fn:ansistring; var fi:TSImg_FPImage; mi:CSImg_MetaInfo);
procedure SImg_ReadImageFile(cc:CSImg_Config; fn:ansistring; var fi:TSImg_FPImage);

procedure SImg_RHeader(cc:CSImg_Config; fn:ansistring; var ri:TSImg_RDImage; mi:CSImg_MetaInfo);                { ri only contains image size }
procedure SImg_RHeader(cc:CSImg_Config; fn:ansistring; mi:CSImg_MetaInfo);

procedure SImg_WriteImageFile(cc:CSImg_Config; const fn:ansistring; ri:TSImg_RDImage; mi:CSImg_MetaInfo);       { write an image (str: filename) }
procedure SImg_WriteImageFile(cc:CSImg_Config; const fn:ansistring; fi:TSImg_FPImage; mi:CSImg_MetaInfo);       { write an image (str: filename) }
procedure SImg_WriteImageFile(cc:CSImg_Config; const fn:ansistring; ri:TSImg_RDImage);                          { write an image (str: filename) }
procedure SImg_WriteImageFile(cc:CSImg_Config; const fn:ansistring; fi:TSImg_FPImage);                          { write an image (str: filename) }


{*************************************************************************}
{******* routines and types defined in ff-pnm.inc ************************}
{*************************************************************************}
function SImg_RPNM(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;
procedure SImg_WPNM(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; mi:CSImg_MetaInfo);


{*************************************************************************}
{******* routines and types defined in ff-sif.inc ************************}
{*************************************************************************}
function  SImg_RSIF(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;
procedure SImg_WSIF(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; mi:CSImg_MetaInfo);


{*************************************************************************}
{******* routines and types defined in ff-tsif.inc ************************}
{*************************************************************************}
const SImg_TSIFHeaderID : array[0..3] of char = 'fIsT';
procedure SImg_RTSIF(cc:CSImg_Config; fd:longint; var ri:TSImg_RDImage; mi:CSImg_MetaInfo; headerOnly:boolean);
procedure SImg_RTSIF(cc:CSImg_Config; fd:longint; var ri:TSImg_RDImage; mi:CSImg_MetaInfo);
procedure SImg_WTSIF(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; mi:CSImg_MetaInfo);


{*************************************************************************}
{******* routines and types defined in ff-fits.inc ***********************}
{*************************************************************************}
function SImg_RFITS(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo; headerOnly:boolean):boolean;
function SImg_RFITS(cc:CSImg_Config; fb:TSImg_FileBuf;var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;
{ write/append function return the output object f without closing it. This is useful for fits files with more than one HDU }
function SImg_WFITS(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; mi:CSImg_MetaInfo):CSImg_Output;
function SImg_WFITSHeader(cc:CSImg_Config; f:CSImg_Output; mi:CSImg_MetaInfo):CSImg_Output;
function SImg_AFITSTable(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; const header:array of string8; mi:CSImg_MetaInfo):CSImg_Output;     { append fits table }


{*************************************************************************}
{******* routines and types defined in ff-sbig.inc ***********************}
{*************************************************************************}
function SImg_RSBIG3(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;


{*************************************************************************}
{******* routines and types from fftw lib ********************************}
{*************************************************************************}
const fftwlib = 'fftw3';
      fftw_timelimit = 0.25;    // time limit for creating a plan in seconds

type fftw_plan = pointer;

//define[FFTW_MEASURE][0]
//define[FFTW_DESTROY_INPUT][(1 shl 0)]
//define[FFTW_UNALIGNED][(1 shl 1)]
//define[FFTW_CONSERVE_MEMORY][(1 shl 2)]
//define[FFTW_EXHAUSTIVE][(1 shl 3)]
//define[FFTW_PRESERVE_INPUT][(1 shl 4)]
//define[FFTW_PATIENT][(1 shl 5)]
//define[FFTW_ESTIMATE][(1 shl 6)]
//define[FFTW_WISDOM_ONLY][(1 1shl 21)]

function fftw_plan_dft_r2c_2d(nx,ny:longint;i:pfftfp;o:pfftfp; flags:dword):fftw_plan; cdecl; external fftwlib name 'fftw_plan_dft_r2c_2d';
function fftw_plan_dft_c2r_2d(nx,ny:longint;i:pfftfp;o:pfftfp; flags:dword):fftw_plan; cdecl; external fftwlib name 'fftw_plan_dft_c2r_2d';
procedure fftw_set_timelimit(seconds:double); cdecl; external fftwlib name 'fftw_set_timelimit';
procedure fftw_destroy_plan(plan:fftw_plan); cdecl; external fftwlib name 'fftw_destroy_plan';
procedure fftw_execute(plan:fftw_plan); cdecl; external fftwlib name 'fftw_execute';
procedure fftw_init_threads; cdecl; external fftwlib name 'fftw_init_threads';
procedure fftw_plan_with_nthreads(nthreads:longint); cdecl; external fftwlib name 'fftw_plan_with_nthreads';
procedure fftw_cleanup; cdecl; external fftwlib name 'fftw_cleanup';
procedure fftw_cleanup_threads; cdecl; external fftwlib name 'fftw_cleanup_threads';
procedure fftw_flops(const p:fftw_plan; add,mul,fmas:pdouble);  cdecl; external fftwlib name 'fftw_flops';

{*************************************************************************}
{******* routines and types defined in fft-sys.inc ***********************}
{*************************************************************************}
const TSImg_FFT_ALIGN = 64;   { FFT data alignment, ready for AVX512 }

type TSImg_FFTImage=record    { image type for Fourier transforms, either storing real data or complex data which is symmetrical on x axis }
        opt      : pointer;   { original pointer }
        pt       : ^fftfp;    { aligned data }
        nxr,nxc  : longint;   { image width in real / complex mode: nxc=nxr div 2+1 }
        ny       : longint;   { image height }
        bytes    : qword;     { amount of allocated memory, must be at least nxc*2*ny*sizeof(fftfp) + TSImg_FFT_ALIGN-1  }
        iid      : longint;   { image id, must be equal to SImg_GetImageTableEntry(self) }
        plan_r2c : fftw_plan; { fftw plan for forward transformation }
        plan_c2r : fftw_plan; { fftw plan for reverse transformation }
        end;
     PSImg_FFTImage=^TSImg_FFTImage;

{ return smallest number larger or equal than 2^n*3^m where m=0..2 }
function SImg_FFT_FindBase(n:longint):longint;

procedure SImg_New(cc:CSImg_Config; var r:TSImg_FFTImage; spt:pointer);  { nx, ny and sl must be set }
procedure SImg_New(cc:CSImg_Config; var r:TSImg_FFTImage);
procedure SImg_New(cc:CSImg_Config; var r:TSImg_FFTImage; nxr,ny:longint);
procedure SImg_New_FindBase(cc:CSImg_Config; var r:TSImg_FFTImage; nx,ny:longint);
procedure SImg_Free(cc:CSImg_Config; var fi:TSImg_FFTImage);
procedure SImg_Move(cc:CSImg_Config; var fi,fj:TSImg_FFTImage);  { fi -> fj }
function SImg_Integrity(fi:TSImg_FFTImage):boolean;
function SImg_OverwriteAlert(fi,fj:TSImg_FFTImage):boolean;
procedure SImg_Duplicate(cc:CSImg_Config; fi:TSImg_FFTImage; var fj:TSImg_FFTImage; dat:boolean);   { fi --> fj; dat=true: copy data }
function SImg_InfoStr(fi:TSImg_FFTImage):string19;
function SImg_InfoStr2(fi:TSImg_FFTImage):string37;

procedure SImg_FFT_R2C(cc:CSImg_Config; fi:TSImg_FFTImage);
procedure SImg_FFT_C2R(cc:CSImg_Config; fi:TSImg_FFTImage);

function SImg_FFT_R2C_flops(cc:CSImg_Config; fi:TSImg_FFTImage):int64;
function SImg_FFT_C2R_flops(cc:CSImg_Config; fi:TSImg_FFTImage):int64;


{*************************************************************************}
{******* routines and types defined in fft-base.inc **********************}
{*************************************************************************}
procedure SImg_Clear(cc:CSImg_Config; fi:TSImg_FFTImage);                                    { fi(*,*):=0; }

{ copy real data }
procedure SImg_CopyR(cc:CSImg_Config; fh:TSImg_FPImage; fi:TSImg_FFTImage; k:longint;xt,yt:longint);    { Copy fh(y-yt,x-xt,k) --> fi(y,x). Rotate. Do not clear undefined pixels. }
procedure SImg_Copy(cc:CSImg_Config; fh:TSImg_FPImage; fi:TSImg_FFTImage; k:longint);                   { Copy fh(y   ,x   ,k) --> fi(y,x). Do not clear undefined pixels. }
procedure SImg_CopyR(cc:CSImg_Config; fi:TSImg_FFTImage; fh:TSImg_FPImage; l:longint;xt,yt:longint);    { Copy fi(y-yt,x-xt) --> fh(y,x,l). Rotate. Do not clear undefined pixels. }
procedure SImg_Copy(cc:CSImg_Config; fi:TSImg_FFTImage; fh:TSImg_FPImage; l:longint);                   { Copy fi(y,x) --> fh(y,x,l). Rotate. Do not clear undefined pixels. }
procedure SImg_CopyC(cc:CSImg_Config; fi:TSImg_FPImage; ti:TSImg_FFTImage; l:longint);                  { Copy fi(y,x) --> fh(y,x,l). Clear undefined pixels. }

{ complex data operations }
procedure SImg_Konj_C(cc:CSImg_Config; fi:TSImg_FFTImage);
procedure SImg_Mul_C(cc:CSImg_Config; fi,fj,fk:TSImg_FFTImage);         { fi*fj --> fk, images may be equal }
procedure SImg_MulK_C(cc:CSImg_Config; fi,fj,fk:TSImg_FFTImage);        {  fi*fj~ --> fk }
procedure SImg_Inv_C(cc:CSImg_Config; fi:TSImg_FFTImage);               { 1/fi --> fi }
procedure SImg_DIV_C(cc:CSImg_Config; fi,fj,fk:TSImg_FFTImage);         { fi/fj --> fk. fi,fj,fk can be the same }
procedure SImg_Exp_C(cc:CSImg_Config; fi:TSImg_FFTImage;n:double);

{ Write real FFT image to file. Not very efficient and should only be used for debugging.
   fn    : filename
   ti    : real image }
procedure SImg_FFTWrite(cc:CSImg_Config; fn:ansistring; ti:TSImg_FFTImage);

{ fj = fi \ast fs
   fi   : input
   fs   : PSF
   fj   : output (must be allocated)
   xc,yc: PSF center coordinates
}
procedure SImg_Convolute(cc:CSImg_Config; fi,fs,fj:TSImg_FPImage; xc,yc: longint);


{*************************************************************************}
{******* routines and types defined in fourier.inc ***********************}
{*************************************************************************}
{ Benchmark:
   img   : test image
   pad   : allow padding
   it    : iterations
   pr    : kpixel/s
   flops : KFLOP/s
   t     : time spent }
procedure SImg_FFTTest(cc:CSImg_Config; img:TSImg_FPImage; pad:boolean; it:longint; var pr,flops,t:double);

{ compute Fourier transform of an image and output the spectrum
    fi   : Source
    fj   : Destination
    cent : True: (0,0) in center
           False: (0,0) top left
   pad   : allow padding}
procedure SImg_Spectrum(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; cent:boolean; pad:boolean);


{*************************************************************************}
{******* routines defined in histogr.inc *********************************}
{*************************************************************************}
{ get the n-th brightest/darkest value
    n         : n
    c         : color, <0 for all
    minv,maxv : results }
procedure SImg_GetMinMaxN(cc:CSImg_Config; fi:TSImg_FPImage; n,c:longint; var minv,maxv:double);

{ get quantiles
    tn       : number of threads
    c        : color, <0 for all
    ps       : percentages
    qs       : quantiles (result, must be allocated and can be the same array as ps)
}
procedure SImg_GetQuantile(cc:CSImg_Config; fi:TSImg_FPImage; tn,c:longint; ps:array of double; var qs:array of double);
function SImg_GetQuantile(cc:CSImg_Config; fi:TSImg_FPImage; q:double):double;
function SImg_GetQuantile(cc:CSImg_Config; fi:TSImg_FPImage; c:longint; q:double):double;

{ crop and scale brightness with linear tonal curve
  lq,uq    : lower/upper quantile
  cut      : cut values
  st       : stretch to [0,1]
  la,ua    : interpret lq/uq as absolute value (not as quantile) }
procedure SImg_HistCut(cc:CSImg_Config; fi:TSImg_FPImage; lq,uq:double; cut,st:boolean; la,ua:boolean);

{ crop and scale brightness with non-linear tonal curve
  lq,mq,uq : lower/mid/upper quantile
  mv       : value for mid quantile
  sp       : saturation preservation (0..1)
  st       : stretch to [0,1]
  la,ma,ua : interpret lq/mq/uq as absolute value (not as quantile) }
procedure SImg_HistCut(cc:CSImg_Config; fi:TSImg_FPImage; lq,mq,uq,mv,sp:double; st:boolean; la,ma,ua:boolean);

{ crop and scale brightness with non-linear tonal curve
  lq,mq,uq : lower/mid/upper quantile
  mv       : value for mid quantile
  sp       : saturation preservation (0..1)
  st       : stretch to [0,1]
  la,ma,ua : interpret lq/mq/uq as absolute value (not as quantile) }
procedure SImg_HistCut(cc:CSImg_Config; fi:TSImg_FPImage; lq,mq,uq,mv, gam, sp:double; st:boolean; la,ma,ua:boolean);

{ crop and scale brightness with regularized non-linear tonal curve
  lq,uq    : lower/mid/upper quantile
  gam,bet  : gamma / beta
  mv       : value for mid quantile
  sp       : saturation preservation (0..1)
  st       : stretch to [0,1]
  sc       : process colors separately
  la,ua    : interpret lq/uq as absolute value (not as quantile) }
procedure SImg_HistCutGamma(cc:CSImg_Config; fi:TSImg_FPImage; lq,uq, gam,bet,sp:double; st:boolean; la,ua:boolean);

{ crop and scale brightness with regularized logarithmic tonal curve
  lq,uq : lower/upper quantile
  reg   : regularization
  ex    : saturation modification exponent
  mq,mv : brightness adjustment of the form 1-(1-y)^b using mid quantile/value. Ignored if one the values if not within the range of 0 to 1.
}
procedure SImg_HistCutLog(cc:CSImg_Config; fi:TSImg_FPImage; lq,uq, reg,ex, mq,mv:double);

{ Logarithmically scale brightness of an image
  lq    : lower quantile
  reg   : regularization
}
procedure SImg_Log(cc:CSImg_Config; fi:TSImg_FPImage; lq,reg:double);

{ crop and scale brightness with regularized non-linear tonal curve
  lq,uq : lower/upper quantile
  reg   : regularization. Valid: >0
  ex    : exponent for the base tonal curve (x+reg)^ex
  cex   : saturation modification exponent. Valid: >0
  mq,mv : brightness adjustment of the form 1-(1-y)^b using mid quantile/value. Ignored if one the values if not within the range of 0 to 1.
}
procedure SImg_HistCutPow(cc:CSImg_Config; fi:TSImg_FPImage; lq,uq, reg,ex,cex, mq,mv:double);


{*************************************************************************}
{******* routines and types defined in is-sys.inc ************************}
{*************************************************************************}
type TSImg_AffineCoeff_vect = array[0..1] of double;
     TSImg_AffineCoeff_mat  = array[0..1,0..1] of double;
     TSImg_AffineCoeff=record                { (ii[x]*fact + offs = ri[Mx+s] }
       ni        : longint;                  { number of entries }
       m         : ^TSImg_AffineCoeff_mat;   { matrix }
       s         : ^TSImg_AffineCoeff_vect;  { shift }
       fact,offs : ^double;                  { factor, offset }
       mse       : ^double;                  { mean square error }
       end;

type CSImg_ImageSet = class
          getmax          : longint;
          cnx,cny,cnc     : boolean;  { compare nx/ny/nc }
          constructor init; virtual;
          destructor done; virtual;
          procedure getimage(cc:CSImg_Config; var img:TSImg_FPImage; i:longint);
          procedure getimage(cc:CSImg_Config; var img:TSImg_FPImage; mi:CSImg_MetaInfo; i:longint);
          function info:shortstring;virtual;
          procedure getMI0(cc:CSImg_Config; mi:CSImg_MetaInfo);         // get meta information of first image
        private
          nx,ny,nc        : longint;
          img0            : TSImg_FPImage;      // buffer of image 0, often read to obtain image info
          mi0             : CSImg_MetaInfo;     // meta info of image 0
          procedure getimage_int(cc:CSImg_Config; var fi:TSImg_FPImage; mi:CSImg_MetaInfo; i:longint); virtual; abstract;
          end;

type CSImg_ImageSet_Files = class(CSImg_ImageSet)
          constructor init(cc:CSImg_Config; const pattern:ansistring);overload;virtual;
          destructor done;override;
          function info:shortstring;override;
        private
          files          : pansistring;
          files_reserved : longint;
          procedure findfiles(const path,pattern:ansistring);
          procedure addfile(const fn:ansistring);
          procedure getimage_int(cc:CSImg_Config; var fi:TSImg_FPImage; mi:CSImg_MetaInfo; i:longint); override;
          end;

function SImg_NumToSetFile(const s:ansistring;n:longint):ansistring;
function SImg_LastFileSetNum(const s:ansistring):longint;


{*************************************************************************}
{******* routines and types defined in is-base0.inc **********************}
{*************************************************************************}
{ Calculate average and noise of an image set
   iset: image set
   fi  : avg output
   fn  : square noise output
   d   : drop brightest and darkest value
   dmin: minimum valid value
   dmax: maximum valid value }
procedure SImg_AvgSet(cc:CSImg_Config; iset:CSImg_ImageSet; var ofi,ofn:TSImg_FPImage; d:boolean; dmin,dmax:double);

{ get (read out) noise from an (bias) image set }
function SImg_GetNoise(cc:CSImg_Config; iset:CSImg_ImageSet): double;
{ get gain from an image set }
function SImg_GetGain(cc:CSImg_Config; iset:CSImg_ImageSet; rdnoise:double): double;

{ Find the the darkest n=1..8 values of each pixels an return their average }
procedure SImg_MinSet(cc:CSImg_Config; iset:CSImg_ImageSet; var fj:TSImg_FPImage; n:longint);

{ Find the the darkest n=1..8 values of each pixels an return their average
  db : drop brightest value }
procedure SImg_MaxSet(cc:CSImg_Config; iset:CSImg_ImageSet; var fj:TSImg_FPImage; n:longint; db:boolean);


{*************************************************************************}
{******* routines and types defined in raw.inc ***************************}
{*************************************************************************}
type CSImg_GetRaw=class            { abstract }
       bufnum  : longint;          { number of buffers }
       bufs    : ^TSImg_RDImage;   { bufnum buffers where the data are stored, created by init }
       constructor init(cc:CSImg_Config; bn, nx,ny,nc:longint; fmt:TSImg_RawFormat); virtual;  { creates bn(=:bufnum) buffers, ... }
       procedure close(cc:CSImg_Config); virtual;   { destroy buffers, calls destroy ... }
       procedure ReadStart(cc:CSImg_Config; i:longint); virtual; abstract;  { start reading data for buffer buffer i }
       procedure ReadEnd(cc:CSImg_Config; i:longint); virtual; abstract;    { waits until data is read and written to buffer buffer i }
       end;

     CSImg_FileRead=class(CSImg_GetRaw)
//       ExposureExt : ^CSImg_Ext_Exposure; { exposure information }
       constructor init(cc:CSImg_Config; const fn:shortstring; bn, nx,ny,nc:longint; fmt:TSImg_RawFormat; cct:boolean);overload;  { fn: file name; ct: create capture thread }
       procedure close(cc:CSImg_Config); override;
       procedure ReadStart(cc:CSImg_Config; i:longint); override;
       procedure ReadEnd(cc:CSImg_Config; i:longint); override;
      protected
       capt_id   : ^dword;  { $ffffffff: capturing finished; buffer with smallest capt_id is captured first}
       fd        : longint;
      private
       ct        : CSImg_Thread;
       capt_cnt  : longint;
       end;

procedure SImg_RawSave(cc:CSImg_Config; data:CSImg_GetRaw; const fo:shortstring; n:longint);   { number of images to save }


{*************************************************************************}
{******* routines and types defined in sources.inc ***********************}
{*************************************************************************}
{ get position and brightness of sources
    fi   : input image
    fw   : weight of input image (1/sqr(noise))
    pl   : source data array
    num  : maximum number of sources
    cmp  : compress, i.e. do no save chains }
procedure SImg_GetSources(cc:CSImg_Config; fi,fw:TSImg_FPImage; var pl:TSImg_RDImage; num:longint; cmp:boolean);

{ plot position and brightness of sources
    fo  : output
    pl  : source data array
    mi  : keys image boundaries IBND* must be set
    rad : half brightness radius
}
procedure SImg_PlotSources(cc:CSImg_Config; var fo:TSImg_FPImage; pl:TSImg_RDImage; mi:CSImg_MetaInfo; rad:double);

{ extract sources from ppm database
    dbd   : database directory
    year  : year for proper motion calculation
    ra,dec,width,height  : extraction area, all parameters in degree
    rot   : rotation of extraction area, cw, 0 mean north is up
    scale : pixel scale in arcsec
    proj  : projection type: -1: gnomonic; 0: stereographic, 1: orthographic
    n     : maximum number of sources
    pl    : source data array
    mi    : meta data (optional)
}
procedure SImg_PPMSources(cc:CSImg_Config; const dbd:ansistring; year:longint; ra,dec,width,height,rot,scale, proj:double; n:longint; var pl:TSImg_RDImage; mi:CSImg_MetaInfo);


{*************************************************************************}
{******* routines and types defined in alignment.inc *********************}
{*************************************************************************}
type TSImg_AlignmentParams = record { alignment parameters, describe how a reference must be transformed to fit to destination }
        { alignment parameters }
        ax,ay               : double;    { shift }
        b11,b12,b21,b22     : double;    { rotation, scaling matrix }
        c1,c2,c3,c4,c5,c6   : double;    { 1st order distortion factors }
        d                   : double;    { 2nd order distortion factors }
        { boundaries }
        xmin,xmax,ymin,ymax : double;    { boundaries of destination image }
        { factor }
        fact                : double;    { brightness scaling factor: how must destination be scaled to fit to reference }
        weight              : double;    { weight, usually be 1 }
        { informal }
        j1,j2,j3            : double;    { correlation at SImg_sources_mindist, SImg_sources_mindist/2 and SImg_sources_mindist/4 }
        srad                : double;    { average square radius of reference source list (used internally for preconditioning) }
        end;
     PSImg_AlignmentParams = ^TSImg_AlignmentParams;

     TSImg_AlignmentConstraints = record { alignment constraints }
        rx,ry : double;   { This position of the reference image must be transformed to a position in the destination image ... }
        dmax  : double;   { ... which is within this radius from to origin (coordinates (0,0), usually the center) }
        nq    : boolean;  { allow non-quadratic pixels }
        flip  : boolean;  { allow flipping }
        smin  : double;   { minimum scale }
        smax  : double;   { maximum scale }
        sh    : double;   { shearing in degrees }
        frc   : double;   { field rotation in degrees }
        frd   : double;   { field rotation difference in degrees. Field rotation fr must satisfy abs(normalize(fr-frc))<=frd }
        end;
     PSImg_AlignmentConstraints = ^TSImg_AlignmentConstraints;

{ return true if alignment parameters are set }
function SImg_ReadAlignmentParams(cc:CSImg_Config; var res:TSImg_AlignmentParams; mi:CSImg_MetaInfo):boolean;

procedure SImg_WriteAlignmentParams(const res:TSImg_AlignmentParams; mi:CSImg_MetaInfo);

{ tries to calculate constraints using wcs data
  rmi : meta info from reference
  dmi : meta info from destination
  tol : tolerance (typically 0.05 to 0.2)
  result: non-empty string on failure containing the reason }
function SImg_GetAlignmentConstraints(cc:CSImg_Config; rmi,dmi:CSImg_MetaInfo; tol:double; var con:TSImg_AlignmentConstraints):ansistring;

{ calculates how reference must be aligned to fit to destination
    pref0 : reference sources
    pdst0 : destination
    dl    : estimate distortion level
    res   : result
    con   : constraints for initial solution }
procedure SImg_GetAlignment(cc:CSImg_Config; pref0,pdst0:TSImg_RDImage; dl:longint; var res:TSImg_AlignmentParams; con:TSImg_AlignmentConstraints);

{ Aligns a bright diffuse source to center
    fi  : input image
    rad : approximate radius of source
    res : result
    cx  : constraint: x position
    cy  : constraint: y position
    cr  : constraint: radius }
procedure SImg_GetAlignment2(cc:CSImg_Config; fi:TSImg_FPImage; rad:double; var res:TSImg_AlignmentParams; cx,cy,cr:integer);


{*************************************************************************}
{******* routines and types defined in stackdb.inc ***********************}
{*************************************************************************}
type TSImg_SDBInfo = record { stack database information }
        { transformation of star database coordinates (x,y) into reference coordinates: ( s11*(x-x0)+s12(y-y0) , s21*(x-x0)+s22(y-y0) }
        x0,y0            : single;      { origin pixel in the output database }
        s11,s12,s21,s22  : single;      { scaling, rotation and flipping matrix}
        xs,ys            : longint;     { size of the sub-grid }
        xso,yso          : longint;     { offset of the sub-grid (non-positive)}
        end;
     PSImg_SDBInfo = ^TSImg_SDBInfo;

procedure SImg_StackDB_InitInfo(var si:TSImg_SDBInfo; scale,rot:single; flip:boolean);                          { Calculate the transformation matrix }
procedure SImg_StackDB_GetInfo(cc:CSImg_Config; var si:TSImg_SDBInfo; mi:CSImg_MetaInfo; nx,ny:longint);        { nx,ny : database dimensions }
procedure SImg_StackDB_WriteInfo(const si:TSImg_SDBInfo; mi:CSImg_MetaInfo);

{ add image to stack database
  si    : stack database info
  fs    : stack database
  fi    : image to add
  fw    : weight image, expected to be 1/sqr(sigma)
  ffl   : FWHM limits
  ffm   : FWHM map
  res   : alignment parameters }
procedure SImg_StackDB_add(cc:CSImg_Config; si:TSImg_SDBInfo; fs,fi,fw,ffl,ffm:TSImg_FPImage; res:TSImg_AlignmentParams);

{ shift reference image to destination
  si    : stack database info
  fr    : reference image
  fi    : Destination image. Only defined pixels are written.
  fw    : Weight image. Optional. If given, undefined pixels are masked out.
  res   : alignment parameters }
function SImg_StackDB_shift(cc:CSImg_Config; si:TSImg_SDBInfo; fr,fi,fw:TSImg_FPImage; res:TSImg_AlignmentParams):boolean;

{ correct value biases between FWHM range }
procedure SImg_StackDB_Correct(cc:CSImg_Config; fs:TSImg_FPImage);

{ convert resolution optimized values and weight
  fi : output image
  fw : weight = 1/sqr(sigma)
  snr: maximum SNR, if <1 interpreted as quantile }
procedure SImg_StackDB_Convert(cc:CSImg_Config; fs:TSImg_FPImage; var fi,fw:TSImg_FPImage; snr:double);

{ convert weighted average
  sc : separate colors }
procedure SImg_StackDB_ConvertVal(cc:CSImg_Config; fs:TSImg_FPImage; var fi:TSImg_FPImage; sc:boolean);

{ convert weighted averages and weight, do not interpolate }
procedure SImg_StackDB_ConvertValWeight(cc:CSImg_Config; fs:TSImg_FPImage; var fi,fw:TSImg_FPImage);

{ convert SNR
  sc : separate colors }
procedure SImg_StackDB_ConvertSNR(cc:CSImg_Config; fs:TSImg_FPImage; var fi:TSImg_FPImage; sc:boolean);

{ convert weight
  sc : separate colors }
procedure SImg_StackDB_ConvertWeight(cc:CSImg_Config; fs:TSImg_FPImage; var fi:TSImg_FPImage; sc:boolean);

{ Calculate a crop region from weight
    thr : threshold. Recommended: 0.1..0.9 }
procedure SImg_StackDB_CropInfo(cc:CSImg_Config; fs:TSImg_FPImage; thr:double; var x1,y1,x2,y2:longint);

{ Calculate FWHM ranges
   si    : stack database info
   fns   : FWHM input files (number of colors must be 1, size can vary)
   res   : Align parameters
   len   : Number of files
   n     : number of ranges
   fj    : output image, will contain n center FWHM and n-1 limits (order: center FWHM, limit, ...) }
procedure SImg_StackDB_FWHMLimits(cc:CSImg_Config; si:TSImg_SDBInfo; const fns:array of ansistring; const res:array of TSImg_AlignmentParams; var fj:TSImg_FPImage; n:longint);

{ Underweight image boundaries
    fw: input and output
    wb: underweight size relative to half image size (0..1) }
procedure SImg_UnderweightBoundaries(cc:CSImg_Config; fw:TSImg_FPImage; wb:double);

{ Unmask unlikely values.
    fi:  background corrected image (input and output)
    fr:  residual input
    fsn: square noise
    fw:  weight (input and output)
    dt:  unmask threshold relative to variance
    dr:  absolute value is unmask threshold relative to reference. If <0 lower threshold is sqrt(1+sqr(dl)).
    adr: ADC resolution in ADU/e }
function SImg_StackDB_unmask(cc:CSImg_Config; fi,fr,fsn,fw:TSImg_FPImage; dt,dr,adr:double):double;

{ Calibrate (destination) image factor using source list of reference
  fd    : destination image
  fw    : pixel weight for destination image
  sl    : source list
  res   : alignment parameters
  fwhm  : star fwhm in order to determine calculation radius
  bbw   : if <>nil it brightness*weight, background*weight and weight are stored to this buffer
  result: calibration factor or -1 if it can't be determined  }
function SImg_StackDB_calib(cc:CSImg_Config; fd,fw:TSImg_FPImage; sl:TSImg_RDImage; res:TSImg_AlignmentParams; fwhm:double; bbw:pdouble):double;

{*************************************************************************}
{******* routines and types defined in colors.inc ************************}
{*************************************************************************}
{ dm    : differential mode
  frad  : smoothing radius
  returns an info string }
function SImg_ColorAdjustInfo(cc:CSImg_Config; fj:TSImg_FPImage; dm:boolean; frad:double):ansistring;

{ dm    : differential mode
  frad  : smoothing radius
  returns an info string }
function SImg_ColorAdjust(cc:CSImg_Config; fi:TSImg_FPImage; dm:boolean; frad:double):ansistring;

procedure SImg_Saturation01(cc:CSImg_Config; fi:TSImg_FPImage;var fj:TSImg_FPImage; sat,maxsat:double);
procedure SImg_Saturation02(cc:CSImg_Config; fi:TSImg_FPImage;var fj:TSImg_FPImage; sat:double);
procedure SImg_SaturationCompensation(cc:CSImg_Config; fi:TSImg_FPImage);

{ Color composition using luminance image
  fi        : input colors images
  fo        : output image }
procedure SImg_Compose(cc:CSImg_Config; fi:array of TSImg_FPImage; var fo:TSImg_FPImage);

{ Color composition using luminance image
  fi        : input colors images
  fj        : input luminance image
  fo        : output image
  fwhm      : FWHM's of input images
  cfwhm     : minimum color FWHM
  result    : color FWHM }
function SImg_LCompose(cc:CSImg_Config; fi:array of TSImg_FPImage; fj:TSImg_FPImage; var fo:TSImg_FPImage; fwhm:array of double; cfwhm:double):double;

{ Extrapolate undefined regions at boundary by using other colors.
  fi    : input and output images
  wi    : weight images
  fo    : output images. Must be Allocated and can be equal to fi or wi.
  b     : boundary width in pixel
  reg   : regularization factor }
procedure SImg_RegBounds(cc:CSImg_Config; fi,wi:array of TSImg_FPImage; b:longint; reg:double);

{*************************************************************************}
{******* routines and types defined in colorize.inc **********************}
{*************************************************************************}
const TSImg_ColorMap_size = 256;
type TSImg_ColorMap=record
        l,u     : simgfp;                                       { limits }
        m       : array[0..TSImg_ColorMap_size,0..3] of simgfp; { map: reference luminance + RGB }
        end;

{ Get color field (1 color image with values between 0 and 1 which is mapped to a color) from ratio of two images
  fu    : uncompressed image. Black level is adjusted to black level of fc.
  fc    : compressed image
  fd    : de-saturation image (optional, if given it acts as input and output)
  fo    : output image: 0 if fc<0 or fu<fc, otherwise 1-fc/fu }
procedure SImg_ColorFieldRatio(cc:CSImg_Config; fu,fc,fd:TSImg_FPImage; var fo:TSImg_FPImage);

{  calculate transformed color map from base color map
    fi  : brightness image
    fc  : color field
    bmap: base color map: n+1 rows (n>=1) consisting in 4 values: 1 position, 3 color values (RGB).
    ml  : quantile of color field where color map starts. If not within range of 0..1, fixed value 0 is used
    mu  : quantile of color field where color map ends. If not within range of 0..1, fixed value 1 is used
    mi  : Invert color map
    me  : Color map histogram equalization. Valid: 0..1
    mb  : Color map histogram balance. <0 overweight smaller values, >0 overweight larger values. No effect if me=0
    ms  : Brightness calibration using saturation reduction. Valid: 0..1, Feasible: 0..0.5
    map : output map }
procedure SImg_TransformColorMap(cc:CSImg_Config; fi,fc:TSImg_FPImage; const bmap:array of simgfp; ml,mu:double; mi:boolean; me,mb,ms:double; var map:TSImg_ColorMap);

{ Colorize 1-color image
  fi: input image used for brightness
  fc: input image used for color, clipped to range: map.l .. map.u
  fd: color de-saturation image, range: 0 (no de-saturation) .. 1 (white)
  fo: output image
  map: color map
  sat: clip to brightest value and correct brightness by de-saturation }
procedure SImg_Colorize(cc:CSImg_Config; fi,fc,fd:TSImg_FPImage; var fo:TSImg_FPImage; const map:TSImg_ColorMap; clip:boolean);
procedure SImg_Colorize(cc:CSImg_Config; fi,fc:TSImg_FPImage; var fo:TSImg_FPImage; const map:TSImg_ColorMap; clip:boolean);

{ Test image
  fo    : output image
  map   : color map
  sat   : saturate to brightest value
  lx,ly : size }
procedure SImg_ColorizeTest(cc:CSImg_Config; var fi:TSImg_FPImage; const map:TSImg_ColorMap; lx,ly:longint);

{ Test image
  fo    : output image
  map   : color map
  sat   : saturate to brightest value
  lx,ly : size
  nl    : number of labels (<2 disables labels)
  lm    : Label mode: >0; number of valid digits, 0: percent, <0 relative to maximum in percent }
procedure SImg_ColorizeLegend(cc:CSImg_Config; var fi:TSImg_FPImage; const map:TSImg_ColorMap; lx,ly,nl,lm:longint);


{*************************************************************************}
{******* routines and types defined in resize.inc ************************}
{*************************************************************************}
procedure SImg_Resize(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; nx,ny:longint);

{ fi    : input
  fo    : output
  sx,sy : scaling factors, >1: down-scaling
  filt  : 0..1: quantile; <0 mean; >1 RMS }
procedure SImg_Resize(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; sx,sy:double; filt:double);

{*************************************************************************}
{******* routines and types defined in approx.inc ************************}
{*************************************************************************}
{ background approximation
    rad    : approximation radii (geometric average of x and y values)
    xyr    : x/y ratio of approximation size.
    cs     : relative approximation size in corner (feasible: 0..1)
    sf     : subtraction factors. If unset and length of rad is 1, the approximation is returned. Otherwise length must be equal to length of rad.
    frad   : radius for quantile pre-filter (<=0 to disable this feature)
    fquant : quantile for pre-filter (0.5 for median)
    fg     : approximate foreground instead of background
}
procedure SImg_FBGApprox(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; mask:TSImg_RDImage; rad:array of double; xyr,cs:double; sf:array of double; frad,fquant:double; fg:boolean);

{ constant background approximation
    res    : result buffer for at least fi.nc values }
procedure SImg_BgApprox2(cc:CSImg_Config; fi:TSImg_FPImage; res:psimgfp);
procedure SImg_BgApprox2(cc:CSImg_Config; fi:TSImg_FPImage);

{ approximate an image by polynomials
    rr   : Factor for radial polynomial
    r1   : Factor for 1st order polynomial (x and y)
    r1b  : Factor for 1st order bi-polynomial (x*y)
    n    : highest order of radial polynomials ((x-xc)^2+(y-yc)^2)^i
    it   : background approximation iterations (cap values to previous approximation)
    dm   : false: approximation replaces input
           true: difference replaces input
    frad  : radius median pre-filter (<=0 to disable this feature)
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
    p_0 to p_2 are orthogonal to all other polynomials }
procedure SImg_Approx(cc:CSImg_Config; fj:TSImg_FPImage; rr,r1,r1b:double; n,it:longint; dm:boolean; frad:double);

{ approximate an image with weight by polynomials
    fi   : input image
    fw   : weight
    pn   : number of polynomials (recommended: 1,3,4,5,..)
    dm   : false: approximation replaces input
           true: difference replaces input
   polynomials:
    p_0(x,y)=1
    p_1(x,y)=x
    p_2(x,y)=y
    p_3(x,y)=x*y
    p_3+i=(x^2+y^2)^i }
procedure SImg_Approx2(cc:CSImg_Config; fi,fw:TSImg_FPImage; pn:longint; dm:boolean);

{ Background subtraction using polynomial approximation
    fi: input and output
    fw: pixel weight
    bg: Approximation method:
        1: constant
        2: gradient
        3: curved
        4+: 1+ radial polynomials }
procedure SImg_BGApprox3(cc:CSImg_Config; fi,fw:TSImg_FPImage; bg:longint);

{ Background approximation by image and constant
    fi: input and output
    fa: approximation image
    fw: pixel weight
    result: image factor
}
function SImg_BGApprox4(cc:CSImg_Config; fi,fa,fw:TSImg_FPImage):double;

{ Approximate fi by fa and polynomials and return the difference.
    fi:  Input image and background corrected output (1 color)
    fa:  Approximation image (1 colors)
    fw:  Weight for approximation image
    fo1: Residual output
    fo2: Background corrected output
    bg:  polynomial approximation method:
         0: none
         1: constant
         2: gradient
         3: curved
         4+: 1+ radial polynomials
    mosaic: boundary overweight factor, >=0
    bc:  brighnes calibration, calibration factor is returned
    result: calibration factor (1 if bc=false) }
function SImg_ImgApprox(cc:CSImg_Config; fi,fa,fw:TSImg_FPImage; var fo1,fo2:TSImg_FPImage; bg:longint; mosaic:double; bc:boolean):double;

{ Approximate fi by polynomials. Border will be overweight (mosaic mode)
    fi: Input and output
    fw: Weight
    bg: polynomial approximation method:
         0: none
         1: constant
         2: gradient
         3: curved
         4+: 1+ radial polynomials
    dm: false: output approximation
        true: output difference
    pc: true if input is already pre-corrected
    mosaic: boundary overweight factor, >0 }
procedure SImg_ApproxMosaic(cc:CSImg_Config; fi,fw:TSImg_FPImage; bg:longint; mosaic:double; dm,pc:boolean);

{ Approximate fi by fa and a constant. Return the factor for fa.
    fi:  Input image and background corrected output (1 color)
    fa:  Approximation image (1 colors)
    fw:  Weight for approximation image
    result: factor for fa
}
function SImg_ImgApprox2(cc:CSImg_Config; fi,fa,fw:TSImg_FPImage):double;

{ Calculate image factor
    fi: input
    fr: reference image
    fw: pixel weight
    bg: Approximation method:
        1: constant
        2: gradient
        3: curved
        4+: 1+ radial polynomials
   oit: object filter iterations. More iterations overweight bright objects/stars. Valid: >= 0. Recommended: 1..3 }
function SImg_ImgFactor(cc:CSImg_Config; fi,fr,fw:TSImg_FPImage; bg:longint):double;

{ constant background approximation. Pixels with a weight less than thr*<weighted average weight> are set to weighted average.
    c      : color
    rad    : median filter radius
    result : result buffer for at least fi.nc values }
function SImg_BgApprox5(cc:CSImg_Config; fi,wi:TSImg_FPImage; c:longint; rad:single):double;


{*************************************************************************}
{******* routines and types defined in hdr.inc ***************************}
{*************************************************************************}
{ Dynamic range compression using non-linear high-pass filter and regularization method 1.
    fi    : input and output image
    mask  : mask, optional, 1-color image
    lq    : black level quantile
    mq    : 25% reduction quantile
    ss    : sizes of details to emphasize
    sp    : saturation preservation (0..1) }
procedure SImg_DRCompress1(cc:CSImg_Config; fi:TSImg_FPImage; mask:TSImg_RDImage; lq,mq:double; ss:array of double; sp:double);

{ Dynamic range compression using non-linear high-pass filter and regularization method 2
    fi    : input and output image
    mask  : mask, optional, 1-color image
    lq    : black level quantile
    hq    : high quantile at which background will be reduced by 50%
    rmin  : Reduction factor limit. 0..1
    ss    : smallest details size to emphasize
    sp    : saturation preservation (0..1) }
procedure SImg_DRCompress2(cc:CSImg_Config; fi:TSImg_FPImage; mask:TSImg_RDImage; lq,hq, rmin, ss, sp:double);

{Dynamic range compression using non-linear high-pass filter and regularization method 3
    fi    : input and output image
    mask  : mask, optional, 1-color image
    lq    : black level quantile or <0 for no black level adjustment. Recommended: 0..0.25
    smin  : smallest detail filter (as radius). Should be larger than supper-resolution factor. At least 1.
    smed  : used median filter below that size
    smax  : largest detail filter (as radius). At least smin*2. If less, it is set to 0.5*min(fi.nx,fi.ny).
    ravg  : average dynamic range compression
    reg   : regularization relative to noise
    sig   : noise floor. Auto-detect if input is <=0
    sp    : saturation preservation (0..1) }
procedure SImg_DRCompress3(cc:CSImg_Config; var fi:TSImg_FPImage; mask:TSImg_RDImage; lq, smin,smed,smax,ravg, reg:double; var sig:double; sp:double);

{ Colorize image using different dynamic range compression for different colors
    fi    : input image
    fo    : output image
    mask  : mask, optional, 1-color image
    lq    : black level quantile (ignored if <0)
    reg   : reduction factor normalization relative to noise
    sa    : Detail filters (as radius). One per color. Last color last colors comes first. Details smaller than first element are added to all colors with factor 1. First element should be at least two times the super-resolution factor.
    ra    : Compression factor per color. Last color colors comes first.
    result: noise floor }
function SImg_DRCompose(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; mask:TSImg_RDImage; lq,reg:double; sa,ra:array of double):double;


{*************************************************************************}
{******* routines and types defined in tiles.inc *************************}
{*************************************************************************}
type TSImg_TileInfo = record
        nx,ny   : longint;      { number of tiles in x any y direction }
        bb      : longint;      { additional border used for blending (total width of transition zone is 1+2*bi }
        bu      : longint;      { additional border which is ignored (only used for processing) }
      end;

{ copy tile into mosaic and blend borders
    ti          : tile info
    ft          : tile image
    fi          : destination (mosaic) image
    tx,tx       : tile position }
procedure SImg_BlendTile(cc:CSImg_Config; ti:TSImg_TileInfo; ft,fi:TSImg_FPImage; tx,ty:longint);

{ copy tile into mosaic and do not blend borders
    ti          : tile info
    ft          : file image
    fi          : destination (mosaic) image
    tx,tx       : tile position }
procedure SImg_CopyTile(cc:CSImg_Config; ti:TSImg_TileInfo; rt,ri:TSImg_RDImage; tx,ty:longint);
procedure SImg_CopyTile(cc:CSImg_Config; ti:TSImg_TileInfo; ft,fi:TSImg_FPImage; tx,ty:longint);

{ Extract tile from mosaic
    ti          : tile info
    ft          : file image
    fi          : destination (mosaic) image
    tx,tx       : tile position }
procedure SImg_GetTile(cc:CSImg_Config; ti:TSImg_TileInfo; ri:TSImg_RDImage; var rt:TSImg_RDImage; tx,ty:longint);
procedure SImg_GetTile(cc:CSImg_Config; ti:TSImg_TileInfo; fi:TSImg_FPImage; var ft:TSImg_FPImage; tx,ty:longint);


{*************************************************************************}
{******* routines and types defined in draw01.inc ************************}
{*************************************************************************}
type TSImg_WriteProc = procedure(cc:CSImg_Config; fi:TSImg_FPImage; x,y:longint; const s:shortstring);
     TSImg_Font = record
       xs,ys  : longint;
       write  : TSImg_WriteProc;
       end;

procedure SImg_Write12x22(cc:CSImg_Config; fi:TSImg_FPImage; x,y:longint; const s:shortstring);
procedure SImg_Write8x16(cc:CSImg_Config; fi:TSImg_FPImage; x,y:longint; const s:shortstring);
procedure SImg_Write8x8(cc:CSImg_Config; fi:TSImg_FPImage; x,y:longint; const s:shortstring);
procedure SImg_Line(cc:CSImg_Config; fi:TSImg_FPImage; x1,y1,x2,y2:longint);
procedure SImg_Bar(cc:CSImg_Config; fi:TSImg_FPImage; x1,y1,x2,y2:longint);
procedure SImg_Rectangle(cc:CSImg_Config; fi:TSImg_FPImage; x1,y1,x2,y2:longint);

const SImg_Font12x22 : TSImg_Font =
        (xs:12; ys:22; write:@SImg_Write12x22);
const SImg_Font8x16 : TSImg_Font =
        (xs:8; ys:16; write:@SImg_Write8x16);
const SImg_Font8x8 : TSImg_Font =
        (xs:8; ys:8; write:@SImg_Write8x8);


{*************************************************************************}
{******* routines and types defined in draw02.inc ************************}
{*************************************************************************}
{ Generate rectangular distributed noise on interval [-0.5,0.5]
   nx,ny,nc  : dimensions }
procedure SImg_Noise_Rect(cc:CSImg_Config; var fi: TSImg_FPImage; nx,ny,nc:longint);

{ Generate normal distributed noise (sigma=1, mu=0)
   nx,ny,nc  : dimensions }
procedure SImg_Noise_Norm(cc:CSImg_Config; var fi: TSImg_FPImage; nx,ny,nc:longint);

{ Draw 3-dimensional Gauss curve (2-dimensional if nc=1)
   nx,ny,cs  : dimensions
   sx,sy,sc  : sigma
   re        : radius exponent (1 for Gauss curve) }
procedure SImg_Gauss(cc:CSImg_Config; var fi: TSImg_FPImage; nx,ny,nc:longint; sx,sy,sc, re:double);

{ Draw 3-dimensional Mexican hat curve (2-dimensional if nc=1)
   nx,ny,cs  : dimensions
   sx,sy,sc  : sigma }
procedure SImg_MexHat(cc:CSImg_Config; var fi:TSImg_FPImage; nx,ny,nc:longint; sx,sy,sc:double);


//define[configvals]
//include[configure.inc]

{*************************************************************************}
{******* implementation **************************************************}
{*************************************************************************}
implementation

uses compatun,errors,baseunix,unix,termio,syscall,math,dateutils,ppmdb,linux;

{$LINKLIB fftw3_threads}
{Required libraries by libfftw3}
{$LINKLIB m}

const dirsep='/';

function SImg_GetLineInfo(addr:pointer):ansistring; forward;
function SImg_GetStackTrace(frame:pointer):ansistring; forward;

//include[utils.inc]
//include[messages.inc]
//include[thread.inc]
//include[configclass.inc]
//include[system.inc]
//include[meta-info.inc]
//include[img-sys.inc]
//include[fp-sys.inc]
//include[rd-sys.inc]
//include[fp-base0.inc]
//include[fp-base1.inc]
//include[rawcorrect.inc]
//include[filter.inc]
//include[denoise.inc]
//include[deconv.inc]
//include[interpolate.inc]
//include[starsubtract.inc]
//include[stars.inc]
//include[info.inc]
//include[bayer.inc]
//include[ff.inc]
//include[ff-pnm.inc]
//include[ff-sif.inc]
//include[ff-tsif.inc]
//include[ff-fits.inc]
//include[ff-sbig.inc]
//include[fft-sys.inc]
//include[fft-base.inc]
//include[fourier.inc]
//include[histogr.inc]
//include[is-sys.inc]
//include[is-base0.inc]
//include[raw.inc]
//include[sources.inc]
//include[alignment.inc]
//include[stackdb.inc]
// include[getshift.inc]
//include[colors.inc]
//include[colorize.inc]
//include[resize.inc]
//include[approx.inc]
//include[hdr.inc]
//include[tiles.inc]

//include[fonts.inc]
//include[draw01.inc]
//include[draw02.inc]

{####### init #########}
const SImg_InitFinished : boolean = false;

initialization
    SImg_InitConfig; // calls SImg_InitMessages;
    SImg_InitThread;
    SImg_InitSys;
    SImg_InitImgSys;
    SImg_InitFFT;
    SImg_InitFinished:=True;

{####### exit #########}
finalization
    if SImg_InitFinished then
    begin
//ifdef[debug]
        SImg_DebugLevel:=0;
        if exitcode=SImg_ExternalExitCode then SImg_DebugMessage('Finalization: Exit by external interrupt signal (Ctrl-C)');
//endif
        SImg_DoneFFT;
        if exitcode<2 then SImg_DoneImgSys(exitcode);
        SImg_DoneSys(exitcode);
        SImg_DoneConfig;
        SImg_DoneMessages(exitcode);
    end
    else writeln(stderr,'Fatal: Error during initialization');
end.
